# BI ALERT PROTECT TOOL - FOR BLUE IRIS ALERTS
#
# THIS TOOL ENSURES THAT PROTECTED BLUE IRIS ALERTS CAN SURVIVE DATABASE/REGENERATE ACTIONS
# THIS MEANS...
#    * THE ALERT'S DAT AND JPG FILES HAVE THEIR READ-ONLY ATTRIBUTES SET
#    * THE METADATA IN THE ALERT'S JPG FILE MATCHES THE TEXT IN THE ALERT'S DATABASE MEMO
#
# THIS TOOL FILTERS ALERTS BASED ON CAMERA, NAMING PATTERN, AND/OR ALERT MEMO
# AND PERFORM ANY/ALL OF THE FOLLOWING ACTIONS: 
#    * LIST (FOR INSPECTION BEFORE PERFORMING THE FOLLOWING ACTIONS)
#    * SET THE FILE READ-ONLY ATTRIBUTES FOR BOTH DAT & JPG FILES
#    * SET THE METADATA IN JPG FILES TO MATCH THE CURRENT TEXT IN THE ALERT'S DATABASE MEMO
#
# Version: 3
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V3 - 2024-03-23 - BREAKING CHANGE: must replace old user-settings file 
#                 - BREAKING CHANGE: Argument -Cams changed to -Cameras
#                 - CHANGE: Argument -Pattern changed to -Patterns (comma-delimited)
#                 - CHANGE: script no longer creates dummy DAT files (text files with *.dat extension)
#                 - NEW: added argument -AllAlerts to process Alerts regardless of 'Protected' flag
#                 - NEW: added argument -Flags to filter Alerts by DB flag, like -Flags 'F','P','!C';
#                   a '!' prefix (like '!C') means 'Not' Confirmed; multiple items must all be true
#                 - NEW: output column 'Flags'; now shows if alerts are 'Confirmed' in the database  
#                 - NEW: added output columns 'DatImgNo' & 'DatImgMemo' that identify the confirmed 
#                   image(s) in existing Blue Iris DAT file (these appear only in the output.csv file)
#                 - NEW: added output column 'HasZD'; shows if alert's db record has a date = zero  
#                 - NEW: the script now reports an issue if an Alert has an empty '' or 'NA' DB memo
#                   while also having a non-blank confirmed image memo in the DAT file (if it exists)
#                 - NEW: the script now repairs Alerts w/ empty '' or 'NA' memos by using the 
#                   memo string from the confirmed image in the DAT file (it it exists)
#                 - NEW: the script can now repair Alerts w/ zero dates - from either a existing DAT or
#                   JPG file's date (if both files exist, priority if given to the DAT file)
#                 - NEW: to aid in the visual identification of repaired Alerts; the script now prepends 
#                   a caret (^) character to the Alert's memo text; this feature may be customized 
#                   via user settings '$apply_custom_memo_mark' & '$custom_memo_mark' 
#                 - NEW: added argument -ClearMarks to facilitate the batch removal the memo caret (^) 
#                   (or '$custom_memo_mark') from all protected Alerts
#                 - NEW: script provides feedback if commandline includes unsupported arguments 
#                 - NEW: the script now tries to grab the dummy/replacement JPG from: 1) the 
#                   DAT file (if it exists & is valid), then 2) the Alert's BVR clip file (if it exists), 
#                   then 3) the Alert's database thumbnail image. 
#                   If all 3 methods fail, the script creates a solid yellow dummy/replacement JPG.
#                   For DAT files, the AI-confirmed image is used (if it exists), otherwise the middle 
#                   image is used.
#                   For BVR files, the image used is grabbed at the Alert's 'offset'; furthermore, these
#                   images may be customized via user settings '$bvr_jpg_image_no' & '$bvr_jpg_height'
#                 - NEW: When grabbing JPGs from BVR files, the script supports multiple Blue Iris 
#                   'New' folders via user-setting '$bi_new_folder_paths'
#                 - NEW: the script can now delete database records for Alerts with zero-dates and 
#                   no JPG or DAT file, via user setting '$delete_unfixable_zerodate_alerts'
#                 - NEW: added various simulator arguments (-SimX) for testing -Fix* arguments 
#                 - IMPROVED: the script now backs up the Alert's JPG file only the FIRST time the 
#                   Alert is repaired (to folder '\_backup_jpgs'); this means that the Alert's 
#                   original memo should always be recoverable by: 1) copying the backup JPG over the 
#                   script's replacement/dummy JPG, and 2) executing a database Delete/Regenerate action
#                 - IMPROVED: the overlay message on replacement JPGs is now optional via user setting 
#                   '$add_jpg_overlay'; additionally, the overlay may be be customized via user settings
#                   '$jpg_overlay_font' & '$jpg_overlay_location'
#                 - IMPROVED: the script is now Blue Iris version aware
#                 - IMPROVED: function BIJsonAlertListParse() now more tolerant of database issues
#                 - IMPROVED: updated function BIJsonCall()
#                 - IMPROVED: updated function FindBytes()
#                 - FIXED: the script now sets the CreationTime & LastAccessTime attributes for 
#                   dummy/replacement JPG & dummy DAT files; previously, the script set only the 
#                   LastWriteTime, which could result in incorrect database timestamps when the 
#                   protected Alerts were recovered during a Blue Iris Database/Regenerate action.
#                   (Note: The 'Date taken' Exif metadata value for dummy/replacement JPGs is 
#                   automatically set by Blue Iris)
#                 - FIXED: fixed handling of 'Invalid session' responses in function BIJsonCall()
#                 - FIXED: fixed handling of large JSON 'alertlist' cmd responses
#                 - FIXED: fixed an untrapped error when the Blue Iris service is not running
#                 - FIXED: fixed occasional false reporting of issues with the script's management 
#                   of the CSV file archive
# V2 - 2023-11-24 - NEW: added progress bar for alertlist parsing (quicker console feedback when the 
#                   script starts)
#                 - NEW: added user setting '$disable_clip_maintenance' to avoid possible simultaneous
#                   Database/Repair actions
#                 - NEW: added user setting '$enable_bg_messages' for temporary use to verify background execution
#                 - IMPROVED: made posted messages to the Blue Iris logfile more informative; now includes
#                   script's runtime,  and counts of issues found & fixed
#                 - IMPROVED: argument -DevMode now outputs more timing information; demos the dummy JPG; 
#                   saves each camera's alertlist
#                 - FIXED: fixed script incorrectly detecting if running in background (via Task Scheduler, 
#                   or Blue Iris)
#                 - FIXED: fixed duplicate entries in the 'Issue(s)' column in the script's log file 
#                   ('fixed_alerts_log.csv')
# V1 - 2023-11-18 - initial release
#
# *** BEFORE USING THIS SCRIPT ***
# *** EDIT ALL REQUIRED VARIABLES IN THE ESSENTIAL COMPANION FILE '{this_script_name}_user_settings.ps1' ***
#
# USAGE: {this_script.name).ps1 [arg1]
# ARGS:  -NoInspect       = (switch) skip inspection of JPG Exif metadata
#        -Cameras         = (array) analyze protected Alerts for specified cameras (comma-delimited)
#        -Patterns        = (array) analyze protected Alerts with descriptor substring
#        -MemoTag         = (string) analyze protected Alerts with database memo aubstring ('tag')
#        -Flags           = (array) analyze protected Alerts with specified flags, like 'F','!C' (and 'P' w/ -AllAlerts)
#        -FilterNoJPG     = (switch) filter protected Alerts without a JPG file
#        -FilterNoDAT     = (switch) filter protected Alerts without a DAT file
#        -FilterNoBoth    = (switch) filter protected Alerts without a JPG file AND without a DAT file
#        -FilterNoRO      = (switch) filter protected Alerts without a read-only JPG DAT or DAT file
#        -FilterNoMatch   = (switch) filter protected Alerts when JPG Exif metadata does not match the alert memo
#        -FilterNoMemo    = (switch) filter protected Alerts that have empty '' or 'NA' memos in the Blue Iris database
#        -FilterZeroDates = (switch) filter protected Alerts that have a zero-date (timestamp) in the Blue Iris database
#        -FilterIssues    = (switch) filter protected Alerts for all issues (-FilterNoJPG, -FilterNoBoth, -FilterNoRO, -FilterNoMatch)
#        -SortAsc         = (switch) sort ascending (default is descending)
#	     -FixJpgExif      = (switch) force Alert JPG file's metadata to match the Alert's memo (in the db)
#        -FixFiles        = (switch) when neither a JPG or DAT exist, create a dummy JPG and DAT file
#	     -FixReadOnly     = (switch) set read-only attribute for JPG & DAT files (for all or filtered alerts)
#        -FixMemos        = (switch) fix Alert database records that have empty '' or 'NA' memos (using DAT confirmed img)
#	     -FixZeroDates    = (switch) fix Alert database records that have a zero-date (timestamp); uses DAT date; DAT file must exist
#	     -FixAll          = (switch) performs all fixes: -FixJpgExit, -FixFiles & -FixReadOnly
#        -SetFlag         = (switch) set the database flag 'Flag' for all protected Alerts
#        -SourcePath      = (string) alternate path to *.jpg and *.dat files (typically the Blue Iris Alerts folder path);
#                            overrides default user-setting '$bi_alerts_folder_path'
#        -ClearMarks      = (switch) remove the script's $custom_memo_mark ('^', by default) from all Alerts
#        -Quiet           = (switch) disable Console status messaging EXCEPT results table, warnings and return object
#        -HideTable       = (switch) disable display of table of all DAT files found/processed
#        -NoBiLog         = (switch) disable/override logging to the Blue Iris logfile
#        -Help            = (switch) disable display of Console usage guide
#        -ShowCLI         = (switch) display output returned when the script is executed in the background
#        -DevMode         = (switch) display developer output
#        -NoRegen         = (switch) prevent script from auto-performing a Blue Iris db Repair/Regenerate action (w/ -DevMode only)
#        -AllAlerts       = (switch) allow script to work all alerts - not just protected ones (w/ -DevMode only)
# RETURNS: (arrays of arrays) @( 
#                                @(alert_labels), @(alert_flags), @(alert_hasjpgs), @(alert_hasdats), @(alert_attribs), @(alert_memos), @(jpg_exif_memos_match) 
#                             )
#          From the console ... shows a table summarizing all the Protected files found
#
# HOW TO AUTOMATE THIS SCRIPT VIA TASK SCHEDULER
#   Open Task Scheduler and create an appropriately named task with one trigger and one action:
#   Task Scheduler
#   'Triggers' tab:
#       'At startup'
#       'Daily' or 'One time' (e.g., at 12:05 AM, every 24 hours)
#       (Note: I find 'One time' to be more reliable)
#   'Actions' tab:
#       Program/script: powershell.exe
#       Add arguments: {your_powershell_scripts_path)\{script_file_name}.ps1 args
#       EXAMPLES:
#       To run script every week
#       ... C:\ps_scripts\BI_clip_usage.ps1
#
#
# DEVELOPER'S TO DO LIST:
# [ ] 
# [ ]

param (
	[Switch]$NoInspect,        # [OPTIONAL] skip inspection of JPG Exif metadata
	[String[]]$Cameras,        # [OPTIONAL] analyze protected Alerts for specified cameras; comma-delimited
	[String[]]$Patterns,       # [OPTIONAL] analyze protected Alerts with descriptor substring
	[String]$MemoTag,          # [OPTIONAL] analyze protected Alerts with database memo aubstring ('tag')
	[String[]]$Flags,          # [OPTIONAL] analyze protected Alerts Alerts w/ specified flags, like 'F','!C'"
	[Switch]$FilterNoJPG,      # [OPTIONAL] filter protected Alerts without a JPG file
	[Switch]$FilterNoDAT,      # [OPTIONAL] filter protected Alerts without a DAT file
	[Switch]$FilterNoBoth,     # [OPTIONAL] filter protected Alerts without a JPG file AND without a DAT file
	[Switch]$FilterNoRO,       # [OPTIONAL] filter protected Alerts without a read-only JPG DAT or DAT file
	[Switch]$FilterNoMatch,    # [OPTIONAL] filter protected Alerts when JPG Exif metadata does not match the alert memo
	[Switch]$FilterNoMemo,     # [OPTIONAL] filter protected Alerts that have empty '' or 'NA' memos in the Blue Iris database
	[Switch]$FilterZeroDates,  # [OPTIONAL] filter protected Alerts that have a zero-date (timestamp) in the Blue Iris database
	[Switch]$FilterIssues,     # [OPTIONAL] filter protected Alerts for all issues (-Filter*)
	[Switch]$SortAsc,          # [OPTIONAL] sort ascending (default is descending)
	[Switch]$FixJpgExif,       # [OPTIONAL] force Alert JPG file's metadata to match the Alert's memo (in the database)
	[Switch]$FixFiles,         # [OPTIONAL] when neither a JPG or DAT exist, create a dummy JPG and DAT file
	[Switch]$FixReadOnly,      # [OPTIONAL] set read-only attribute for filtered DAT files and their associated JPG files
	[Switch]$FixMemos,         # [OPTIONAL] fix Alert db records that have empty '' or 'NA' memos (using DAT confirmed img)
	[Switch]$FixZeroDates,     # [OPTIONAL] fix Alert db records that have a zero-date (timestamp); uses DAT date; DAT file must exist
	[Switch]$FixAll,           # [OPTIONAL] performs all fixes: -FixJpgExit, -FixFiles & -FixReadOnly, 
	[Switch]$SetFlag,          # [OPTIONAL] set the 'Flag' flag for all protected Alerts
	[String]$SourcePath,       # [OPTIONAL] Blue Iris Alerts folder path
    [Switch]$ClearMarks,       # [OPTIONAL] remove the script's $custom_memo_mark from all Alerts; ignored with -Fix* args
	[Switch]$Quiet,            # [OPTIONAL] disable Console status messaging EXCEPT results table, warnings and return object
	[Switch]$HideTable,        # [OPTIONAL] disable display of table of all DAT files found/processed
	[Switch]$NoBiLog,          # [OPTIONAL] disable/override logging to the Blue Iris logfile
	[Switch]$Help,             # [OPTIONAL] disable display of Console usage guide
	[Switch]$ShowCLI,          # [OPTIONAL] display output returned when the script executed in the background
	[Switch]$DevMode,          # [OPTIONAL] display developer output
	[Switch]$NoRegen,          # [OPTIONAL] prevent auto-performing a Blue Iris database Repair/Regenerate action (w/ -DevMode only)
	[Switch]$AllAlerts,        # [OPTIONAL] allow script to work all alerts - not just protected ones
	[Switch]$Sim1,  # [OPTIONAL] remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
	[Switch]$Sim2,  # [OPTIONAL] remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response (per iteration)
	[Switch]$Sim3,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ all 'memo'='NA' params
	[Switch]$Sim4,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ 3 random 'memo'='NA' params (per iteration)
	[String]$Sim5,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ zero-date (timestamp) values; arg value is CamName
	[Int]$Sim6     # [OPTIONAL] simulate zero-date (timestamp); adds/subtracts $Sim6 seconds; Works only if output is < 3 Alerts
)

# Verify that singular (ambiguous) arguments work as well
#Write-Host "FixFiles = $FixFiles"
#Write-Host "FixMemos = $FixMemos"
#Write-Host "FixZeroDates = $FixZeroDates"

Set-StrictMode  -Version 3
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

# GATHER SCRIPT INFO 
#$ScriptName = $MyInvocation.MyCommand.Name
#$ScriptName = (Get-Variable MyInvocation).Value.MyCommand.Name
$ScriptName = Split-Path -leaf $PSCommandpath
$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
$ScriptBaseName = $ScriptName.Replace('.ps1', '')
$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

# For debugging
#Write-Host "ScriptName = $ScriptName"
#Write-Host "ScriptPath = $ScriptPath"
#Write-Host "ScriptBaseName = $ScriptBaseName"
#Write-Host "ScriptPath = $ScriptPath"

# INITIALIZE VARS
$PsVersion = $PSVersionTable.PSVersion.Major
# For testing
#Write-Host "ScriptName = $ScriptName"
#Write-Host "PsVersion = $PsVersion"

# NOTE: Scripts running from the following hosts have process names as follows:
#       Task Scheduler is 'svchost', Powershell-ISE is 'explorer', Windows Terminal is 'WindowsTerminal'
#       https://stackoverflow.com/a/65894924
$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
# ... create lookup table (used by script log ($app_path\fixed_alerts_log.csv)
$HostLabels = @{}
$HostLabels['svchost'] = 'TaskScheduler'
$HostLabels['explorer'] = 'Powershell or Powershell-ISE'
$HostLabels['WindowsTerminal'] = 'WindowsTerminal'

# For testing
#Write-Host "ParentProcess = $ParentProcess"

# NOTE: Scripts called by another script have a 'PSCommandPath'.
#       Method 'PSCommandPath' returns null if script is executed via the console
#       https://stackoverflow.com/a/23370618/13499984
$CallingScript = $MyInvocation.PSCommandPath

# For testing
#Write-Host "CallingScript = $CallingScript"

# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
$fromConsole = !(($ParentProcess -eq 'svchost' -or $ParentProcess -eq 'BlueIris'))

if ($DevMode) {
	$fromConsole = $true
}

# Get the current server domain (hostname) and username
# Used for background messaging via msg.exe
# ... tip: from the console, you can type 'hostname' or 'whoami'
$server_name = (whoami).Split('\')[0]
$server_username = (whoami).Split('\')[1]

# For testing
#msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nPowerShell version is... $PsVersion`nParent process is... '$ParentProcess'`nCalling script name... $CallingScript"
#msg.exe $server_username /TIME:5 "From script... $ScriptName`nfromConsole = $fromConsole"

# CHECK IF SCRIPT'S PATH IS IN ENVIRONMENT 'SYSTEM PATH' VARIABLE
$EnvPath = [Environment]::GetEnvironmentVariable('Path', 'Machine')
if ($null -ne $EnvPath) {
	if (!($EnvPath -split ';' -contains $ScriptPath)) {

		# For debugging
		#Write-Host "ScriptPath = $ScriptPath"
		#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

		return "ERR/this script's path '$ScriptPath' does not exist in the Environment System Path variable"
		
		# Attempt repair
		# NOTE: when I tried this it added MANY other/duplicate entries
		#[Environment]::SetEnvironmentVariable("Path", $EnvPath + ";$ScriptPath", "Machine")
		#Write-Warning "Added this script's path '$ScriptPath' to the Environment Path"
	}
} else {
	Write-Warning "Unable to inspect Environment variable; be sure script's path is in the Environment Path"
}
# For debugging
#Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

# For debugging
#Write-Host "Entered script: $ScriptName"	
#Write-Host ("Cameras type = " + $Cameras.GetType())
#Write-Host "Cameras = $Cameras"
#Write-Host "Pattern = $Pattern"
#Write-Host "NoInspect = '$NoInspect'"
#Write-Host "SourcePath = '$SourcePath'"
#Write-Host "FixMemos = FixMemos"
#Write-Host "MemoTag = $MemoTag"
#Write-Host "SortAsc = $SortAsc"
#Write-Host "Help = $Help"
#Write-Host "HideTable = $HideTable"
#Write-Host "ShowCLI = $ShowCLI"
#Write-Host "DevMode = $DevMode"
#Write-Host "PsVersion = $PsVersion"

# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
#Write-Host "Extra Args w/ values = $Args"
$cmdline_args_extra = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
if ($cmdline_args_extra) {
	$cmdline_args_extra = [String[]]$cmdline_args_extra
} else {
	$cmdline_args_extra = @()
}
#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
#Write-Host "Cmdline Args Extra = $cmdline_args_extra"
if ($cmdline_args_extra.Count -gt 0) {
	if ($fromConsole) {
		Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_extra -Join ', ') + " ]") -ForegroundColor Green
	}
	
}

# ==============================================================
# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	$err_msg = "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
	return $err_msg
}

# VALIDATE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
	# ... Handle REQUIRED user settings
	$variable_test = Get-Variable -Name bi_alerts_folder_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_alerts_folder_path"
		return $err_msg
	} else {
		if ($bi_alerts_folder_path.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-settings file setting `$bi_alerts_folder_path is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_new_folder_paths -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_new_folder_paths"
		return $err_msg
	} else {
		$bi_new_folder_paths_ary = $bi_new_folder_paths -Split ','
		if ($bi_new_folder_paths_ary[0].StartsWith('YOUR')) {
			$err_msg = "ERR/user-settings file setting `$bi_new_folder_paths is not set"
			return $err_msg
		}
	}
	$i = 0
	foreach ($item in $bi_new_folder_paths_ary) {
		$bi_new_folder_paths_ary[$i] = $item.Trim().TrimEnd('\')
		if (!(Test-Path -LiteralPath "$item")) {
			$err_msg = "ERR/user-setting `$bi_new_folder_paths contains a non-existent path: '$item'"
			return $err_msg
		}
		$i++
	}
	# For debugging
	#Write-Host ("bi_new_folder_paths_ary = " + ($bi_new_folder_paths_ary | ConvertTo-Json))
	if ($bi_new_folder_paths_ary.Count -gt 1) {
		if ($DevMode) {
			if ($fromConsole) {
				Write-Host "(dev mode) Blue Iris appears to be using MULTIPLE 'NEW' FOLDERS" -ForegroundColor Cyan
			}
		}
	}

	$variable_test = Get-Variable -Name bi_db_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		#$bi_db_path = "C:\BlueIris\db"  # DEFAULT
		$err_msg = "ERR/user-settings file is missing setting `$bi_db_path"
		return $err_msg
	} else {
		if ($bi_db_path.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-settings file setting `$bi_db_path is not set"
			return $err_msg
		}
	}
	$bi_db_path = $bi_db_path.TrimEnd('\')

	$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$app_path"
		return $err_msg
	} else {
		if ($app_path.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-settings file setting `$app_path is not set"
			return $err_msg
		}
	}
	$app_path = $app_path.TrimEnd('\')

	$variable_test = Get-Variable -Name bi_webserver -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_webserver"
		return $err_msg
	} else {
		if ($bi_webserver.StartsWith('YOUR_')) {  
			$err_msg = "ERR/user-settings file setting '`$bi_webserver'is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_user_name"
		return $err_msg
	} else {
		if ($bi_user_name.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-settings file setting `$bi_user_name is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$bi_user_pw"
		return $err_msg
	} else {
		if ($bi_user_pw.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-settings file setting `$bi_user_pw is not set"
			return $err_msg
		}
	}

	# ... Handle OPTIONAL user settings
	$variable_test = Get-Variable -Name apply_custom_memo_mark -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$apply_custom_memo_mark = $true
	}
	$variable_test = Get-Variable -Name custom_memo_mark -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$custom_memo_mark = '^'
	}
	$variable_test = Get-Variable -Name append_memo_tag -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_memo_tag = $true
	}
	$variable_test = Get-Variable -Name delete_unfixable_zerodate_alerts -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$delete_unfixable_zerodate_alerts = $false
	}
	$variable_test = Get-Variable -Name bvr_jpg_image_no -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bvr_jpg_image_no = 1
	} else {
		if ($bvr_jpg_image_no -lt 1 -or $bvr_jpg_image_no -gt 10) {
			$err_msg = "ERR/user-setting '`$bvr_jpg_image_no' must be in range 1 to 10"
			return $err_msg
		}
	}
	$variable_test = Get-Variable -Name bvr_jpg_height -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bvr_jpg_height = 0
	}
	$variable_test = Get-Variable -Name add_jpg_overlay -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$add_jpg_overlay = $true
	}
	$variable_test = Get-Variable -Name jpg_overlay_font -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$jpg_overlay_font = 'Courier New'
	}
	$variable_test = Get-Variable -Name jpg_overlay_location -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$jpg_text_loc = 'Center'
	} else {
		$locations_ary = @('TopLeft','TopMiddle','TopRight','BottomLeft','BottomMiddle','BottomRight','Center')
		if ($locations_ary -contains $jpg_overlay_location) {
			$jpg_text_loc = $jpg_overlay_location
		} else {	
			$jpg_text_loc = 'Center'  # Default
			if ($FromConsole) {
				Write-Warning "unexpected value for user-setting `$jpg_overlay_location ('$jpg_overlay_location'); using default"
			}
		}
	}
	$variable_test = Get-Variable -Name auto_open_csv_file -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$auto_open_csv_file = $false
	}
	$variable_test = Get-Variable -Name auto_open_csv_file_in_def_app -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$auto_open_csv_file_in_def_app = $false
	}
	$variable_test = Get-Variable -Name auto_open_csv_file_min_rows -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$auto_open_csv_file_min_rows = 0
	}
	$variable_test = Get-Variable -Name timestamp_output_csv_files_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$timestamp_output_csv_files_flg = $true
	}
	$variable_test = Get-Variable -Name archive_lifetime_days -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$archive_lifetime_days = 365
	}
	$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_flg = $false
	}
	$variable_test = Get-Variable -Name post_w_creds_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_w_creds_flg = $false
	}
	$variable_test = Get-Variable -Name post_to_BI_logfile_fixes_only_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_fixes_only_flg = $false
	}
	$variable_test = Get-Variable -Name post_to_BI_logfile_in_background_only_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_in_background_only_flg = $false
	}
	$variable_test = Get-Variable -Name disable_clip_maintenance -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$disable_clip_maintenance = $false
	}
	$variable_test = Get-Variable -Name enable_bg_messages -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$enable_bg_messages = $false
	}
	if ($Quiet) {
		if ($enable_bg_messages) {
			$enable_bg_messages = $false
			if ($fromConsle) {
				Write-Warning "Background messaging disabled when using the -Quiet argument"
			}
		}
	}
}
# ==============================================================

function BIJsonCall() {
# PURPOSE:   To get a Blue Iris JSON interface response from the Blue Iris web server.
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -JsonCmd        = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
#            -ForceFlg       = [OPTIONAL] (boolean) forces new login (gets a new authenticated session id)
#            -RetryNo        = [OPTIONAL] (int) if -gt 0, this call is a retry attempt
#            -QuietFlg       = [OPTIONAL] (boolean) $true suppresses verbose output to the console
#            -FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS:   (object) PowerShell PSCustomObject
# REFS:      1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
#            2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
#            3) https://ipcamtalk.com/threads/json-interface.21491/
#            4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
# NOTES:
# 1) Saves the BI API session id to a file for reuse (to avoid unnecessary repeated 
#    logins while the session is still active); the file stored in folder 
#    'c:\temp\$ScriptBaseName' (created if necessary)
# 2) In the event that the JSON call fails with reason 'Invalid session',
#    the script will automatically attempt to repeat the call up to $max_retries
#
# Version: 3 (2024-02-17)
# Created by jaydeel at ipcamtalk.com

	param (
		[Hashtable]$JsonCmd,
		[Boolean]$ForceFlg,
		[Int]$RetryNo,
		[Boolean]$QuietFlg,
		[Boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name() ..."	
	#Write-Host "`Entered function $fn_name()`n-----------------------------"
	#Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"

	# For debugging
	#Write-Host "$fn_name() retry_cntr = $retry_cntr"	

	# System vars
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	$PsVersion = $PSVersionTable.PSVersion.Major
	# Get the current server domain (hostname) and username
	# Used for background messaging via msg.exe
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$server_name = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	function encryptMD5($str) {
	# PURPOSE: Get MD5 hash for a string.
	# ARGS:    $str = (string) string to be encrypted
	# RETURNS: (string) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
	# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
	# NOTES:   (from REF)
	#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
	#	However, PowerShell offers a cmdlet to generate MD5 hash for a file: Get-FileHash. 
	# 	The cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
	# UTILITY: https://www.md5hashgenerator.com/
	# ... Example: 'MD5string' returns '0753a0e20c9027ccab70f4291a21aa08'

		$stringAsStream = [System.IO.MemoryStream]::new()
		$writer = [System.IO.StreamWriter]::new($stringAsStream)
		$writer.write($str)
		$writer.Flush()
		$stringAsStream.Position = 0
		$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
		
		# For debugging
		#$str = $hash_obj | Out-String
		#Write-Host "hash_obj = $str"

		return $hash_obj.Hash
	}
	
	# MAX RETRIES FOR JSON CMD
	# ... This function will recursively call itself if a JSON cmd's previous
	#     $json_response.result value is 'fail' and the reason is 'Invalid session'
	$max_retries = 3
	# ... Wait time before attempting a 'Invalid session' recursive call 
	$retry_delay_ms = 100
	
	if (!$RetryNo) {
		$retry_cntr = 0
	} else {
		$retry_cntr = $RetryNo
	}

	# Abort if we've exceeded the retries 'Invalid session' break variable
	if ($retry_cntr -ge $max_retries) {
		$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' unable to create a valid session"
		return $err_msg
	}
		
	# Show progress bar as a heads-up visualization that retries are being attempted
	<# if ($retry_cntr -gt 0) {
		if ($fromConsole) {
			$WriteProgressParams = @{
				Activity         = "Retrying BI JSON Call..."
				Status           = "Progress: retry $retry_cntr of $max_retries max"
				PercentComplete  = (($retry_cntr / $max_retries) * 100)
			}
			Write-Progress @WriteProgressParams
			# Note: If loop is too fast, you might need a delay to get the progress bar to show
			Start-Sleep -Milliseconds 2000				
		}
	} #>

	# HANDLE FILES FOR SAVED SESSION IDS AND RETRY COUNTER
	# ... Perform existence test for the *.txt file used to save session ids
	#     Create its path if it does not exist
	#     NOTE: if a path is not specified, the default Windows location is 'C:\Users\<username>"
	$path = "c:\temp\$ScriptBaseName"
	if (!(Test-Path -LiteralPath $path)) {
		New-Item -ItemType Directory -Force -Path $path
	}
	# ... NOTE: we use a different filename when running the script via different processes;
	#     this avoids collisions with scripts running via:
	#     'scvhost' (Task Scheduler), 'WindowsTerminal', 'explorer' (Powershell ISE)
	$base_name = "$path\bi_session_id"
	$session_filespec = $base_name + "_" + $ParentProcess + ".txt"
	$found_saved_session_file_flg = (Test-Path -LiteralPath "$session_filespec")

	# PARSE COMMANDLINE ARGS
	if ($JsonCmd -eq $null) {
		$err_msg = "ERR/expecting an argument: JsonCmd (PowerShell hashtable)"
		return $err_msg
	} else {
		# For debugging
		#Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
	}
	# ... Save original JsonCmd (in case we need to make a recursive call)
	$json_cmd_orig = $JsonCmd
	# ... Override reuse of saved session_id
	if ($ForceFlg) {
		$found_saved_session_file_flg = $false
	}

	# ========================
	# PERFORM API CALLS

	# Make the URL
	$uri = "http://" + $script:bi_webserver + "/json";

	# For debugging
	#Write-Host "$fn_name() uri = $uri"

	# Flag 'logout' JSON cmd requests
	# ... NOTE: All JSON 'logout' cmds should be performed by the calling script!
	if ($JsonCmd.cmd -eq 'logout') {
		$logging_out_flg = $true
	} else {
		$logging_out_flg = $false
	}
	
	# For debugging
	if ($logging_out_flg) {
		#Write-Host "$fn_name() detected JSON cmd to log out of BI API session"
	} else {
		#Write-Host "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
		#msg.exe * /TIME:5 "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
	}

	# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
	$dev_uuid = $null  # Default
	$json_cmd_keys = $JsonCmd.Keys
	if ($json_cmd_keys -contains 'uuid') {
		if ($JsonCmd.uuid -ne $null) {
			$dev_uuid = $JsonCmd.uuid
		}
	}

	# For debugging
	if ($dev_uuid) {
		#Write-Host "$fn_name() dev_uuid = $dev_uuid"
	}

	# EXECUTE API CALL
	
	if (!$found_saved_session_file_flg) {
		
		# ================================================================
		# 1.1 PRELIMINARY LOGIN
		
		# Create, then encode the JSON 'login' command
		$parameters = @{
			cmd = 'login'
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-host "$fn_name() (1.1) login #1 json_params... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$jsonResponseObj"
		
		# Get the session id
		$session = $jsonResponseObj.session
		
		# ================================================================
		# 1.2 AUTHENTICATE THE LOGIN

		# Create MD5 hash-encoded credentials string
		$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
		$response_hash = encryptMD5 $response_str

		# Create, then encode a modified JSON 'login' command
		if ($dev_uuid) {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
				uuid = $dev_uuid
			}
		} else {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
			}
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-Host "$fn_name() (1.2) authorizing json login... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) authorizing json login ... `n$json_params"
		
		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#Write-Host ("$fn_name() (1.2) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) jsonResponse... `n$jsonResponseObj"
		
		# Handle failure
		if ($jsonResponseObj.result -eq 'fail') {
			$err_msg = "ERR/unable to authenticate Blue Iris JSON interface`n"
			return $err_msg
		}
		
		# Get the authenticated session id
		$auth_session_id = $jsonResponseObj.session

		# For debugging
		#msg.exe $server_username /TIME:5 "$fn_name() auth_session_id... $auth_session_id"
	
		# Save the authenticated session id
		# NOTE: the session id is always saved, whether used or not
		$auth_session_id | Set-Content "$session_filespec" -NoNewline

	} else {
	
		# Retrieve the saved session id
		# KEY POINT: 
		#   Force typing the output of 'Get-Content' to [String] was essential when running script in
		#   in Powershell ISE. Otherwise the output was similar to an object.
		#   *** But in actuality it is more complicated ***
		#   Per https://stackoverflow.com/a/55881978/13499984
		#      When 'Get-Content' is assigned to a variable, its output is an array of .NET [String] objects, 
		#      each representing a line from the input file. However, Get-Content also decorates these string 
		#      objects with additional properties that provide helpful metadata about where the strings came from.
		$saved_session_id = [String](Get-Content -Path "$session_filespec")

		$auth_session_id = $saved_session_id

		# For debugging
		if (!$logging_out_flg) {
			#Write-Host ("$fn_name() saved_session_id type = " + $saved_session_id.GetType())
			#Write-Host "$fn_name() saved_session_id... $saved_session_id"
			#msg.exe * /TIME:5 "$fn_name() saved_session_id... $saved_session_id"
		}
		
	}

	# ================================================================
	# 1.3 EXECUTE THE REQUESTED JSON COMMAND

	if ($dev_uuid -eq $null) {
		
		# In the event that this JSON call fails with reason 'Invalid session',
		# the script will automatically attempt up to $max_retries
	
		# Modify, then encode the requested JSON login command
		$JsonCmd.session = $auth_session_id
		$json_params = $JsonCmd | ConvertTo-Json

		# Execute the JSON request
		$params = @{
			Uri                = $uri
			#Headers            = @{"Content-Type"  = "application/json"}
			Method             = 'POST'
			#Body               = $JsonCmd | ConvertTo-Json
			Body               = $json_params
			ContentType        = 'application/json'
		}
		
		$json_response_obj = Invoke-RestMethod @params

		# Alternate syntaxes (pick one)
		#$json_response_obj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
		#$json_response_obj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#if ($JsonCmd.cmd -match 'update') {
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "'")
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "' ...`n" + $json_params)
			#Write-Host ("$fn_name() json_response_obj = `n" + ($json_response_obj | ConvertTo-Json))
		#}

		if ($json_response_obj.result -eq 'fail') {
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response.result = fail")
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response = `n" + ($json_response_obj | ConvertTo-Json))

			# Try to determine reason
			if ($json_response_obj.data) {

				$data_obj = $json_response_obj.data

				# For debugging
				#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data = " + ($data_obj | ConvertTo-Json))
				
				if ($data_obj.reason) {
					$reason = $data_obj.reason
				
					# For debugging
					#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data.reason = $reason")

					if ($reason -match 'Invalid session') {
						# Update retry counter
						$retry_cntr++
						
						# RECURSIVE CALL
						#Start-Sleep -Milliseconds 500
						$json_response_obj = `
							BIJsonCall `
							-JsonCmd $JsonCmd `
							-ForceFlg $true `
							-RetryNo $retry_cntr `
							-QuietFlg $QuietFlg `
							-FromConsoleFlg $FromConsoleFlg
						
					} elseif ($reason -match 'Access denied') {
						$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' requires an Administrator account"
						return $err_msg
						
					} else {
						$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed with reason '" + $reason + "'"
						return $err_msg
					}
					
				} else {
					$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed without a reason"
					return $err_msg
				}
			}
		} 
	}

	# For debugging
	if ($json_response_obj) {
		#Write-Host ("$fn_name() (1.3) jsonResponse type = " + $json_response_obj.GetType())
		#Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $json_response_obj"
	} 

	# Retain for error testing
	#$json_response_obj = $null
	#$json_response_obj = ''

	if ($json_response_obj -eq $null) {
		$err_msg ="ERR/null response from Blue Iris JSON interface"
		return $err_msg
	} 
	if ($json_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris JSON interface"
		return $err_msg
	}
	if ($json_response_obj.result -ne $null) {

		# For debugging
		#Write-Host ("$fn_name() (1.3) json_response_obj.result... " + ($json_response_obj.result))
		
		if ($json_response_obj.result -eq 'fail') {
			
			# Show the contents of the $json_response_obj 'data' key

			$json_data_obj = $json_response_obj.data
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) json_response_obj.data... " + ($json_response_obj.data))

			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jdo_str = $json_data_obj | ConvertTo-Json -Depth 16
			} else {
				$jdo_str = $json_data_obj | ConvertTo-Json
			}
			
			# 'Minify' the json string.
			if ($jdo_str) {
				if (!($jdo_str -match 'null')) {
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is 'result:fail'"
					return $err_msg
				} else {					
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is null (DB record not found)"
					return $err_msg
				}
			}
		}
	}

	return $json_response_obj

}

function IsBiVersion() {
# PURPOSE:   To confirm if the current Blue Iris version exceeds a specified version
# ARGUMENTS: $CompareVersion = [REQUIRED] (string) compare version, '5.7.2.3'
#            Note: supports version strings 'X.000.000.000' to 'X.999.999.999'            
# REQUIRES:  function BIJsonCall()
	
	param (
		[String]$CompareVersion,
		[Boolean]$QuietFlg,       # used by BIJsonCall()
		[Boolean]$FromConsoleFlg  # used by BIJsonCall()
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "Entered $fn_name()..."
	#Write-Host "$fn_name() CompareVersion = $CompareVersion"
	
	$rtn_flg = $false  # Assume failure

	# GET CURRENT VERSION
	$jsonCmd = @{
		cmd = 'login'
	}
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))

	# Abort if GET failed
	if ($json_response_obj -match 'ERR/') {
		if ($FromConsoleFlg) {
			Write-Host "$fn_name() ERR/JSON request failed"
			Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
		}
		return $false
	}
	if ($json_response_obj.result -eq 'fail') {
		$x  = $json_response_obj.result
		if ($FromConsoleFlg) {
			Write-Host "$fn_name() ERR/JSON request failed ('result' = fail)"
			Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
		}
		return $false
	}

	# Extract data from the JSON command response
	if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

		$data_obj = $json_response_obj.data

		# Get the current version
		$bi_version = $data_obj.version
		
		# Make comparator integer arrays
		[int[]]$my_ver_ary = $CompareVersion.Split('.')
		[int[]]$bi_ver_ary = $bi_version.Split('.')

		# Make 'prime.subprime' version comparators, like 5.8 vs 5.7
		$my_ver_prime_dot = $my_ver_ary[0] + $my_ver_ary[1]/1000
		$bi_ver_prime_dot = $bi_ver_ary[0] + $bi_ver_ary[1]/1000

		# For debugging
		#Write-Host "bi_version = $bi_version"
		#Write-Host ("my_ver_ary = " + ($my_ver_ary | ConvertTo-Json))
		#Write-Host ("bi_version_ary = " + ($bi_ver_ary | ConvertTo-Json))
		#Write-Host "my_ver_prime_dot = $my_ver_prime_dot"
		#Write-Host "bi_ver_prime_dot = $bi_ver_prime_dot"
		
		if ($my_ver_prime_dot -eq $bi_ver_prime_dot) {

			# Make sub-subprime version comparators, 5.001 for 5.8.5.1, 999.020 for 5.8.999.20

			if ($my_ver_ary[2] -eq 0 -and $my_ver_ary[3] -eq 0) {
				$my_ver_sub = 0.000
			} else {
				$my_ver_sub = $my_ver_ary[2] + $my_ver_ary[3]/1000
			}

			if ($bi_ver_ary[2] -eq 0 -and $bi_ver_ary[3] -eq 0) {
				$bi_ver_sub = 0.000
			} else {
				$bi_ver_sub = $bi_ver_ary[2] + $bi_ver_ary[3]/1000
			}

			# For debugging
			#Write-Host ("my_ver_sub = " + ("{0:n3}" -f $my_ver_sub))
			#Write-Host ("bi_ver_sub = " + ("{0:n3}" -f $bi_ver_sub))

			$rtn_flg = ($bi_ver_sub -ge $my_ver_sub) 

		} else {
			$rtn_flg = ($bi_ver_prime_dot -ge $my_ver_prime_dot)
		}
	}

	return $rtn_flg
}

function BIHttpCall() {
# PURPOSE: Send Blue Iris HTTP Interface command to the Blue Iris web server.
# ARGS:    -HttpCmd  = (string) like '/admin?camera=x&trigger=1'
#          -DestPath = (string) full filespec for commands that downlaod JPGs
#                      like 'C:\temp/image.jpg'
# RETURNS: (string) web server plain text output,
#          or error message
#          or JPG file 
#
# Version: X
# Created by jaydeel at ipcamtalk.com

	param (
		[String]$HttpCmd,
		[String]$DestPath
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -HttpCmd = '$HttpCmd'"
	#Write-Host "$fn_name() -DestPath = '$DestPath'"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major

	# PARSE COMMANDLINE ARGS
	if ($null -eq $HttpCmd) {
		$err_msg = "$fn_name() ERR/expecting an argument: HttpCmd"
		return $err_msg
	}

	# Check for image URLS
	$image_url_flg = ($HttpCmd -match 'alerts' -or $HttpCmd -match 'thumbs' -or $HttpCmd -match 'clips')


	# ========================
	# PERFORM API CALLS

	# Make the URL
	if ($use_bi_creds) {
		$uri = "http://" + $bi_webserver + $HttpCmd + "&user=" + $bi_user_name + "&pw=" + $bi_user_pw
	} else {
		$uri = "http://" + $bi_webserver + $HttpCmd	
	}
	# For debugging
	#Write-Host "$fn_name() uri = '$uri'"

	# ================================================================
	# EXECUTE THE REQUESTED HTTP COMMAND 
	$web_response_obj = 'fail'
	try {
		if ($image_url_flg) {

			$web_response = Invoke-WebRequest -Uri $uri -Outfile $DestPath

			# REF: https://stackoverflow.com/a/22448954
			#$wc_obj = New-Object System.Net.WebClient
			#$web_response = $wc_obj.DownloadFile($uri, $DestPath)

		} else {

			#$web_response = Invoke-WebRequest -Uri $uri -UseBasicParsing
			$web_response = Invoke-WebRequest -Uri $uri

		}
	} 
	catch {

		# For debugging
		#Write-Host $error
		#Write-Host $_.ScriptStackTrace
		#Write-Host $PSitem

		Write-Host ("$fn_name() " + $_.Exception.Message) -ForegroundColor Red
		return 'fail'
	}

	if (!$image_url_flg) {
		# For debugging
		# REF: https://4sysops.com/archives/powershell-invoke-webrequest-parse-and-scrape-a-web-page/
		if ($web_response_obj) {
			#Write-Host ("Web Response type = " + $web_response_obj.GetType())
			#$web_response_obj| Get-Member
			#$web_response_obj.Content
			#$web_response_obj.Headers
			#$web_response_obj.StatusCode
			#$web_response_obj.StatusDescription	
		} 

		# For testing
		#$web_response_obj = $null
		#$web_response_obj = ''

		if ($null -eq $web_response_obj) {
			Write-Host "$fn_name() ERR/null response from Blue Iris HTTP interface" -ForegroundColor Red
			return 'fail'
		} 
		if ($web_response_obj -eq '') {
			Write-Host "$fn_name() ERR/empty response from Blue Iris HTTP interface" -ForegroundColor Red
			return 'fail'
		}
		If ($web_response_obj -match "<!DOCTYPE html>") {

			# For debugging
			#$HTML = New-Object -Com "HTMLFile"
			#$HTML.write([ref]$web_response_obj)
			#$TextOnly = $Html.body.innerText
			#Write-Host $TextOnly -ForegroundColor Cyan

			Write-Host "$fn_name() ERR/unexpected response" -ForegroundColor Red
			return 'fail'
		}
	}
	
	# For debugging
	#Write-Host $web_response_obj.Content
	
	return "success"

}

function BIJsonCamListParse() {
# PURPOSE:   Get Blue Iris web server JSON 'camlist' command response (for all or subset of BI cameras)
# PARAMS:    -CamProps = [OPTIONAL] (list) like 'bps,bps2', comma delimited
# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'camProp':value}
#		                                   @{camName:{[camProp1:value1],[camProp2:value2],...},...}"
#
# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
# 	"returns a list of cameras on the system ordered by group. Cameras not belonging to any
#	group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
#	the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
#	describing a camera or a camera group."
#
# Version: X
# Created by jaydeel at ipcamtalk.com

	param(
		[String]$CamProps,  # [OPTIONAL] list
		[Boolean]$QuietFlg,
		[Boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() CamProps = $CamProps"

	#Set-StrictMode  -Version 3
	#Start-Transcript  # default destination is C:\Users\{user}l\Documents

	# PARSE COMMANDLINE ARGS
	if ($CamProps) {
		$cam_props_ary = $CamProps -Split ','
		#msg.exe * /TIME:5 "camProps... $cam_props_ary"
	} else {
		$cam_props_ary = @()
	}
	#Write-Host "cam_props_ary = $cam_props_ary"

	# PERFORM API CALLS
	$jsonCmd = @{
		cmd = 'camlist'
	}
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
	#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response... $json_response_obj"

	# Abort if GET failed
	$fail_msg = "ERR/JSON request failed"
	if ($json_response_obj -match 'ERR/') {
		if ($FromConsoleFlg) {
			Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj = $json_response_obj"
		}
		return $fail_msg
	}
	$fail_msg = "ERR/JSON request failed ('result' = fail)"
	if ($json_response_obj.result -eq 'fail') {
		$x  = $json_response_obj.result
		if ($FromConsoleFlg) {
			Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj.result = $x"
		}
		return $fail_msg
	}

	# Extract data from the JSON command response
	if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {
	
		$data_obj = $json_response_obj.data

		# For error testing
		#$data_obj = $null

		# Note: null/blank $data_obj.ram observed on occasion
		$data_cnt = $data_obj.Count
		
		# For debugging
		#Write-Host ("data_obj cnt = " + $data_cnt)
		#Write-Host ("data_obj[0] = `n" + $data_obj[0])
		#Write-Host ("data_obj[" + ($data_cnt - 1) +"] = `n" + $data_obj[$data_cnt-1])

		$rtn_obj = [ordered]@{}

		for ($i=0; $i -lt $data_obj.Length; $i++) { 
			$item_obj = $data_obj[$i]
			$option_disp = $item_obj.optionDisplay;
			$option_val  = $item_obj.optionValue;
			
			# For debugging
			#Write-Host "$i : option_val = $option_val, option_disp = $option_disp"
			#Write-Host ("    option_val starts w/ @ = " + $option_val.StartsWith("@"))
			#Write-Host ("    option_disp starts w/ + = " + $option_disp.StartsWith("+"))
			
			if ($option_val -ne "Index" -and !$option_disp.StartsWith("+") -and !$option_val.StartsWith("@")) {
				$new_obj = [ordered]@{}
				
				if ($cam_props_ary.Length -gt 0) {
					# GET REQUESTED PROPERTIES
					
					foreach ($cam_prop in $cam_props_ary) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
							#$new_obj.$cam_prop = $x
						} else {
							if ($item_obj.$cam_prop -eq 0) {
								$new_obj.Add($cam_prop, 0)
							} else {
								$new_obj.Add($cam_prop, $null)
							}
						}
					}
				} else {
					# GET ALL PROPERTIES
					
					# Get all property names in the JSON response (a PowerShell PSCustomObject)
					# ... NOTE: Extracting 'keys' from the PSCustomObject was bit tricky
					#           REF: https://stackoverflow.com/a/18780001/13499984
					$item_obj_keys = @()
					foreach ($item in $item_obj) {
						$item_obj_keys += $($item | Get-Member -MemberType *Property).Name
					}

					# For debugging
					#Write-Host ("item_obj_keys keys = " + $item_obj_keys)
					
					foreach ($cam_prop in $item_obj_keys) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
						} else {
							if ($item_obj.$cam_prop -eq 0) {
								$new_obj.Add($cam_prop, 0)
							} else {
								$new_obj.Add($cam_prop, $null)
							}
						}
					}
				}
				$rtn_obj[$option_val] = $new_obj
			}
		}

		# For testing
		#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
		#Write-Host "rtn_obj[DW1] = "$rtn_obj.DW1
		#Write-Host "rtn_obj[DW1]['BPS'] = "$rtn_obj['DW1']['BPS']
		#Write-Host "rtn_obj keys = "$rtn_obj.Keys
		
	} else {
		#Write-Host "sorry - JSON query returned null/blank`n"
	}

	return $rtn_obj

}

function MakeBitArray() {
# PURPOSE:    To factor a decimal number to a specified number of bits N (i.e., 2^0, 2^1 ... 2^N).
# PARAMS:     -DecValue = base 10 number
#             -NoBits   = size of return array (NOTE: index starts at zero)
# RETURNS:    (integer array) of 0's and 1's for each bit (1 if set, 0 if not).
#             Example: dec 403767296 to bin returns 29 bits 11000000100010000000000000000
#                      but we want 32 bits               00011000000100010000000000000000

	param (
		[Int64]$DecValue, 
		[Int]$NoBits
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() DecValue = $DecValue"
	#Write-Host "$fn_name() NoBits = $NoBits"

	$bin_str = [convert]::ToString($DecValue,2)

	# For debugging
	#Write-Host "$DecValue to bin = $bin_str"

	if ($bin_str.Length -le $NoBits) {
		$str = $bin_str.PadLeft($NoBits,'0')
	} else {
		$str = $bin_str.Substring(0,$NoBits)
	}
	$bin_ary = $str.ToCharArray()
	
	# Reverse the array
	[Array]::Reverse($bin_ary)
	
	# Convert string array to integer array
	$rtn_ary = @()
	foreach($item in $bin_ary) {
		if ($item -match "1") {
			$rtn_ary += 1
		} else {
			$rtn_ary += 0
		}
	}

	# For debugging
	#Write-Host ("str = " + $str)
	#Write-Host ("rtn_ary cnt = " + $rtn_ary.Count)
	#Write-Host ("rtn_ary = " + $rtn_ary)
	
	return $rtn_ary;
}

function BIJsonAlertListParse() {
# PURPOSE:   To get and parse the Blue Iris JSON interface cmd 'alertlist' response for a single camera
#            To retrieve all or, if requested, a subset of 'alertlist' properties for each Alert
#            NOTE: returns non-existent requested 'alertlist' properties as 'alert_prop':'NA'
# PARAMS:    -CamName        = [REQUIRED] camera shortname
#            -Pattern        = process Alerts with descriptor substring; must be used w/ -CamName
#            -AlertProps     = (array) like 'memo,offset', comma delimited
#              			       options: camera, clip, color, date, file, filesize, flags memo, 
#                                       newalerts, newalerttime, offset, path, res, zones
#            -MemoTag        = (string) process only those protected Alerts with descriptor substring
#            -AlertPath      = (string) process single Alert with alertlist [path] like '@123456789'
#            -DatabaseView   = (string) default is 'alerts'
#                              options: all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
#                                       flagged, export, archive, people, vehicles, confirmed, or canceled                             
#                              Note: When view='flagged', clip items are included too, but they don't have msec metadata.
#            -FromConsoleFlg = (boolean) $true notifies function that calling command was executed from the console
#            -QuietFlg       = (boolean) true suppresses verbose informational output to the console
#            -SilentFlg      = (boolean) true suppresses all output to the console, including warnings
#            -DevModeFlg     = (boolean) true performs devmode actions
#            -DevModePath    = (string) destination full path for devmode actions
# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'alert_prop':value}
# REQUIRES: Functions: BIJsonCall(), IsBiVersion(), MakeBitArray()
#
# Created by jaydeel at ipcamtalk.com
# Version: 4 (2024-02-18)
#
# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
#    "returns a list of cameras on the system ordered by group. Cameras not belonging to any
#    group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
#    the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
#    describing a camera or a camera group."

	param(
		[String]$CamName,           # [REQUIRED] camera shortname
		[String]$Pattern,           # must be used with -CamName
		[String[]]$AlertProps,      # list of BI JSON 'alertlist' cmd properties (parameters)
		[String]$MemoTag,           # substring in Alert's memo
		[String]$DatabaseView,      # default is 'alerts'
		[String]$AlertPath,         # like '@123456789' (JSON alertlist [path] value)
		[Boolean]$FromConsoleFlg,
		[Boolean]$QuietFlg,
		[Boolean]$SilentFlg,
		[Boolean]$DevModeFlg,
		[String]$DevModePath
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() CamName = $CamName"
	#Write-Host "$fn_name() Pattern = $Pattern"
	#Write-Host ("$fn_name() AlertProps type = " + $AlertProps.GetType())
	#Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
	#Write-Host "$fn_name() MemoTag = $MemoTag"
	#Write-Host "$fn_name() DatabaseView = $DatabaseView"
	#Write-Host "$fn_name() AlertPath = $AlertPath"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() DevModeFlg = $DevModeFlg"
	#Write-Host "$fn_name() DevModePath = $DevModePath"

	#Set-StrictMode  -Version 3
	# PARSE ARGS
	if ($Pattern -and !$CamName) {
		$err_msg = "ERR/argument -Pattern must be used with -CamName [$fn_name]"
		return $err_msg
	}
	if (!$CamName) {
		$err_msg = "ERR/expecting at least 1 argument (-CamName) [$fn_name]"
		return $err_msg
	}

	# If searching with $AlertMemo, we can avoid GETTING all alerts from the database
	# ... via the JSON 'alertlist' cmd parameter 'search'
	$search_param_val = $null  # Default
	if ($MemoTag) {
		$search_param_val = $MemoTag
	}
	#Write-Host "$fn_name() search_param_val = $search_param_val"
	
	if (!$DatabaseView) {
		$DatabaseView = 'all'
	}
	if (!$AlertPath) {
		$AlertPath = "*"
	}

	# Initialize master collector for all alerts in database 
	$al_props_obj = @{}
	$al_props_obj.Add( 'cameras', @() )
	$al_props_obj.Add( 'paths', @() )
	$al_props_obj.Add( 'clips', @() )
	$al_props_obj.Add( 'files', @() )
	$al_props_obj.Add( 'memos', @() )
	$al_props_obj.Add( 'offsets', @() )
	$al_props_obj.Add( 'flagss', @() )  # double 's' intentional!
	$al_props_obj.Add( 'ress', @() )    # double 's' intentional!
	$al_props_obj.Add( 'zoness', @() )  # double 's' intentional!
	$al_props_obj.Add( 'dates', @() )
	$al_props_obj.Add( 'colors', @() )
	$al_props_obj.Add( 'filesizes', @() )
	
	# For debugging
	#Write-Host ("$fn_name() al_props_obj type = " + $al_props_obj.GetType())
	#Write-Host ("$fn_name() al_props = " + ($al_props_obj | ConvertTo-Json))

	# ---------------------------------------------------------------------------------------
	# KEY POINT: Prior to version 5.7.9.3 BI's 'alertlist' json cmd returned only 1000 items; 
	#            As of version 5.7.9.3 BI's 'alertlist' json cmd returns 2048 items; 
	#            so we must re-send the json cmd until we have retrieved all alerts.
	#            REF: https://ipcamtalk.com/threads/blue-iris-ui3.23528/post-788744
	
	# For debugging
	#Write-Host ("BI version > 5.7.9.3 = " + (IsBiVersion '5.7.9.3'))

	if (IsBiVersion '5.7.9.3') {
		$json_alertlist_max = 2048;
		#$json_cliplist_max = 4096;
	} else {
		$json_alertlist_max = 1000;  # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
		#$json_cliplist_max = 1200;   # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
	}
	# ---------------------------------------------------------------------------------------

	# Initialize WHILE LOOP variables
	[Array]$paths_ary = @('x')	# primary break loop variable... starts with a dummy value
	$oldest_date = 0  # oldest date found 
	$newest_date = 0  # newest date found 
	$skipped_next_json_call_flg = $false
	
	# Initialize WHILE LOOP break variables
	$prev_loop_alerts_cnt = 0
	$max_tries = 10
	$iter = 1;

	# WHILE LOOP - Get JSON 'alertlist' data in batches until the cmd returns no new data
	while ($paths_ary.Count -gt 0) {
		
		if ($DevModeFlg) {
			if ($FromConsoleFlg) {
				if ($iter -eq 1) {
					Write-Host ""
					Write-Host "(dev mode) ----- $CamName" -ForegroundColor Cyan
				} else {
					Write-Host "(dev mode) -----" -ForegroundColor Cyan
				}
				Write-Host "(dev mode) JSON alertlist iter #$iter" -ForegroundColor Cyan
			}
		}

		# Initialize collector this JSON 'alertlist' call
		$tmp_obj = @{}
		$tmp_obj.Add( 'cameras', @() )
		$tmp_obj.Add( 'paths', @() )
		$tmp_obj.Add( 'clips', @() )
		$tmp_obj.Add( 'files', @() )
		$tmp_obj.Add( 'memos', @() )
		$tmp_obj.Add( 'offsets', @() )
		$tmp_obj.Add( 'flagss', @() )  # double 's' intentional!
		$tmp_obj.Add( 'ress', @() )    # double 's' intentional!
		$tmp_obj.Add( 'zoness', @() )  # double 's' intentional!
		$tmp_obj.Add( 'dates', @() )
		$tmp_obj.Add( 'colors', @() )
		$tmp_obj.Add( 'filesizes', @() )
		#$tmp_obj.Add( 'AAAA', @() )
		
		# For debugging
		#Write-Host ("$fn_name() tmp_obj type = " + $tmp_obj.GetType())
		#Write-Host ("$fn_name() tmp_obj = " + ($tmp_obj | ConvertTo-Json))
		#Write-Host ("$fn_name() tmp_obj.paths type  = " + ($tmp_obj.paths).GetType())

		# GET JSON DATA FROM THE BLUE IRIS WEB SERVER

		# First call params - Default
		$jsonCmd = @{
			cmd = 'alertlist'
			camera = $CamName
			search = $search_param_val
			view = $DatabaseView
		}
		
		# Subsequent call params
		if ($iter -gt 1) {
			
			# For debugging
			#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
			#Write-Host "$fn_name() oldest_date = $oldest_date"
			
			if ($oldest_date -gt 0) {  #Just in case
				# Params for the next set of 'alertlist' data
				$jsonCmd = @{
					cmd = 'alertlist'
					camera = $CamName
					search = $search_param_val
					view = $DatabaseView
					enddate = $oldest_date   # oldest time found (so far); per Ken @ Blue Iris Support 
				}
			}
		}
		
		# For debugging
		#Write-Host ("$fn_name() jsonCmd = " + ($jsonCmd | ConvertTo-Json))

		# FIRST JSON CALL
		# Returns a powershell PSCustomObject
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-QuietFlg $QuietFlg `
			-FromConsoleFlg $FromConsoleFlg

		if ($FromConsoleFlg) {
			if ($iter -gt 1) {
				#Write-Host ("$fn_name() jsonCmd response type = " + $json_response_obj.GetType())
				#Write-Host ("$fn_name() jsonCmd response = " + ($json_response_obj | ConvertTo-Json))
			}
		}

		if ($json_response_obj -match 'success') {

			# Test for $json_response_obj.data object
			if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

				$data_obj = $json_response_obj.data

				if ($data_obj) {
					
					# For debugging
					#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
					#Write-Host ("$fn_name() data_obj cnt = " + ($data_obj.Count))
					#Write-Host ("$fn_name() data_obj.path type = " + ($data_obj.path).GetType())
					#Write-Host ("$fn_name() data_obj.memo type = " + ($data_obj.memo).GetType())

					# KEY POINT: $paths_ary is the WHILE LOOP's primary break variable
					try {
						$paths_ary = [String[]]$data_obj.path  					
					}
					catch {
						$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'path' entries (database corrupt?)"
						return $err_msg
					}
					
					# Get the alertlist dates array
					# KEY POINT: using try:catch as a redundant quality test
					try {
						$dates_ary = [int[]]$data_obj.date  					
					}
					catch {
						$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'date' entries (database corrupt?)"
						return $err_msg
					}

					# PERFORM SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
					# WHY? Per Blue Iris support...
					#     Prior to version 5.8.2.4, the alertlist 'memo' field was included only if it was not blank.
					#     Starting with version 5.8.3.4 it is now always included (REF: 12/18/2023 email from Ken)
						
					# For debugging
					#Write-Host ("$fn_name() BI version > 5.8.2.4 = " + (IsBiVersion '5.8.2.4'))
					#Write-Host ("$fn_name() data_obj.memo = " + ($data_obj.memo | ConvertTo-Json))
					
					$memo_fix_cntr = 0

					if (IsBiVersion '5.8.2.4') {
						$memos_ary = [String[]]$data_obj.memo
						#Write-Host ("$fn_name() memos_ary** = " + ($memos_ary | ConvertTo-Json))
					} else {
						# For debugging
						#Write-Host "$fn_name() Special handling of Alert memos" 
						
						try {
							# KEY POINT: This code line fails if ANY Alert in the JSON 'alertlist' response lacks a 'memo'
							$memos_ary = [String[]]$data_obj.memo  
						} 
						catch {
							# REPAIR THE 'BAD' ALERTS
							
							# For debugging
							Write-Warning "$fn_name() Found alerts w/ empty memos; preparing to change values to 'NA'..." 
							
							$i = 0
							foreach ($item in $data_obj) {
								try {
									$x = $item.memo
								}
								Catch {
									$data_obj[$i] | Add-Member -MemberType NoteProperty -Name "memo" -Value "NA"
									$memo_fix_cntr++
								}
								$i++
							}
							$memos_ary = [String[]]$data_obj.memo
							
							if ($data_obj['memos'].Count -ne $tmp_obj['paths'].Count) {
								$err_msg = "ERR/unable to fix missing 'memo':'value' pairs in the JSON 'alertlist' response"
								return $err_msg
							}
							
							if ($FromConsoleFlg -and !$SilentFlg) {
								Write-Warning "repaired $memo_fix_cntr Alerts in the JSON 'alertlist' response that lacked a 'memo' key; now 'memo':'NA'"
							}
						}
					}

					# For debugging
					if ($memo_fix_cntr -gt 0) {
						Write-Host ("$fn_name() memos_ary cnt AFTER repair = " + $memos_ary.Count)
						Write-Host ("$fn_name() memos_ary AFTER repair = " + ($memos_ary | ConvertTo-Json))
					}

					# For debugging
					#Write-Host ("$fn_name() memos_ary = " + ($memos_ary | ConvertTo-Json))

					# Initialize vars
					$this_loop_alerts_cnt = $paths_ary.Count
					$this_loop_dates_cnt = $dates_ary.Count
					
					# Shows $dates_ary endpoints
					if ($DevModeFlg -and  $FromConsoleFlg) {
						Write-Host ("(dev mode) fist date = dates_ary[0] = " + $dates_ary[0]) -ForegroundColor Cyan 
						Write-Host ("(dev mode) last date = dates_ary[" + ($this_loop_dates_cnt-1) + "] = " + $dates_ary[$this_loop_dates_cnt-1]) -ForegroundColor Cyan
					}

					# Break loop if this iteration is a single result with same oldest date
					# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
					if ($this_loop_alerts_cnt -eq 1) {
						if ($dates_ary[$this_loop_dates_cnt-1] -eq $oldest_date) {
							
							if ($DevModeFlg) {
								if ($FromConsoleFlg) {
									Write-Host ("(dev mode) JSON alertlist this_loop_alerts_cnt = " + $this_loop_alerts_cnt) -ForegroundColor Cyan
									Write-Host ""
								}
							}
							break
						}
					}

					# Break loop if item's 'path' and 'date' parameter counts do not match
					# SOMETHING'S WRONG
					if ($this_loop_alerts_cnt -ne $this_loop_alerts_cnt) {
						if ($DevModeFlg -and $FromConsoleFlg) {
							Write-Host "(dev mode) terminating 'alertlist' call; response is unexpected ('path' & 'date' counts unequal)" -ForegroundColor Cyan
						}
						break
					}

					# Break loop if JSON 'alertlist' call returns the same number of alerts as the 
					# previous result (<2048 items)
					if ($this_loop_alerts_cnt -lt $json_alertlist_max -and ($this_loop_alerts_cnt -eq $prev_loop_alerts_cnt)) {
						if ($al_props_obj['paths'] -contains $paths_ary[0]) {
							if ($DevModeFlg -and $FromConsoleFlg) {
								Write-Host "(dev mode) ignoring JSON alertlist response w/ same alert count as previous response" -ForegroundColor Cyan
							}
							break
						}
					}
					
					$prev_loop_alerts_cnt = $this_loop_alerts_cnt
					
					if ($DevModeFlg -and $FromConsoleFlg) {
						Write-Host "(dev mode) JSON alertlist this_loop_alerts_cnt = $this_loop_alerts_cnt" -ForegroundColor Cyan
					}

					# For iterations > 1, remove the first alert in the response
					# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
					if ($iter -gt 1) {
						
						if ($this_loop_alerts_cnt -ge 1 -and $this_loop_alerts_cnt -le $json_alertlist_max) {
							
							# REMOVE THE FIRST OBJECT ELEMENT 
							$data_obj = $data_obj[1 .. ($data_obj.Count - 1)]

							# SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
							$memos_ary = $memos_ary[1 .. ($memos_ary.Count - 1)]
							
							# KEY POINT: 
							# Note that we do not update var $paths_ary at this point
							# Why? Because it is the while loop's break variable and we still have work to do
							
							# For debugging
							#Write-Host ("$fn_name() data_obj cnt AFTER = " + ($data_obj.Count))
						}
					}

					# For debugging
					#if ($iter -eq 2) {
						#$data_obj_keys += $($data_obj[0] | Get-Member -MemberType *Property).Name
						#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
						#Write-Host ("$fn_name() data_obj keys = " + ($data_obj_keys | ConvertTo-Json))
						#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
						#Write-Host ("$fn_name() data_obj.date type = " + ($data_obj.date).GetType())
					#}

					# Get all dates for this set of Alerts
					# ... This SHOULD BE an array of UTC timestamps, ordered largest to smallest
					# ... NOTE: this array is also used for QUALITY testing
					[Int[]]$dates_ary = $data_obj.date
					$dates_cnt = $dates_ary.Count

					# HANDLE SIMULATIONS
					if ($script:Sim5) {  #simulate response w/ zero-date (timestamp) values; arg value is CamName
						if ($Cameras -contains $script:Sim5) {
							# Set first date to zero
							#$dates_ary[0] = 0
							#$data_obj[0].date = 0

							# Set last date to zero
							#$dates_ary[$dates_cnt-1] = 0
							#$data_obj[$dates_cnt-1].date = 0
						
							# Set 2nd date to zero
							$dates_ary[1] = 0
							$data_obj[1].date = 0

							# Set 2nd to last date to zero
							$dates_ary[$dates_cnt-2] = 0						
							$data_obj[$dates_cnt-2].date = 0
						
							# For debugging
							Write-Host ("SIM5 fist date = dates_ary[0] = " + $dates_ary[0])
							Write-Host ("SIM5 last date = dates_ary[" + ($dates_cnt-1) + "] = " + $dates_ary[$dates_cnt-1])
						}
					}
					
					# Get the oldest/newest dates for this set of Alerts 
					# ... The oldest date is needed for the next JSON 'alertlist' call, if necessary
					$newest_date = $dates_ary[0]
					$oldest_date = $dates_ary[$dates_cnt-1]
					$max_date = ($dates_ary | Measure-Object -Maximum).Maximum
					$min_date = ($dates_ary | Measure-Object -Minimum).Minimum
					$dates_sum = ($dates_ary | Measure-Object -Sum).Sum
					
					# For debugging
					#Write-Host ("$fn_name() dates_ary type = " + $dates_ary.GetType())
					#Write-Host "$fn_name() dates_cnt = $dates_cnt"
					#Write-Host "$fn_name() dates_ary = $dates_ary"
					#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
					#Write-Host "$fn_name() newest_date = $newest_date"
					#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
					#Write-Host "$fn_name() oldest_date = $oldest_date"
					#Write-Host "$fn_name()dates_sum = $dates_sum"
					#Write-Host ("$fn_name() data_obj.date = " + ($data_obj.date | ConvertTo-Json))

					# Check for more QUALITY issues...
					# ... 1. Check if all dates are zero (FATAL)
					if ($dates_sum -eq 0 ) {
						$skipped_next_json_call_flg = $true
						if ($FromConsoleFlg -and !$SilentFlg) {
							Write-Warning "JSON 'alertlist' response for '$CamName' contains only zero 'date' entries (database corrupt?); find w/ -DevMode"
						}
						break
					}
					# ... 2. Check if the first array element is the max date (IT SHOULD BE)
					$first_date_is_not_max = ($newest_date -lt $max_date)
					# ... 3. Check that the first and last array elements are NOT zero (THEY SHOULD BE)
					$first_date_is_zero = ($newest_date -eq 0)
					$last_date_is_zero = ($oldest_date -eq 0)
					$zero_dates_cntr = 0  # Initialize counter
					# ... 4. Check if any OTHER alerts have dates with zero values
					#        Make hashtable like @("date1":count, "date2":count ...}
					#        if zero-dates are found, $ht will have and entry like {..., "0": 23, ...}
					# ... REF: https://stackoverflow.com/a/12640991
					$dates_ary | Group | % { $ht = @{} } { $ht[$_.Name] = $_.Count } { $ht } | Out-Null
					
					# For debugging
					#Write-Host ("ht = " + ($ht | ConvertTo-Json))
					#Write-Host ("ht['0'] = " + $ht['0'])
					
					if ($ht['0']) {
						$zero_dates_cntr += $ht['0']
						$dates_between_zero_cntr = $ht['0']
						
						# For debugging
						#Write-Host ("ht['0'] type = " + ($ht['0']).GetType())
						
					} else {
						$dates_between_zero_cntr = 0
					}
					if ($first_date_is_zero -and $dates_between_zero_cntr -gt 0) {
						$dates_between_zero_cntr--
					}
					if ($last_date_is_zero -and $dates_between_zero_cntr -gt 0) {
						$dates_between_zero_cntr--
					}
					if ($FromConsoleFlg -and !$SilentFlg) {
						if ($first_date_is_zero) {
							if ($last_date_is_zero) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins and ends with an Alert with zero-date values"
							} else {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins with an Alert with a zero-date value"
							}
						} else {
							if ($first_date_is_not_max) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' does not have the newest date in the first entry (as expected)"
							}
							if ($last_date_is_zero) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' ends with an Alert with a zero-date value"
							}
						}
						if ($dates_between_zero_cntr -gt 0) {
							Write-Warning "JSON 'alertlist' response #$iter for '$CamName' has $zero_dates_cntr zero-date values; find w/ -DevMode"
						}
					}
					
					# For debugging
					#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
					#Write-Host "$fn_name() newest_date = $newest_date"
					#Write-Host ("$fn_name() newest_date for iter #" + $iter + " ($this_loop_alerts_cnt alerts) = " + $newest_date)
					#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
					#Write-Host "$fn_name() oldest_date = $oldest_date"

					# Load collectors
					$tmp_obj['cameras']   = $data_obj.camera  # an array (adds a column of values to the hashtable object)
					$tmp_obj['paths']     = $data_obj.path
					$tmp_obj['clips']     = $data_obj.clip
					$tmp_obj['files']     = $data_obj.file
					#$tmp_obj['memos']     = $data_obj.memo   # We handle '$data_obj.memo' differently 
					$tmp_obj['memos']     = $memos_ary
					$tmp_obj['offsets']   = $data_obj.offset
					$tmp_obj['flagss']    = $data_obj.flags
					$tmp_obj['ress']      = $data_obj.res
					$tmp_obj['zoness']    = $data_obj.zones
					$tmp_obj['dates']     = $data_obj.date	# Note: $date_ary[] is reverse-ordered.
					$tmp_obj['colors']    = $data_obj.color
					$tmp_obj['filesizes'] = $data_obj.filesize

					# For debugging
					#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))
	
					# HANDLE SIMULATIONS
					if ($script:Sim1) {  # remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
						foreach ($item in $data_obj) {
							$item.PSObject.Properties.Remove('memo')
						}
						if ($FromConsoleFlg) {
							Write-Host "(test mode) Simulation #1: removed ALL 'memo':'value' pairs from JSON 'alertlist' response" -Foregroundcolor Green
							#Write-Host ("(test mode) Simulation #1: data_obj AFTER = " + ($data_obj | ConvertTo-Json)) -Foregroundcolor Green
						}
					}
					if ($script:Sim2) {  # remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response
						try {
							$tmp_obj['memos'] = [Array]$data_obj.memo
							$total_cnt = $tmp_obj['paths'].Count
							$chgd_memos_cnt = [Math]::min(3,$alerts_cnt-1)
							$chgd_idxs_ary = @()
							$chgd_alerts_obj = @()
							$i = 0
							while ($i -lt $chgd_memos_cnt) {
								$idx = Get-Random -Maximum $total_cnt
								# Only protected alerts
								#Write-Host ("flags = " + ($data_obj[$idx]).flags)
								$flags_ary = `
									MakeBitArray `
									-DecVal ($data_obj[$idx]).flags `
									-NoBits 32
								# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
								if ($flags_ary[2] -eq 1) {
									if (!($chgd_idxs_ary -contains $idx)) {
										$data_obj[$idx].PSObject.Properties.Remove('memo')
										$chgd_alerts_obj += $data_obj[$idx]
										$chgd_idxs_ary += $idx
										$i++
									}
								}
							}
							if ($FromConsoleFlg) {
								Write-Host "(test mode) Simulation #2: removed JSON alertlist 'memo' entry from $chgd_memos_cnt random protected Alerts" -Foregroundcolor Green
								Write-Host ("(test mode) Simulation #2: changed alertlist entries = " + ($chgd_alerts_obj | ConvertTo-Json)) -Foregroundcolor Green
							}
						} catch {
							if ($FromConsoleFlg) {
								Write-Host "(test mode) Simulation #2: unable to run simulation; no Alerts have a 'memo':'value' pair"
							}
						}
					}
					if ($script:Sim3) {  # simulate response w/ all 'memo'='NA' params
						$tmp_obj['memos'] = 1..$tmp_obj['paths'].Count | % { 'NA' }
						if ($FromConsoleFlg) {
							Write-Host "(test mode) Simulation #3: set 'memo'='NA' for all Alerts in $fn_name() response" -Foregroundcolor Green
							Write-Host ("(test mode) Simulation #3: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
						}
					}
					if ($script:Sim4) {  # simulate response w/ 3 random 'memo'='NA' params
						$memos_cnt = $tmp_obj['memos'].Count
						if ($memos_cnt -gt 0) {
							#Write-Host "memos_cnt = $memos_cnt"
							$memos_cnt = $memos_cnt
							$chgd_memos_cnt = [Math]::min(3,$memos_cnt-1)
							$chgd_idxs_ary = @()
							$chgd_alerts_ary = @()
							$i = 0
							while ($i -lt $chgd_memos_cnt) {
								$idx = Get-Random -Maximum $memos_cnt
								#Write-Host "idx = $idx"
								# Only protected alerts
								#Write-Host ("flags = " + ($data_obj[$idx]).flags)
								$flags_ary = `
									MakeBitArray `
									-DecVal ($data_obj[$idx]).flags `
									-NoBits 32

								# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
								if ($flags_ary[2] -eq 1) {
									if (!($chgd_idxs_ary -contains $idx)) {
										if ($tmp_obj['memos'][$idx] -ne 'NA') {
											$tmp_obj['memos'][$idx] = 'NA'
											$chgd_alerts_ary += (($data_obj[$idx]).file).Replace('.jpg','')
											$chgd_idxs_ary += $idx
											$i++
										}
									}
								}
							}
							if ($FromConsoleFlg) {
								Write-Host "(test mode) Simulation #4: set 'memo'='NA' for $chgd_memos_cnt random Alerts in $fn_name() response" -Foregroundcolor Green
								#Write-Host ("(test mode) Simulation #4: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
								Write-Host ("(test mode) Simulation #4: modified AlertIDs = " + ($chgd_alerts_ary | ConvertTo-Json)) -Foregroundcolor Green
							}
						}
					}

					# For debugging
					
					#Write-Host ("$fn_name() tmp_obj['paths'] cnt = " + $tmp_obj['paths'].Count)
					#Write-Host ("$fn_name() tmp_obj['paths'] = " + ($tmp_obj['paths'] | ConvertTo-Json))
					#Write-Host ("$fn_name() tmp_obj['flagss'] cnt = " + $tmp_obj['flagss'].Count)
					#Write-Host ("$fn_name() tmp_obj['flagss'] = " + ($tmp_obj['flagss'] | ConvertTo-Json))
					#Write-Host ("$fn_name() tmp_obj['memos'] dnt = " + $tmp_obj['memos'].Count)
					#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))

					# Merge results with the master accumulator
					$al_props_obj['cameras'] = $al_props_obj['cameras'] + $tmp_obj['cameras']
					$al_props_obj['paths']   = $al_props_obj['paths']   + $tmp_obj['paths']
					$al_props_obj['clips']   = $al_props_obj['clips']   + $tmp_obj['clips']
					$al_props_obj['files']   = $al_props_obj['files']   + $tmp_obj['files']
					$al_props_obj['memos']   = $al_props_obj['memos']   + $tmp_obj['memos']
					$al_props_obj['offsets'] = $al_props_obj['offsets'] + $tmp_obj['offsets']	
					$al_props_obj['flagss']  = $al_props_obj['flagss']  + $tmp_obj['flagss']	
					$al_props_obj['ress']    = $al_props_obj['ress']    + $tmp_obj['ress']	
					$al_props_obj['zoness']  = $al_props_obj['zoness']  + $tmp_obj['zoness']	
					$al_props_obj['dates']   = $al_props_obj['dates']   + $tmp_obj['dates']	
					$al_props_obj['colors']  = $al_props_obj['colors']  + $tmp_obj['colors']	
					$al_props_obj['filesizes']   = $al_props_obj['filesizes']   + $tmp_obj['filesizes']	
					
					# Stop requesting data if the last date is zero
					if ($oldest_date -eq 0) {
						$skipped_next_json_call_flg = $true
						if ($FromConsoleFlg -and !$SilentFlg) {
							Write-Warning "skipping next JSON 'alertlist' call because this call's oldest date is zero"
						}
						break
					}
				
				} else {
					break
				}
			} else {
				break
			}
		} else {
			break
		}

		# Avoid infinite loop
		if ($iter -gt $max_tries) {
			break
		}
		
		$iter++
	}
	
	# For debugging
	#Write-Host "Broke out of While loop"

	# TEST FOR MORE QUALITY ISSUES (indicators of a possibly corrupt database)
	# ... Test if 'clip' values contain '@-1.jpg' entries (WARNING)
	[String[]]$clips_ary = $al_props_obj['clips']
	if ($clips_ary -contains '@-1.jpg') {
		$clips_cnt = $clips_ary.Count
		$bad_clip_cntr = 0
		foreach($item in $clips_ary) {
			if ($item -eq '@-1.jpg') {
				$bad_clip_cntr++
			}
		}
		if ($FromConsoleFlg -and !$SilentFlg) {
			Write-Warning "JSON 'alertlist' response for '$CamName' has $bad_clip_cntr/$clips_cnt 'clip'='@-1.jpg' entries; find w/ -DevMode"
		}
	}

	# For debugging
	#Write-Host ("$fn_name() al_props_obj type = " + ($al_props_obj.GetType()))
	#Write-Host ("$fn_name() al_props_obj keys = " + $al_props_obj.Keys)
	#Write-Host ("$fn_name() al_props_obj = " + ($al_props_obj | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['cameras'] = " + ($al_props_obj['cameras'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['memos'] = " + ($al_props_obj['memos'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj['flagss'] = " + ($al_props_obj['flagss'] | ConvertTo-Json))
	#Write-Host ("$fn_name() al_props_obj.paths = " + $al_props_obj.paths)
	
	# DEV MODE - Save each camera's 'alertlist' data
	if ($DevModeFlg) {
		if (!$script:Sim1 -and !$script:Sim2 -and !$script:Sim3 -and !$script:Sim4) {
			
			$al_objects_ary = @()  # Initialize accumulator; an array of hashtables
			$keys = [Array]$al_props_obj.Keys
			$all_alerts_cnt = $al_props_obj['paths'].Count
			$alertlist_filespec = "$DevModePath\alertlist.$CamName.json"
			
			# For debugging
			#Write-Host ("$fn_name() keys = " + ($keys | ConvertTo-Json))
			#Write-Host "$fn_name() all_alerts_cnt = $all_alerts_cnt"
			#Write-Host "$fn_name() alertlist_filespec = '$alertlist_filespec'"
			
			if ($all_alerts_cnt -gt 0) {

				# Possible keys in $al_props_obj
				$keys1 = @('cameras','memos','paths','clips','files','filesizes','dates','offsets')
				$keys2 = @('flagss','zoness')

				for ($i=0; $i -lt $all_alerts_cnt; $i++) {
					$ht = @{}  # Initialize hashtable
					# Modify $al_props_obj key for use in hashtable 
					
					#Write-Host ""
					
					foreach ($key in $keys) {
						if ($keys1 -contains $key) {
							$key_new = $key.TrimEnd('s')
						} else {  # $keys_2 -contains $key
							$key_new = $key.Replace('ss','s')
						}
						
						# For debugging
						#Write-Host "key = $key"
						#Write-Host ("al_props_obj[$key] type = " + ($al_props_obj[$key]).GetType())
						#Write-Host ("al_props_obj[$key] = " + ($al_props_obj[$key] | ConvertTo-Json))
						#Write-Host ("al_props_obj[$key][0] = " + $al_props_obj[$key][0])
						#Write-Host ("al_props_obj[$key][" + ($all_alerts_cnt-1) + "] = " + $al_props_obj[$key][$all_alerts_cnt-1])
						#Write-Host ("al_props_obj[$key][$i] = " + $al_props_obj[$key][$i])
						
						$ht.Add($key_new, $al_props_obj[$key][$i])	
						#exit

						
					}
					$al_objects_ary += $ht
				}

				# For debugging
				#Write-Host ("al_list_ary = " + ($al_list_ary | ConvertTo-Json))

				$alertlist_saved_flg = $false
				if (!$Cameras -and !$Pattern) {
					$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"
					$alertlist_saved_flg = $true
				} else {
					if ($Cameras) {
						$match_str = ($Cameras -Join '|')
						if ($match_str -match $CamName) {
							$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
							$alertlist_saved_flg = $true
						}
					}				
					if ($Pattern) {
						if ($Pattern -match $CamName) {
							$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
							$alertlist_saved_flg = $true
						}
					}				
				}
				if ($alertlist_saved_flg) {		
					if ($FromConsoleFlg) {
						Write-Host "(dev mode) 'alertlist' response saved to disk ($all_alerts_cnt items) --> '$alertlist_filespec'" -ForegroundColor Cyan
						#Write-Host ""
					}
				}
				
				# Top open this *.json file in Excel...
				# 1. Data > Get Data > From File > From JSON
				# 2. Navigate to *json file
				# 3. Click 'Import' 
				# 4. Click 'To Table'
				# 5. Click 'OK'
				# 6. Click icon by 'Column1'
				# 7. Click 'OK'
				# 8. Click 'Close & Load'
				
			}  else {
				if ($FromConsoleFlg -and !$SilentFlg) {
					Write-Warning "no alerts; unable to save alertlist to disk ($alertlist_filespec)"
				}
			}
		} else {
			if ($FromConsoleFlg) {
				Write-Host "(dev mode) HEADS-UP: cam 'alertlist' *NOT* saved to disk when using arguments -SimX" -ForegroundColor Cyan
			}
		}
	}

	# Get all alerts files, paths
	$all_alerts_paths = $al_props_obj.paths
	$all_alerts_files = $al_props_obj.files
	$all_alerts_cnt = $all_alerts_paths.Count
	
	$all_alerts_zones = $al_props_obj.zoness

	# For debugging
	#Write-Host ("$fn_name() AlertPath = " + $AlertPath)
	#Write-Host ("$fn_name() all_alerts_paths cnt = " + $all_alerts_paths.Count)
	#Write-Host ("$fn_name() all_alerts_paths = " + ($all_alerts_paths | ConvertTo-Json))
	#Write-Host ("$fn_name() all_alerts_files = " + ($all_alerts_files | ConvertTo-Json))
	#Write-Host ("$fn_name() all_alerts_zones type = " + $all_alerts_zones.GetType())
	#Write-Host ("$fn_name() all_alerts_zones cnt = " + $all_alerts_zones.Count)
	#Write-Host ("$fn_name() all_alerts_zones = " + ($all_alerts_zones | ConvertTo-Json))
	#Write-Host "$fn_name() Found $all_alerts_cnt total items`n"

	# Initialize collectors
	$rtn_obj = [ordered]@{}
	$new_obj = [ordered]@{}
	if ($AlertProps.Count -gt 0) {
		foreach ($alert_prop in $AlertProps) {
			$new_obj.Add($alert_prop, @())
		}
	} else {
		foreach ($al_props_key in $al_props_obj_keys) {
			$new_obj.Add($al_props_key, @())
		}
	}
	
	# For debugging
	#Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

	# WALK THROUGH ALERTS LIST
	for ($i=0; $i -lt $all_alerts_cnt; $i++) { 
	
		# For debugging
		#Write-Host "`nalert idx = $i..."
			
		if ($AlertPath -ne "*") {
			if (!($all_alerts_paths[$i] -match $AlertPath)) {
				Continue
			} else {
				# For debugging
				#Write-Host "Found match for AlertPath ($AlertPath)"
			}
		}
		
		if ($Pattern.Length -gt 0) {
			if (!($all_alerts_files[$i] -match $Pattern)) {
				Continue
			} else {
				# For debugging
				#Write-Host ("Pattern '$Pattern' is a match for '" + $all_alerts_files[$i] + "'")
			}
		}

		if ($AlertProps.Count -gt 0) {
			
			# GET REQUESTED PROPERTIES

			foreach ($alert_prop in $AlertProps) {

				$key = $alert_prop + "s"  # $al_props_obj's keys are plural
				
				[Array]$arr = $al_props_obj.$key
				
				# For debugging
				#if ($key -eq 'memos') {
					#Write-Host ("$fn_name() arr[$key] = " + ($arr | ConvertTo-Json))
				#}
				
				if (!($null -eq $arr)) {  # KEY POINT: $arr can be a single element array @(0)
					
					$val = $arr[$i]
					
					# For debugging
					#if ($arr[$i] -eq '') {
						#Write-Host ($key + " = '" + $arr[$i] + "'")
					#}

					# Append value to collector
					if ($arr[$i] -or $arr[$i] -eq 0) {
						$new_obj.$alert_prop += $arr[$i]				
					} else {
						$new_obj.$alert_prop += ''
					}
					
					# For debugging
					#if ($arr[$i] -eq '') {
						#Write-Host ("new_obj.$alert_prop = " + ($new_obj.$alert_prop | ConvertTo-Json))
					#}

					# For debugging
					#Write-Host ("arr = " + $arr)
					#Write-Host ("item #" + ($i+1) + " $alert_prop = " + $arr[$i])

				} else {
					if ($i -eq 0) {
						if ($FromConsoleFlg) {
							Write-Host ("WARNING/nonexistent AlertProp '$alert_prop', expecting: [" + ($AlertProps -Join ',') + "]")
						}
					}
				}
			}
			
		} else {
			
			# GET ALL PROPERTIES
			
			# Get all property names
			if ($i -lt 2) {
				
				# NOTE: $al_props_obj's keys are plural
				$al_props_obj_keys = $al_props_obj.Keys

				# For debugging
				#Write-Host ("al_props_obj_keys = " + $al_props_obj_keys)
				
			}
			
			foreach ($al_props_key in $al_props_obj_keys) {

				# For debugging
				#Write-host ("new_obj.$al_props_key type = " + ($new_obj.$al_props_key).GetType())

				# Get all values for the property (an array)
				$vals = $al_props_obj.$al_props_key

				# Remove the terminal 's' from $al_props_key
				$new_obj_key = $al_props_key.Substring(0,$al_props_key.Length-1) 
				
				if ($vals) {
					$new_obj.$new_obj_key += $vals[$i]					
				} else {
					#$new_obj.$new_obj_key += $null
					$new_obj.$new_obj_key += ''
				}
				
				#For debugging
				#Write-Host ("item #" + ($i+1) + " $new_obj_key = " + $vals[$i])
			}
		}
			
		# For debugging
		#Write-host ("new_obj = " + ($new_obj | ConvertTo-Json))
	}
	
	$rtn_obj[$CamName] = $new_obj

	# For debugging
	#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
	#Write-Host "rtn_obj keys = "$rtn_obj.Keys
	#Write-Host ("rtn_obj keys = " + ($rtn_obj | ConvertTo-Json))
	#Write-Host ("rtn_obj = " + ($rtn_obj | ConvertTo-Json))s
	#Write-Host ("rtn_obj.FE = " + ($rtn_obj.FE | ConvertTo-Json))
	#Write-Host ("rtn_obj.DW1.BPS = '" + $rtn_obj.DW1.BPS + "'")
	#Write-Host ("rtn_obj.GT.path = '" + $rtn_obj.GT.path + "'")
	#Write-Host ("rtn_obj.GT.memo = '" + $rtn_obj.GT.memo + "'")
	#Write-Host ("rtn_obj.GT.flags = '" + $rtn_obj.GT.flags + "'")

	return $rtn_obj
}

function BIJsonGetAlertList() {
# PURPOSE:    To get properties from the Blue Iris JSON interface cmd 'alert'
# USAGE:      {this_script_name}.ps1 prop cam(s) prop(s)
# PARAMS:     -Cameras        = (string) filter results for Alerts with specified camera shortnames(comma-delimited),  
#                                        like 'DW1','DW2,'FD'
#             -Patterns       = (string) filter results for Alerts with specified descriptor substrings
#             -MemoTag        = (string) filter results for Alerts with specified memo substrings
#             -AlertProps     = (string) return only these JSON 'alertlist' cmd params (comma-delimited )
#                                        like 'enabled,profile,motion'
#             -FromConsoleFlg = (boolean) $true notifies function that calling command was executed from the console
#             -QuietFlg       = (boolean) true suppresses verbose informational output to the console
#             -DevModeFlg     = (boolean) true performs devmode actions
#             -DevModePath    = (string) destination path for devmode actions
# RETURNS:    (hashtable) @{ Cam1: @{ prop1:[values], ... , propN:[values] }, Cam2: @{ prop1:[values], ... , propN:[values] }, ... }
# REQUIRES:   Functions: 'BIJsonAlertListParse()'
#
# Created by jaydeel at ipcamtalk.com
# Version: 4 (2024-02-18)
#

	param(
		[String[]]$Cameras,
		[String[]]$Patterns,
		[String[]]$AlertProps,
		[String]$MemoTag,
		[Boolean]$FromConsoleFlg,
		[Boolean]$QuietFlg,
		[Boolean]$DevModeFlg,
		[String]$DevModePath
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "Entered function $fn_name() ..."
	#Write-Host "$fn_name() RecordID = $RecordID"
	#Write-Host "$fn_name() Cameras = $Cameras"
	#Write-Host "$fn_name() Patterns = $Patterns"
	#Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
	#Write-Host "$fn_name() MemoTag = $MemoTag"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	#Write-Host "$fn_name() DevModeFlg = $DevModeFlg"
	#Write-Host "$fn_name() DevModePath = $DevModePath"
	#Write-Host "$fn_name() Sim1 = $script:Sim1"
	#Write-Host "$fn_name() Sim3 = $script:Sim3"

	# HANDLE ARGS
	if ($Cameras) {
		$my_cams_ary = $Cameras -Split ','
	} else {
		$my_cams_ary = @()
	}
	$my_cams_cnt = $my_cams_ary.Count

	if ($Patterns) {
		$my_patterns_ary = $Patterns -Split ','
	} else {
		$my_patterns_ary = @('')
	}
	$my_patterns_cnt = $my_patterns_ary.Count

	if ($AlertProps.Count -eq 0) {
		$my_props_ary = @('camera', 'path', 'clip', 'file', 'memo', 'offset', 'flags', 'res', 'zones', 'date', 'color', 'filesize')
	} else {
		$my_props_ary = $AlertProps -Split ','
	}
	$my_props_cnt = $my_props_ary.Count

	# For debugging
	#Write-Host ("$fn_name() my_cams_ary type = " + $my_cams_ary.GetType())
	#Write-Host ("$fn_name() my_cams_ary = " + $my_cams_ary)
	#Write-Host ("$fn_name() my_cams_ary cnt = " + $my_cams_ary.Count)
	#Write-Host "my_patterns_ary = $my_patterns_ary"
	#Write-Host "my_flags_ary = $my_flags_ary"
	#Write-Host ("$fn_name() my_props_ary cnt = " + $my_props_ary.Count)
	#Write-Host ("$fn_name() my_props_ary = " + ($my_props_ary | ConvertTo-Json))

	# INITIALIZE
	$rtn_obj = @{}
	$progress_bar_cntr = 0
	#$cntr = 1
	$patterns_found_ary = @()
	$i = 0
	foreach ($my_cam in $my_cams_ary) {

		# For debugging
		#Write-Host ("$fn_name() " + $my_cam + ":")
		
		# Show progress bar
		if ($FromConsoleFlg) {
			$progress_bar_cntr++
			$WriteProgressParams = @{
				Activity         = "Parsing alertlist data..."
				Status           = "Progress:  cam '$my_cam' ($progress_bar_cntr of requested $my_cams_cnt cams)"
				PercentComplete  = (($progress_bar_cntr / $my_cams_cnt) * 100)
			}
			# Note: If loop is too fast, you might need a delay to get the progress bar to update
			#Start-Sleep -Milliseconds 150				
			Write-Progress @WriteProgressParams
		}
		
		$pattern_cntr = 0
		foreach ($pattern in $my_patterns_ary) {
			
			# For debugging
			#Write-Host "`n----- processing pattern '$pattern' -----"
			
			# Skip pattern search if already found
			if ($patterns_found_ary -contains $pattern) {
				continue
			}

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
			# ... returns an ordered hashtable
			$response_obj = `
				BIJsonAlertListParse `
				-CamName $my_cam `
				-Pattern $pattern `
				-AlertProps $my_props_ary `
				-MemoTag $MemoTag `
				-FromConsoleFlg $FromConsoleFlg `
				-QuietFlg $QuietFlg `
				-SilentFlg ($pattern_cntr -gt 0) `
				-DevModeFlg $DevModeFlg `
				-DevModePath $DevModePath

			# For debugging
			#Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
			#Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))
			
			if ($response_obj.GetType() -match 'string') {
				if ($response_obj.StartsWith('ERR/')) {
					return $response_obj
				} else {
					$err_msg = "ERR/unexpected response from AlertListParse(); $response_obj"
					return $err_msg
				}
			}

			# $response_obj is an ordered hashtable, like
			# @{
			# 	DW1: @{
			#		"file": [
			#           filename1,
			#           ...,
			#           filenameN,
			#       ],
			#       "flags": [
			#           int1,
			#           ...,
			#           int2,
			#       ],
			#       "memo": [
			#           string1,
			#           ...,
			#           stringN,
			#       ],
			#	}
			# }

			# INITIALIZE
			$x = ''

			# EXTRACT PROPERTY VALUE(S)
			# ... Note: null/blank $response_obj.ram observed on occasion
			if ($response_obj) {	

				$response_obj_keys = $response_obj.Keys

				# For debugging
				#Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
				#Write-Host ("$fn_name() response_obj cnt = " + $response_obj.Count)
				#Write-Host ("$fn_name() response_obj keys = " + $response_obj_keys)
				#Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))

				if ($response_obj_keys -contains $my_cam) {
					
					$cam = $my_cam
					$item_obj = $response_obj[$cam]
					$item_obj_keys = $item_obj.Keys

					# $item_obj (and $new_obj below) are System.Collections.Specialized.OrderedDictionary 
					# that should look like...
					<# 
					  "camera": ["DW1"],
					  "path": ["@224803430.bvr"],
					  "clip": ["@224669140.bvr"],
					  "file": ["DW1.20240128_020000.3152644.5-0.jpg"],
					  "memo": ["LD/C;vehicle:93%"],
					  "offset": ["3157680"],
					  "flags": [402980864],
					  "res": ["2688x1520"],
					  "zones": [65],
					  "date": [1705777428],
					  "color": [16711808],
					  "filesize": ["22 sec (523K)"]
					}
					#>					

					# For debugging
					#Write-Host ("$fn_name() item_obj type = " + $item_obj.GetType())
					#Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
					#Write-Host ("$fn_name() item_obj_keys = " + ($item_obj_keys | ConvertTo-Json))
					#Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
					#rite-Host ("$fn_name() item_obj[0] type = " +  ($item_obj[$item_obj_keys[0]]).GetType())
					#Write-Host ("$fn_name() item_obj[0] = " +  $item_obj[$item_obj_keys[0]])
					
					# Test for empty result
					if (!$item_obj[$item_obj_keys[0]]) {  # looking for $item_obj['camera'] = @()
						continue
					} else {
						if ($patterns_ary.Count -gt 1) {
							$patterns_found_ary += $pattern
						}
					}

					$new_obj = [ordered]@{}	
					
					if ($my_props_ary.Length -gt 0) {
						$props_ary = $my_props_ary   # RETURN REQUESTED PROPERTIES
					} else {
						$props_ary = $item_obj_keys  # RETURN ALL PROPERTIES			
					}

					# RETURN PROPERTIES

					foreach ($cam_prop in $props_ary) {
						
						$item_obj_ary = $item_obj.$cam_prop  # expect $item_obj.path = @(value)

						# For debugging
						#Write-Host "$fn_name() Processing property '$cam_prop' ..."
						#Write-Host ("$fn_name() item_obj.$cam_prop type = " + ($item_obj.$cam_prop).GetType())
						#Write-Host ("$fn_name() item_obj.$cam_prop ary = " + ($item_obj_ary | ConvertTo-Json))
						#Write-Host ("$fn_name() cam_prop = $cam_prop, val = '" + $item_obj_ary[0] + "'")
						
						# KEY POINT: $null must put on the left side of operator!!
						# REF: https://stackoverflow.com/a/49411823
						if ($null -eq $item_obj_ary) {
							$new_obj.Add($cam_prop, @('NA'))
							$x = 'NA'  # For single camera, single property requests								
						} else {
							try {
								if ($null -eq $item_obj_ary[0]) {
									$new_obj.Add($cam_prop, @('NA'))
									$x = 'NA'  # for single camera, single property requests
								} else {
									$new_obj.Add($cam_prop, $item_obj_ary)
									$x = $item_obj_ary[0]  # for single camera, single property requests
								}
							}
							catch {
								$new_obj.Add($cam_prop, @('NA'))
								$x = 'NA'  # for single camera, single property requests
							}
						}
						
						# For debugging
						#Write-Host ("$fn_name() " + $cam_prop + " = '"  + $x + "'")	
					}
					
					# For debugging
					#Write-Host ("$fn_name() new_obj = " + $new_obj)
					#Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

					if ($my_patterns_cnt -lt 2) {
						$rtn_obj[$cam] = $new_obj
					} else {
						# Handling multiple -Patterns is a bit more complex
						foreach ($item_obj_key in $item_obj_keys) {
							if ($pattern_cntr -eq 0) {
								# First time...
								if (!$rtn_obj[$cam]) {
									# Initialize the $rtn_obj's $cam keys 
									$rtn_obj.Add($cam, @{})
								}
								# Add & initialize the $rtn_obj[$cam]'s key's prop-value pairs
								$rtn_obj[$cam].Add($item_obj_key, @{})
								$rtn_obj[$cam][$item_obj_key] = @()
							}
							$rtn_obj[$cam][$item_obj_key] += $new_obj[$item_obj_key]
						}
					}

				} else {
					if ($FromConsoleFlg) {
						$err_msg = "ERR/nonexistent camera"
						return $err_msg
					}
					
					$rtn_obj[$cam] = 'NA'
				}
				
			} else {
				if ($FromConsoleFlg) {
					Write-Warning "JSON query returned null/blank`n"
				}
			}

			# Handle single camera, single property request
			if ($rtn_obj -ne 'NA') {
				if ($my_cams_ary.Length -eq 1 -and $my_props_ary.Length -eq 1) {
					$rtn_obj = $x
				}
			}
			
			$pattern_cntr++
		}
		$i++
	}

	# LOG OUT OF API SESSION
	$jsonCmd = @{cmd = 'logout'}
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd
		
	if ($FromConsoleFlg) {
		if ($json_response_obj.result -eq 'fail') {
			Write-Warning "BI API: log out fail; connection not terminated"
		} else {
			if (!$QuietFlg) {
				#Write-Host "BI API: log out successful; connection terminated"	
			}
		}
	}
	
	# For debugging
	#Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))

	return $rtn_obj
}

function BIJsonClipstats() {
# PURPOSE:    To get blue iris json 'clipstats' cmd properties
# PARAMS:     $Clips          = [REQUIRED] (string) comma-delimited QUOTED DB locators, like "@926894649","@926871636"
#                                NOTE: extensions accepted, but not required, e.g., "@926894649.bvr",
#             Props           = [OPTIONAL] (string) comma-delimited JSON properties, like 'name,push,count'
#             $QuietFlg       = [OPTIONAL] (boolean) true suppresses verbose informational output to the console
#             $FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS:    1. (ordered hashtable) if multiple clips and/or properties requested
#             2. (string) if a single clip and a single property requested
#             3  'NA' if the clip is specified and the property does not exists
# USAGE:      BIJsonClipstats -Clips list -Props list -Quiet
# USAGE EXAMPLE
#		$response_obj = BIJsonClipstats -Clips -Props 'camera,path,file,msec,alertmsec,res'
#
#       From the response, set these variables:
#		$clipstats_result = $response_obj[0] 
#                         like, {"result": "success"}, or {"result": "fail"}
#		$clipstats_data = $response_obj[1]
#                         like {"@152866668.bvr": "NA"}
#		                  or   {"@152866668.bvr": {
#			                       "camera": "DW1",
#			                       "path": "@152720932.bvr",
#			                       "file": "DW1.20230530_060000.1965063.17-0.jpg",
#			                       "msec": 3599166,
#			                       "alertmsec": 27426,
#			                       "res": "2688x1520"
#			                       "dummy": "NA"
#		                        }
#
# Version: 2 - 2023-12-10
# Created by jaydeel at ipcamtalk.com

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[String]$Clips,
		[String]$Props,
		[Boolean]$FromConsoleFlg = $false,
		[Boolean]$QuietFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() Clips = '$Clips'"
	#Write-Host "$fn_name() Props = '$Props'"
	
	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$Clips) {
		$err_msg = "ERR/$fn_name() expecting an argument (-Clips)"
		return $err_msg
	}
	$my_clips_ary = $Clips.split(',')
	
	# HANDLE OPTIONAL: COMMANDLINE ARGS
	$my_props_ary = @()
	if ($Props) {
		$my_props_ary = $Props.split(',')
	}

	# For debugging
	#Write-Host ("$fn_name() my_clips_ary = " + $my_clips_ary)
	#Write-Host ("$fn_name() my_clips_ary cnt = " + $my_clips_ary.Length)
	#Write-Host ("$fn_name() my_props_ary = " + $my_props_ary)

	# INITIALIZE
	$rtn_obj = [ordered]@{}
	<# $rtn_obj = [ordered]@{
		success = $null
		data = $null
	} #>
	$x = ""

	if ($my_clips_ary.Length -gt 0) {
		$i = 1;
		foreach ($clip in $my_clips_ary) {
			
			if ($FromConsoleFlg -and !$QuietFlg) {
				if ($i -eq 1) {
					#Write-Host "$fn_name() processing clip '$clip'"
				} else {
					#Write-Host "`n$fn_name() processing clip '$clip'"
				}
			}

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
			$jsonCmd = @{
				cmd = 'clipstats'
				path = $clip
			}
			
			# Returns a Powershell PSCustomObject
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd `
				-QuietFlg $QuietFlg `
				-FromConsoleFlg $FromConsoleFlg

			# For debugging
			#Write-Host ("$fn_name() json_response_obj type = " + $json_response_obj.GetType())
			#Write-Host ("$fn_name() json_response_obj** = " + ($json_response_obj | ConvertTo-Json))
			#Write-Host ("json_response_obj.result** = " + ($json_response_obj.result))

			# If GET failed, process next clip
			if ($json_response_obj -match 'ERR/') {
				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host $json_response_obj
				}
				$rtn_obj[$clip] = 'NA'
				break
			}
			if ($json_response_obj.result -eq "fail") {

				# For debugging
				#Write-Host ("$fn_name() json_response_obj.result type = " + (($json_response_obj.result).GetType()))
				#Write-Host ("$fn_name() son_response_obj.result = " + ($json_response_obj.result | ConvertTo-Json))

				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host "$fn_name() json_response_obj.result = fail"
					Write-Host ($json_response_obj | ConvertTo-Json)
				}
				$rtn_obj[$clip] = 'NA'
				break
			}

			# Extract data from the JSON command response
			if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

				$data_obj = $json_response_obj.data

				# For debugging
				#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
				#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))

				# $data_obj is a PSCustomObject like
				# @{
				#	camera=DW1;
				#	path=@746553655.bvr;
				#	file=DW1.20230428_220000.1572042.17-0.jpg; --OR-- 2023-04-28\\DW2\\DW1.20230428_220000.1572042.17-0.jpg
				#	offset=1572009;
				#	date=1682744400;
				#	color=16711808;
				#	flags=134217985;
				#	zones=0;
				#	memo=;
				#	plate=;
				#	res=2688x1520;
				#	msec=2814017;
				#	alertmsec=24520:
				#	camprofile=7;
				#	camschedule=1;
				#	camsmart=2;
				#	camzones=1;
				#	campreset=30;
				#	filesize=46m55s (419.5M);
				#	filetype=bvr H264 New
				# }

				# Convert PSCustomObject into a hashtable
				# REF: https://tinyurl.com/ddatr635
				$hashtable = @{}
				foreach ($property in $data_obj.psobject.properties.name) {
					$hashtable[$property] = $data_obj.$property
				}
				$data_obj_keys = $hashtable.Keys

				# For debugging
				#Write-Host ("$fn_name() hashtable = " + ($hashtable | ConvertTo-Json))
				#Write-Host "$fn_name() data_obj_keys = $data_obj_keys"

				$new_obj = [ordered]@{}		
				if ($my_props_ary.Length -gt 0) {
					
					# LIST REQUESTED PROPERTIES
					foreach ($clipProp in $my_props_ary) {
						if ($data_obj.$clipProp -or $data_obj.$clipProp -eq 0) {
							if ($FromConsoleFlg -and !$QuietFlg) {
								#Write-Host ($clipProp + " = "  + $data_obj.$clipProp)
							}
							$new_obj.Add($clipProp, $data_obj.$clipProp)
							$x = $data_obj.$clipProp
						} else {
							if ($FromConsoleFlg -and !$QuietFlg) {
								#Write-Host ($clipProp + " = NA")
							}							
							$new_obj.Add($clipProp, "NA")
							$x = "NA"
						}
					}
					$rtn_obj[$clip] = $new_obj
					
				} else {
					
					# LIST ALL PROPERTIES
					foreach ($clip_prop in $data_obj_keys) {
						if ($data_obj.$clip_prop -or $data_obj.$clip_prop -eq 0) {
							if ($FromConsoleFlg -and !$QuietFlg) {
								#Write-Host ($clip_prop + " = "  + $data_obj.$clip_prop)
							}							
							$new_obj.Add($clip_prop, $data_obj.$clip_prop)
						} else {
							if ($FromConsoleFlg -and !$QuietFlg) {
								#Write-Host ($clip_prop + " = NA")
							}							
							$new_obj.Add($clip_prop, "NA")
						}
					}
					$rtn_obj[$clip] = $new_obj
				}
				$i++
			}
		}

		# LOG OUT OF API SESSION
		$jsonCmd = @{cmd = 'logout'}
		BIJsonCall `
			-JsonCmd $jsonCmd `
			-QuietFlg $QuietFlg `
			-FromConsoleFlg $FromConsoleFlg
			
	} else {
		<# $data_obj_keys = @()
		foreach ($item in $data_obj) {
			$data_obj_keys += $($item | Get-Member -MemberType *Property).Name
		}
		foreach ($key in $data_obj_keys) {
			if ($FromConsoleFlg -and !$QuietFlg) {
				#Write-Host ($key + " = "  + $data_obj.$key)					
			}
			$rtn_obj.Add($key, $data_obj.$key)
		} #>
	}

	# Handle single clip, single property request
	if ($my_clips_ary.Length -eq 1 -and $my_props_ary.Length -eq 1) {
		$rtn_obj = $x
	}
								
	<# if ($FromConsoleFlg -and !$QuietFlg) {
		Write-Host "Clipstats:"
		Write-Host "----------"
		Write-Host ($rtn_obj | ConvertTo-Json)
	} #>

	return $rtn_obj
}

function MakeTable() {
# PURPOSE:   To create a CSV file showing filtered DAT files, with following header:
#            AlertID, Flags, HasJPG, HasDAT, RO, HasZeroDate, AlertMemo, JpgsExitMemo, Issues 
#            (This is accomplished by using a temporary CSV file; see note 3 below.)
# PARAMS:    -Path                = (string) path for temporary CSV file is to display hashtable
#            -SortColNo           = (int) column to sort on (first column no = 1)
#            -SortAscFlg          = (boolean) sort ascending
#            -LabelsList          = (list) Alert descriptor
#            -AlertFlagsList      = (list) flags on Alert associated with DAT file (F=flagged,P=protected)
#            -AlertJpgsList       = (list) Alert has JPG file
#            -AlertDatsList       = (list) Alert has DAT file
#            -AlertAttribsList    = (list) JPG and DAT file read-only attributes set
#            -AlertZeroDatesList  = (list) Alert has a date = 0 (zero)
#            -AlMemosList         = (list) Alert memos found for each entry in $FilesList
#            -JpgExifMemos        = (list) extracted 'Title' from JPG Exif metadata
#            -DatMemosList        = (list) confirmed image 'memos' from DAT files
#            -DatMemoImageNosList = (list) confirmed image numbers from DAT files
#            -Issues              = (list) Issues found while inspecting protect Alerts
#            -TimestampFlg        = (boolean) prefix output.csv files with timestamp, like '2023-10-29_1215'
#
# RETURNS:  (string) filespec
# NOTES:    1. Using CSV files is very convenient; Powershell automatically handles the table formatting.
#           2. The temporary CSV files are created in directory "$app_path\_output" ($app_path via the user_settings file)
#           3. The temporary CSV file can be displayed to the console using the following command:
#              --> Import-Csv "output.csv" | Sort-Object "xxx" | Format-Table 
#              A challenge with this approach is that the Sort-Object column has only the -Length property.
#              Because of this, successfully displaying the 'AgeHrs' & 'ImgsCnt' columns in sorted order requires that their values  
#              i) have an identical number of digits after the decimal, and 
#              ii) be left-padded to create strings of identical length.
	param(
		[String]$Path,
		[Int]$SortColNo,
		[Boolean]$SortAscFlg,
		[Array]$LabelsList,
		[Array]$AlertFlagsList,
		[Array]$AlertJpgsList,
		[Array]$AlertDatsList,
		[Array]$AlertAttribsList,
		[Array]$AlertZeroDatesList,
		[Array]$AlMemosList,
		[Array]$JpgExifMemos,
		[Array]$DatMemosList,
		[Array]$DatMemoImageNosList,
		[Array]$IssuesList,
		[Boolean]$TimestampFlg
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() Path = $Path"
	#Write-Host "$fn_name() SortColNo = $SortColNo"
	#Write-Host ("$fn_name() LabelsList type = " + $LabelsList.GetType())
	#Write-Host "$fn_name() LabelsList = $LabelsList"
	#Write-Host ("$fn_name() LabelsList = " + ($LabelsList | ConvertTo-Json))
	#Write-Host ("$fn_name() LabelsList cnt = " + $LabelsList.Length)
	#Write-Host ("$fn_name() AlertFlagsList type = " + $AlertFlagsList.GetType())
	#Write-Host ("$fn_name() AlertFlagsList cnt = " + $AlertFlagsList.Length)
	#Write-Host "$fn_name() AlertFlagsList = $AlertFlagsList"
	#Write-Host ("$fn_name() AlertAttribsList type = " + $AlertAttribsList.GetType())
	#Write-Host ("$fn_name() AlertAttribsList cnt = " + $AlertAttribsList.Length)
	#Write-Host "$fn_name() AlertAttribsList = $AlertAttribsList"
	#Write-Host ("$fn_name() AlertZeroDatesList type = " + $AlertZeroDatesList.GetType())
	#Write-Host ("$fn_name() AlertZeroDatesList cnt = " + $AlertZeroDatesList.Length)
	#Write-Host "$fn_name() AlertZeroDatesList = $AlertZeroDatesList"
	#Write-Host ("$fn_name() AlMemosList type = " + $AlMemosList.GetType())
	#Write-Host ("$fn_name() AlMemosList cnt = " + $AlMemosList.Length)
	#Write-Host "$fn_name() AlMemosList = $AlMemosList"
	#Write-Host ("$fn_name() AlMemosList = " + ($AlMemosList | ConvertTo-Json))
	#Write-Host ("$fn_name() JpgExifMemos type = " + $JpgExifMemos.GetType())
	#Write-Host ("$fn_name() JpgExifMemos cnt = " + $JpgExifMemos.Length)
	#Write-Host "$fn_name() JpgExifMemos = $JpgExifMemos"
	#Write-Host ("$fn_name() DatMemoImageNosList type = " + $DatMemoImageNosList.GetType())
	#Write-Host ("$fn_name() DatMemoImageNosList cnt = " + $DatMemoImageNosList.Count)
	#Write-Host "$fn_name() DatMemoImageNosList = $DatMemoImageNosList"
	#Write-Host ("$fn_name() DatMemoImageNosList = " + ($DatMemoImageNosList | ConvertTo-Json))
	#Write-Host ("$fn_name() DatMemosList type = " + $DatMemosList.GetType())
	#Write-Host ("$fn_name() DatMemosList cnt = " + $DatMemosList.Count)
	#Write-Host "$fn_name() DatMemosList = $DatMemosList"
	#Write-Host ("$fn_name() DatMemosList = " + ($DatMemosList | ConvertTo-Json))
	#Write-Host ("$fn_name() IssuesList type = " + $IssuesList.GetType())
	#Write-Host ("$fn_name() IssuesList cnt = " + $IssuesList.Length)
	#Write-Host "$fn_name() IssuesList = $IssuesList"
	#Write-Host ("$fn_name() IssuesList = " + ($IssuesList | ConvertTo-Json))
	
	# Abort of nothing to do
	if ($LabelsList.Count -eq 0) {
		return
	}

	if ($Path.EndsWith('\') ){
		$Path = $Path.TrimEnd('\')
	}

	# Create directory for temporary CSV file (if necessary).
	# ... CSV file is used to display a hashtable
	If (!(Test-Path -PathType container $Path)) {
		New-Item -ItemType Directory -Path "$Path" | Out-Null
	}
		
	if ($TimestampFlg) {
		$ts_str = Get-Date -Format "yyyy-MM-dd_HHmmss"
		$ts_str = "_" + $ts_str
	} else {
		$ts_str = ""
	}

	# Initialize list collectors; each collector is a column in the displayed output table.
	$labels        = New-Object System.Collections.Generic.List[System.Object]
	$flags         = New-Object System.Collections.Generic.List[System.Object]
	$attribs       = New-Object System.Collections.Generic.List[System.Object]
	$hasjpgs       = New-Object System.Collections.Generic.List[System.Object]
	$hasdats       = New-Object System.Collections.Generic.List[System.Object]
	$haszerodates  = New-Object System.Collections.Generic.List[System.Object]
	$almemos       = New-Object System.Collections.Generic.List[System.Object]
	$imglabels     = New-Object System.Collections.Generic.List[System.Object]
	$jpgmemos      = New-Object System.Collections.Generic.List[System.Object]
	$datmemos      = New-Object System.Collections.Generic.List[System.Object]
	$datmemoimgnos = New-Object System.Collections.Generic.List[System.Object]
	$issues        = New-Object System.Collections.Generic.List[System.Object]
	
	$i = 0
	foreach($lbl in $LabelsList) {
		$labels.Add($lbl)
		$flags.Add($AlertFlagsList[$i])
		$hasjpgs.Add($AlertJpgsList[$i])
		$hasdats.Add($AlertDatsList[$i])
		$attribs.Add($AlertAttribsList[$i])
		$haszerodates.Add($AlertZeroDatesList[$i])
		$almemos.Add($AlMemosList[$i])
		$jpgmemos.Add($JpgExifMemos[$i])
		try {
			$datmemoimgnos.Add($DatMemoImageNosList[$i])
		} 
		catch {
			$datmemoimgnos.Add('')
		}
		try {
			$datmemos.Add($DatMemosList[$i])
		} 
		catch {
			$datmemos.Add('')
		}
		<# if ($DatMemosList[$i].Length -gt 0) {
			$datmemos.Add($DatMemosList[$i])
		} else {
			$datmemos.Add('')
		} #>

		if ($IssuesList.Count -gt 0) {
			if ($IssuesList[$i]) {
				$issues.Add($IssuesList[$i])
			} else {
				$issues.Add('')
			}
		} else {
			$issues.Add('')
		}
		$i++
	}
	
	#Write-Host ("$fn_name() almemos cnt = " + $almemos.Count)
	#Write-Host "$fn_name() almemos = $almemos"
	#Write-Host ("$fn_name() jpgmemos cnt = " + $jpgmemos.Count)
	#Write-Host "$fn_name() jpgmemos = $jpgmemos"
	#Write-Host ("$fn_name() datmemos cnt = " + $datmemos.Count)
	#Write-Host "$fn_name() datmemos = $datmemos"
	#Write-Host ("$fn_name() datmemoimgnos cnt = " + $datmemoimgnos.Count)
	#Write-Host "$fn_name() datmemoimgnos = $datmemoimgnos"

	# Create & populate a hashtable with formated entries.
	# ... each element will be a column in the final table.
	# ... a hashtable is required for creating the temporary CSV file.
	$output = [ordered]@{}
	$output.Add("AlertID", $labels)
	$output.Add("Flags", $flags)
	$output.Add("HasJPG", $hasjpgs)
	$output.Add("HasDAT", $hasdats)
	$output.Add("RO", $attribs)
	$output.Add("HasZD", $haszerodates)
	$output.Add("AlertMemo", $almemos)
	$output.Add("JpgExifMemo", $jpgmemos)
	$output.Add("DatImgNo", $datmemoimgnos)
	$output.Add("DatImgMemo", $datmemos)
	$output.Add("Issues", $Issues)

	# USE A TEMPORARY CSV FILE TO DISPLAY A NICE LOOKING TABLE.

	# EXPORT THE HASHTABLE AS A CSV FILE.
	# REFS: 1. Microsoft everything about hashtables...https://tinyurl.com/3eezjed9
	#       2. https://tommymaynard.com/hash-table-to-csv-revisited/
	#       3. PSObject vs PSCustomObject... https://stackoverflow.com/a/50196654/13499984

	# Get the maximum number of entries ($cnt) from all the columns in the CSV table.
	# ... Note: calculating this is *not* necessary in our case, b/c we already know that all
	#     the columns have the same number of entries; however, it is a good practice in general.
	$cnt  = $output.Values |
			ForEach-Object { $_.Count } |
			Sort-Object |
			Select-Object -Last 1

	# Get the CSV table column headers.
	$keys = $output.Keys

	# For debugging.
	#Write-Host ("column hdrs = " + $output.Keys)
	#Write-Host ("column cnts = " + ($output.Values | ForEach-Object { $_.Count } ))

	$csv_filespec = "$output_path\output" + $ts_str + ".csv"

	# Use hashtable $output to create a custom object for exporting
	# ... this is necessary to save it correctly to CSV
	# ... Export-Csv works with objects. It expects properties and values.
	# Build CSV file, row-by-row. 
	0..($cnt-1) | ForEach-Object {
		$tmp_tbl = [ordered]@{}
		foreach ($key in $keys) {
			$tmp_tbl[$key] = $output[$key][$_]
		}
		New-Object -Type PSObject -Property $tmp_tbl
	} | Export-Csv "$csv_filespec" -NoTypeInformation

	return "$csv_filespec"
}

function ShowTable() {
# PURPOSE:   To output to the console a 'pretty' table of DAT files showing: 
#            filenames, file ages, JPG counts, AI models used, AI labels found, and Alert memos.
# PARAMS:    -Path         = (string) path for temporary CSV file to be displayed
#            -SortColNo    = (int) column to sort on (first column no = 1)
#            -SortAscFlg   = (boolean) sort ascending
# RETURNS:  nothing - function creates output to Console only
# NOTES:    1. Using CSV files is very convenient; Powershell automatically handles the table formatting.
#           2. The temporary CSV files are created in directory "$app_path\_output" ($app_path via the user_settings file)
#
	param(
		[String]$Path,
		[Int]$SortColNo,
		[Boolean]$SortAscFlg
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() Path = $Path"
	#Write-Host "$fn_name() SortColNo = $SortColNo"
	#Write-Host "$fn_name() SortAscFlg = $SortAscFlg"
	#Write-Host "$fn_name() TimestampFlg = $TimestampFlg"

	# Create directory for temporary CSV file (if necessary).
	# ... CSV file is used to display a hashtable
	If (!(Test-Path -LiteralPath $Path)) {
		$err_msg = "ERR/nonexistent CSV file '$Path'"
		return $err_msg
	}

	# DISPLAY THE CSV FILE

	if ($SortColNo -eq 1) {
		if ($SortAscFlg) {
			#Import-Csv "$Path" | Sort-Object "AlertID" | Format-Table 
			Import-Csv "$Path" | Select * -ExcludeProperty "DatImgNo" | Select * -ExcludeProperty "DatImgMemo" | Format-Table 
		} else {
			#Import-Csv "$Path" | Sort-Object "AlertID" -Descending | Format-Table 
			Import-Csv "$Path" | Select * -ExcludeProperty "DatImgNo" | Select * -ExcludeProperty "DatImgMemo" | Sort-Object "AlertID" -Descending | Format-Table 
		}
		
	} else {
		#Import-Csv "$Path" | Format-Table 
		Import-Csv "$Path" | Select * -ExcludeProperty "DatImgNo" | Select * -ExcludeProperty "DatImgMemo" | Format-Table 
	}

}

function MakeDummyJpgFile() {
# PURPOSE:  Create a 'dummy' JPG file
# ARGS:     see param lines below
# RETURNS:  (string) 'success' or 'fail'
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	
    param(
		[String]$JpgFilespec,     # [REQUIRED] full filespec pattern for the dummy JPG file
		[String]$Color,           # background color (default is 'Yellow')
		[Boolean]$AddOverlayFlg,  # apply overlay to dummy JPG image
        [String]$FontName,        # font to use (default is 'Courier New')
		[String]$FontColor         # font color to use (default is 'Black')
 	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() JpgFilespec = $JpgFilespec"
	#Write-Host "$fn_name() Color = $Color"
	#Write-Host "$fn_name() AddOverlayFlg = $AddOverlayFlg"
	#Write-Host "$fn_name() FontName = $FontName"
	
	# Delete existing JPG
	if (Test-Path -LiteralPath "$JpgFilespec") {

		# Just in case, unset the JPG file's read-only attribute
		$obj = Get-ItemProperty -LiteralPath "$JpgFilespec" | Select-Object IsReadOnly
		if ($obj.IsReadOnly) {	
			Set-ItemProperty (Get-Item "$JpgFilespec") -Name IsReadOnly -Value $false
		}

		Remove-Item -LiteralPath "$JpgFilespec" -Force
	}
	
	# Parse params
	if (!$Color) {
		$bg_color = 'Yellow'
	} else {
		$bg_color = $Color
	}
	if (!$FontName) {
		$overlay_font = 'Courier New'
	} else {
		$overlay_font = $FontName
	}
	if (!$FontColor) {
		$font_color = 'Black'
	} else {
		$font_color = $FontColor
	}
					
	# Used in overlay
	$ScriptName = Split-Path -leaf $PSCommandpath

	# @@@ DEV - Used to avoid Windows 11 issues
	$os_ver = [System.Environment]::OSVersion.Version.Major

	# Default (@@@ for now - don't yet know how to get original missing JPG's dimensions)
	$dummy_jpg_width = 848
	$dummy_jpg_height = 480
	$overlay_size = 12
		
	# Load assemblies
	Add-Type -AssemblyName System.Drawing | Out-Null
	Add-Type -AssemblyName System.Windows.Forms | Out-Null	
	
	# Create objects
	$bmp = new-object System.Drawing.Bitmap($dummy_jpg_width,$dummy_jpg_height)
	$brushBg = New-Object Drawing.SolidBrush([System.Drawing.Color]::$bg_color)
	$graphics = [System.Drawing.Graphics]::FromImage($bmp) 
	$graphics.FillRectangle($brushBg,0,0,$bmp.Size.Width,$bmp.Size.Height) 
	
	# For debugging
	#Write-Host ("bmp = " + ($bmp | ConvertTo-Json))
	#Write-Host ("bmp.Size.Width type = " + ($bmp.Size.Width).GetType())
	#Write-Host ("bmp.Size.Width = " + $bmp.Size.Width)
	#Write-Host ("bmp.Size.Height = " + $bmp.Size.Height)
	#Write-Host ("brushBg = " + ($brushBg | ConvertTo-Json))

	if ($AddOverlayFlg) {
		$font = new-object System.Drawing.Font($overlay_font,$overlay_size,[System.Drawing.FontStyle]::Regular)
		$brushFg = New-Object Drawing.SolidBrush([System.Drawing.Color]::$font_color)
		
		# Center the two lines of text displayed the JPG
		$str1 = "dummy JPG file"
		$str2 = "(created by $ScriptName)"  # this string is longer
		# ... get string lengths in pixels
		#     REF: https://stackoverflow.com/a/59518307
		$size_obj_str1 = [System.Windows.Forms.TextRenderer]::MeasureText($str1, $font)						
		$size_obj_str2 = [System.Windows.Forms.TextRenderer]::MeasureText($str2, $font)
		#    $size_obj_str1 is like = {
		#      "IsEmpty": false,
		#      "Width": 148,
		#      "Height": 18
		#    }
		$str_x1 = [Math]::Floor(($dummy_jpg_width - $size_obj_str1.Width)/2)
		$str_x2 = [Math]::Floor(($dummy_jpg_width - $size_obj_str2.Width)/2)
		$str_y1 = [Math]::Floor($dummy_jpg_height/2) - $size_obj_str1.Height
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
		$graphics.DrawString($str1,$font,$brushFg,$str_x1,$str_y1)  # X,Y pixels (from top left corner)
		$graphics.DrawString($str2,$font,$brushFg,$str_x2,$str_y2) 
	}

	# Make the image file
	$bmp.Save($JpgFilespec,[System.Drawing.Imaging.ImageFormat]::Jpeg)
	
	# Release resources
	if ($AddOverlayFlg) {
		if ($null -ne $font) {
			$font.Dispose()
		}
		if ($null -ne $brushFg) {
			$brushFg.Dispose()
		}
	}
	if ($null -ne $brushBg) {
		$brushBg.Dispose()
	}
	if ($null -ne $graphics) {
		$graphics.Dispose()
	}
	if ($null -ne $bmp) {
		$bmp.Dispose()
	}
	
	# For debugging
	#Write-Host ("$fn_name() size_obj_str1 = " + ($size_obj_str1 | ConvertTo-Json))
	#Write-Host ("$fn_name() size_obj_str1.Height = " + $size_obj_str1.Height)
	#Write-Host ("$fn_name() size_obj_str1.Width = " + $size_obj_str1.Width)
	#Write-Host ("$fn_name() size_obj_str2.Width = " + $size_obj_str2.Width)
	#Write-Host "$fn_name() str_x1 = $str_x1"
	#Write-Host "$fn_name() str_x2 = $str_x2"
	#Write-Host "$fn_name() str_y1 = $str_y1"
	#Write-Host "$fn_name() str_y2 = $str_y2"

	# Test if the dummy JPG exists
	If (Test-Path -LiteralPath "$jpg_filespec") {
		$rtn_str = 'success'

		# For debugging
		#Invoke-Item $jpg_filespec

	} else {
		$rtn_str = 'fail'
	}
	
	return $rtn_str

}

Function AddTextToImage {
# PURPOSE:  Add text to a JPG file
# ARGS:     see param lines below
# RETURNS:  (string) 'success', 'fail' or 'ERR/message'
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
# NOTES: 
#   [1] Modified from SOURCE: https://riro.wordpress.com/2010/03/05/add-text-to-images-with-powershell/
#       SOURCE points to original code from http://www.ravichaganti.com/blog/?p=1012
#
# ********** WARNING ****************************************************
# If file $DestPath already exists and you want to retain the original, 
# BE SURE TO BACK IT UP BEFORE USING THIS SCRIPT
# ***********************************************************************

    [CmdletBinding()]
    param (
        [Parameter(Mandatory=$true)][String] $SourcePath,
        [Parameter(Mandatory=$true)][String] $DestPath,
        [Parameter(Mandatory=$true)][String] $Text1,
        [Parameter()][String] $Text2,
        [Parameter()][String] $Location,     # 'TopLeft','TopMiddle','TopRight','BottomLeft','BottomMiddle','BottomRight','Center' (default is 'Center')
        [Parameter()][String] $FontName,     # font to use (default is 'Courier New')
        [Parameter()][Int] $FontSize1,       # font size (default is 20)
        [Parameter()][Int] $FontSize2,       # font size (default is 20)
        [Parameter()][String] $FontStyle,    # font style ('Regular','Bold','Italics') (default is 'Regular')
        [Parameter()][String] $ForeColor,    # text color ('White','black','Red','Blue','Green','Yellow') (default is 'Yellow')
        [Parameter()][String] $BackColor,    # text background color ('White','black','Red','Blue','Green','Yellow') (default is 'Black')
		[Parameter()][Boolean] $FromConsoleFlg
  )
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() SourcePath = $SourcePath"
	#Write-Host "$fn_name() DestPath = $DestPath"
	#Write-Host "$fn_name() Text1 = '$Text1'"
	#Write-Host "$fn_name() Text2 = '$Text2'"
	#Write-Host "$fn_name() Location = '$Location'"
	#Write-Host "$fn_name() FontName = '$FontName'"
	#Write-Host "$fn_name() FontSize1 = $FontSize1"
	#Write-Host "$fn_name() FontSize2 = $FontSize2"
	#Write-Host "$fn_name() FontStyle = $FontStyle"
	#Write-Host "$fn_name() ForeColor = $ForeColor"
	#Write-Host "$fn_name() BackColor = $BackColor"
	
	# ==================================
	# VALIDATE ARGS
	if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
		if (!(Test-Path -LiteralPath "$SourcePath") ){
			$err_msg = "ERR/argument -SourcePath file does not exist ('$SourcePath')"
			if ($FromConsoleFlg) {
				Write-Host "$fn_name() $err_msg" -ForegroundColor Red
			}
			return $err_msg
		}

		if ($DestPath -eq $SourcePath){
			$err_msg = "ERR/arguments -SourcePath and -DestPath cannot be the same"
			if ($FromConsoleFlg) {
				Write-Host "$fn_name() $err_msg" -ForegroundColor Red
			}
			return $err_msg
		}

		$locations_ary = @('TopLeft','TopMiddle','TopRight','BottomLeft','BottomMiddle','BottomRight','Center')
		if (!$Location) {
			$text_loc = 'Center'
		} else {
			if ($locations_ary -contains $Location) {
				$text_loc = $Location
			} else {	
				$text_loc = 'Center'  # Default
				$err_msg = "ERR/unexpected value for argument -Location ('$Location'); using default"
				if ($FromConsoleFlg) {
					Write-Host "$fn_name() $err_msg" -ForegroundColor Red
				}
			}
		}

		if (!$FontName) {
			$text_font = 'Courier New'
		} else {
			$text_font = $FontName
		}

		if (!$FontSize1) {
			$text_size1 = 20
		} else {
			$text_size1 = $FontSize1
		}

		if (!$FontSize2) {
			$text_size2 = $text_size1 - 2
		} else {
			$text_size2 = $FontSize2
		}
		$text_styles_ary = @('Regular','Bold','Italics')

		if (!$FontStyle) {
			$text_style = 'Regular'
		} else {
			if ($text_styles_ary -contains $FontStyle) {
				$text_style = $FontStyle
			} else {	
				$err_msg = "ERR/unexpected value for argument -FontStyle [$text_styles_ary]; using default"
				if ($FromConsoleFlg) {
					Write-Host "$fn_name() $err_msg" -ForegroundColor Red
				}
			}
		}

		$text_colors_ary = @('White','black','Red','Blue','Green','Yellow')
		if (!$ForeColor) {
			$text_color = 'Yellow'
		} else {
			if ($text_colors_ary -contains $ForeColor) {
				$text_color = $ForeColor
			} else {	
				$err_msg = "ERR/unexpected value for argument -ForeColor [$text_colors_ary]; using default"
				if ($FromConsoleFlg) {
					Write-Host "$fn_name() $err_msg" -ForegroundColor Red
				}
			}
		}

		if (!$BackColor) {
			$bg_color = 'black'
		} else {
			if ($text_colors_ary -contains $BackColor) {
				$bg_color = $BackColor
			} else {	
				$err_msg = "ERR/unexpected value for argument -BackColor [$text_colors_ary]; using default"
				if ($FromConsoleFlg) {
					Write-Host "$fn_name() $err_msg" -ForegroundColor Red
				}
			}
		}
	}
	# ==================================

 	# Load assemblies
	Add-Type -AssemblyName System.Drawing | Out-Null
	Add-Type -AssemblyName System.Windows.Forms | Out-Null	

	# Get the image from $SourcePath
	$srcImg = [System.Drawing.Image]::FromFile($SourcePath)
	
	# For debugging
	#Write-Host ("srcImg type = " + $srcImg.GetType())
	#Write-Host ("srcImg = " + ($srcImg | ConvertTo-Json -Depth 32))
	 
	# Create bitmap
	$bmp_obj = new-object System.Drawing.Bitmap([Int]($srcImg.width)),([Int]($srcImg.height))
	
	# For debugging
	#Write-Host ("bmp_obj type = " + $bmp_obj.GetType())
	#Write-Host ("bmp_obj = " + ($bmp_obj | ConvertTo-Json -Depth 32))
 
	# Initialize graphics
	$graphics = [System.Drawing.Graphics]::FromImage($bmp_obj)
	
	# For debugging
	#Write-Host ("Graphics type = " + $graphics.GetType())
	#Write-Host ("Graphics = " + ($graphics | ConvertTo-Json -Depth 32))

	$graphics.SmoothingMode = "AntiAlias"
	$rectangle_obj = New-Object Drawing.Rectangle 0, 0, $srcImg.Width, $srcImg.Height
	$graphics.DrawImage($srcImg, $rectangle_obj, 0, 0, $srcImg.Width, $srcImg.Height, ([Drawing.GraphicsUnit]::Pixel))
 
	# Make graphics font objects
	$font_obj1 = new-object System.Drawing.Font($text_font,$text_size1,[System.Drawing.FontStyle]::$text_style)
	$font_obj2 = new-object System.Drawing.Font($text_font,$text_size2,[System.Drawing.FontStyle]::$text_style)
	
	# For debugging
	#Write-Host ("font_obj1 = " + ($font_obj1 |ConvertTo-Json))
	#Write-Host ("font_obj2 = " + ($font_obj2 |ConvertTo-Json))
 
	# Get string dimensions in pixels
	# REF: https://stackoverflow.com/a/59518307
	# e.g.,  
	#    $size_obj_str1 is like = {
	#      "IsEmpty": false,
	#      "Width": 148,
	#      "Height": 18
	#    }
	$size_obj_str1 = [System.Windows.Forms.TextRenderer]::MeasureText($Text1, $font_obj1)						
	$size_obj_str2 = [System.Windows.Forms.TextRenderer]::MeasureText($Text2, $font_obj2)

	# For debugging
	#Write-Host ("size_obj_str1 = " + ($size_obj_str1 | ConvertTo-Json))
	#Write-Host ("size_obj_str2 = " + ($size_obj_str2 | ConvertTo-Json))

	# @@@ DEV - FUTURE MOD - Determine dimensions for background box
	#$text_W = [Math]::Max([Int]$size_obj_str1.Width, [Int]$size_obj_str2.Width) + 10
	#$text_H = [Int]$size_obj_str1.Height + [Int]$size_obj_str2.Height + 10

	if ($text_loc -match 'TopLeft') {
		$str_x1 = 5
		$str_x2 = 5
		$str_y1 = 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} elseif ($text_loc -match 'TopMiddle') {
		$str_x1 = [Math]::Floor(($srcImg.Width - $size_obj_str1.Width)/2)
		$str_x2 = [Math]::Floor(($srcImg.Width - $size_obj_str2.Width)/2)
		$str_y1 = 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} elseif ($text_loc -match 'TopRight') {
		$str_x1 = $srcImg.width - $size_obj_str1.Width - 5
		$str_x2 = $srcImg.width - $size_obj_str2.Width - 5
		$str_y1 = 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} elseif ($text_loc -match 'BottomLeft') {
		$str_x1 = 5
		$str_x2 = 5
		$str_y1 = $srcImg.height - 1.5*$size_obj_str2.Height - $size_obj_str2.Height - 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} elseif ($text_loc -match 'BottomMiddle') {
		$str_x1 = [Math]::Floor(($srcImg.Width - $size_obj_str1.Width)/2)
		$str_x2 = [Math]::Floor(($srcImg.Width - $size_obj_str2.Width)/2)
		$str_y1 = $srcImg.height - $size_obj_str2.Height - 1.5*$size_obj_str2.Height - 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} elseif ($text_loc -match 'BottomRight') {
		$str_x1 = $srcImg.width - $size_obj_str1.Width - 5
		$str_x2 = $srcImg.width - $size_obj_str2.Width - 5
		$str_y1 = $srcImg.height - $size_obj_str2.Height - 1.5*$size_obj_str2.Height - 5
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	} else {  
		# Center
		$str_x1 = [Math]::Floor(($srcImg.width - $size_obj_str1.Width)/2)
		$str_x2 = [Math]::Floor(($srcImg.width - $size_obj_str2.Width)/2)
		$str_y1 = [Math]::Floor($srcImg.height/2) - $size_obj_str1.Height
		$str_y2 = $str_y1 + 1.5*$size_obj_str2.Height
	}

	# For debugging
	#Write-Host "str_x1 = $str_x1"
	#Write-Host "str_y1 = $str_y1"

	# Make Brush object
	#$Brush = New-Object Drawing.SolidBrush ([System.Drawing.Color]::FromArgb(255, 0, 0,0))
	$brush_obj = New-Object Drawing.SolidBrush ([System.Drawing.Color]::$text_color)
	
	# For debugging
	#Write-Host ("brush_obj = " + ($brush_obj | ConvertTo-Json))
	
	# Draw Text1
	$graphics.DrawString($Text1, $font_obj1, $brush_obj, $str_x1, $str_y1)
 
	# Draw Text2
	if ($null -ne $Text2) {
		#$brush_obj = New-Object Drawing.SolidBrush ([System.Drawing.Color]::FromArgb(120, 0, 0, 0))
		#$brush_obj = New-Object Drawing.SolidBrush ([System.Drawing.Color]::$text_color)
		$graphics.DrawString($Text2, $font_obj2, $brush_obj, $str_x2, $str_y2)
	}

	# Delete existing file if in same location
	# ... Avoids error: Exception calling "Save" with "2" argument(s): "A generic error occurred in GDI+."
	if (Test-Path -LiteralPath "$DestPath") {
		Remove-Item -Path "$DestPath" -Force		
	}
	 
    # Save the markup file
	try {
		$bmp_obj.save($DestPath, [System.Drawing.Imaging.ImageFormat]::Jpeg)
		$rtn_str = 'success'
	} 
	catch {
		if ($FromConsoleFlg) {
			Write-Host $_.Exception.Message -ForegroundColor Red
		}
		$rtn_str = 'fail'
	}

	# Release resources
	$graphics.Dispose()
    $srcImg.Dispose()
    $bmp_obj.Dispose()
	
	return $rtn_str
}

function Get-FileMetaData {
	# SOURCE: https://evotec.xyz/getting-file-metadata-with-powershell-similar-to-what-windows-explorer-provides/
    <#
    .SYNOPSIS
    Small function that gets metadata information from file providing similar output to what Explorer shows when viewing file

    .DESCRIPTION
    Small function that gets metadata information from file providing similar output to what Explorer shows when viewing file

    .PARAMETER File
    FileName or FileObject

    .EXAMPLE
    Get-ChildItem -Path $Env:USERPROFILE\Desktop -Force | Get-FileMetaData | Out-HtmlView -ScrollX -Filtering -AllProperties

    .EXAMPLE
    Get-ChildItem -Path $Env:USERPROFILE\Desktop -Force | Where-Object { $_.Attributes -like '*Hidden*' } | Get-FileMetaData | Out-HtmlView -ScrollX -Filtering -AllProperties

    .NOTES
    #>
    [CmdletBinding()]
    param (
        [Parameter(Position = 0, ValueFromPipeline)][Object] $File,
        [switch] $Signature
    )
    Process {
        foreach ($F in $File) {
            $MetaDataObject = [ordered] @{}
            if ($F -is [String]) {
                $FileInformation = Get-ItemProperty -Path $F
            } elseif ($F -is [System.IO.DirectoryInfo]) {
                #Write-Warning "Get-FileMetaData - Directories are not supported. Skipping $F."
                continue
            } elseif ($F -is [System.IO.FileInfo]) {
                $FileInformation = $F
            } else {
                #Write-Warning "Get-FileMetaData - Only files are supported. Skipping $F."
                continue
            }
            $ShellApplication = New-Object -ComObject Shell.Application
            $ShellFolder = $ShellApplication.Namespace($FileInformation.Directory.FullName)
            $ShellFile = $ShellFolder.ParseName($FileInformation.Name)
            $MetaDataProperties = [ordered] @{}
            0..400 | ForEach-Object -Process {
                $DataValue = $ShellFolder.GetDetailsOf($null, $_)
                $PropertyValue = (Get-Culture).TextInfo.ToTitleCase($DataValue.Trim()).Replace(' ', '')
                if ($PropertyValue -ne '') {
                    $MetaDataProperties["$_"] = $PropertyValue
                } else {
					#$MetaDataProperties["$_"] = ''
				}
            }
            foreach ($Key in $MetaDataProperties.Keys) {
                $Property = $MetaDataProperties[$Key]
                $Value = $ShellFolder.GetDetailsOf($ShellFile, [Int] $Key)
                if ($Property -in 'Attributes', 'Folder', 'Type', 'SpaceFree', 'TotalSize', 'SpaceUsed') {
                    continue
                }
                If (($null -ne $Value) -and ($Value -ne '')) {
                    $MetaDataObject["$Property"] = $Value
                }
            }
            if ($FileInformation.VersionInfo) {
                $SplitInfo = ([String] $FileInformation.VersionInfo).Split([char]13)
                foreach ($Item in $SplitInfo) {
					
					# For debugging
					#Write-Host "Item = '$Item'"
					#Write-Host ("Property = " + ($Property | ConvertTo-Json))

					if ($item) {
						[String[]]$Property = $Item.Split(":").Trim()
					
						# For debugging
						#Write-Host ("Property = " + ($Property | ConvertTo-Json))

						if ($Property[0] -and $Property[1] -ne '') {
							$MetaDataObject["$($Property[0])"] = $Property[1]
						}
					}
                }
            }
            $MetaDataObject["Attributes"] = $FileInformation.Attributes
            $MetaDataObject['IsReadOnly'] = $FileInformation.IsReadOnly
            $MetaDataObject['IsHidden'] = $FileInformation.Attributes -like '*Hidden*'
            $MetaDataObject['IsSystem'] = $FileInformation.Attributes -like '*System*'
            if ($Signature) {
                $DigitalSignature = Get-AuthenticodeSignature -FilePath $FileInformation.Fullname
                $MetaDataObject['SignatureCertificateSubject'] = $DigitalSignature.SignerCertificate.Subject
                $MetaDataObject['SignatureCertificateIssuer'] = $DigitalSignature.SignerCertificate.Issuer
                $MetaDataObject['SignatureCertificateSerialNumber'] = $DigitalSignature.SignerCertificate.SerialNumber
                $MetaDataObject['SignatureCertificateNotBefore'] = $DigitalSignature.SignerCertificate.NotBefore
                $MetaDataObject['SignatureCertificateNotAfter'] = $DigitalSignature.SignerCertificate.NotAfter
                $MetaDataObject['SignatureCertificateThumbprint'] = $DigitalSignature.SignerCertificate.Thumbprint
                $MetaDataObject['SignatureStatus'] = $DigitalSignature.Status
                $MetaDataObject['IsOSBinary'] = $DigitalSignature.IsOSBinary
            }
            [PSCustomObject] $MetaDataObject
        }
    }
}

function FindBytes() {
# PURPOSE:  Find all starting indices in $Bytes where $Search was found
# ARGS:     see param lines below
# RETURNS:  (array) integers, or empty array @()
# Source:   https://stackoverflow.com/a/62511302/13499984
# Version  2 (2/27/2024)

	param (
		[Byte[]]$Bytes, 
		[Byte[]]$Search, 
		[Int]$Start, 
		[Switch]$All		
	)
	
	[Int[]]$rtn_ary = @()
	
    for ($idx = $Start; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
        for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
        if ($i -ge $Search.Length) { 
            $rtn_ary += $idx
            if (!$All) { 
				return $rtn_ary
			}
        } 
    }
	
	# KEY POINT: By default, Powershell functions will return $null if attempting to return a empty array.
	# e.g., return @()
	# To avoid this, the options are:
	# 1) return ,$rtn_ary  ... where the comma is the Powershell 'array construction operator'
	# 2) write-output $arr -NoEnumerate 
	# REF: https://stackoverflow.com/a/18477004

	return ,$rtn_ary
}

function GetJpgExifMetadata() {
# PURPOSE: To get selected Exif metadata for Blue Iris Alert JPG files
# PARAMS:  -FilePath = (string) full filespec for the BI Alert JPG file
# RETURNS: (hashtable) @{'title':str, 'subject':str, 'comments':str, 'datetaken':str, 'width':int, 'height':int}
# For reference... the $exif_obj for a Blue Iris JPG is like:
<# exif_obj = {
	  "Name": "DW2.20240120_110000.1822971.5-0.jpg",
	  "Size": "81.6 KB",
	  "ItemType": "ACDSee JPG Image",
	  "DateModified": "1/22/2024 3:26 PM",
	  "DateCreated": "1/22/2024 3:26 PM",
	  "DateAccessed": "1/22/2024 3:26 PM",
	  "PerceivedType": "Image",
	  "Owner": "Administrators",
	  "Kind": "Picture",
	  "DateTaken": "‎1/‎20/‎2024 ‏‎11:30 AM",
	  "Rating": "Unrated",
	  "Title": "vehicle:93%xxx",
	  "Subject": "vehicle:93%xxx",
	  "Comments": "1.1827877.399365-64..22474.22402.1705779033000",
	  "Dimensions": "‪848 x 480‬",
	  "ProgramName": "Blue Iris",
	  "Computer": "DELL-9020 (this PC)",
	  "FileExtension": ".jpg",
	  "Filename": "DW2.20240120_110000.1822971.5-0.jpg",
	  "BitDepth": "24",
	  "HorizontalResolution": "‎96 dpi",
	  "Width": "‎848 pixels",
	  "VerticalResolution": "‎96 dpi",
	  "Height": "‎480 pixels",
	  "Shared": "No",
	  "FolderName": "Alerts",
	  "FolderPath": "D:\\BlueIris\\Alerts",
	  "Path": "D:\\BlueIris\\Alerts\\DW2.20240120_110000.1822971.5-0.jpg",
	  "LinkStatus": "Unresolved",
	  "SharingStatus": "Not shared",
	  "File": "D",
	  "Debug": "False",
	  "Patched": "False",
	  "PreRelease": "False",
	  "PrivateBuild": "False",
	  "SpecialBuild": "False",
	  "Attributes": 33,
	  "IsReadOnly": true,
	  "IsHidden": false,
	  "IsSystem": false
	}
#>

	param(
		[String]$FilePath
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	
	# Initialize return variable
	$rtn_obj = @{
		'title'     = ''
		'subject'   = ''
		'comments'  = ''
		'datetaken' = ''
		'width'     = 0
		'height'    = 0
	}

	if (Test-Path -LiteralPath "$FilePath") {
		
		# GET EXIF METADATA AS SHOWN BY WINDOWS EXPLORER
		# ... SOURCE: https://evotec.xyz/getting-file-metadata-with-powershell-similar-to-what-windows-explorer-provides/
		
		$exif_obj = Get-ChildItem -Path $FilePath -Force | Get-FileMetaData
		
		# For debugging
		#Write-Host ("$fn_name() exif_obj 'as is' = " + ($exif_obj | ConvertTo-Json))

		# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		# ALTERNATIVE
		# Cmdlet 'Get-Member' errs if Exit 'Title' is blank, or nonexistent
		#$exif_title_obj = Get-Member -InputObject $exif_obj -name "Title" -Membertype Properties

		#Write-Host ("$fn_name() exif_title_obj = " + $exif_title_obj)
		
		# $exif_title_obj is like
		#exif_title_obj = {
			#"TypeName": "System.Management.Automation.PSCustomObject",
			#"Name": "Title",
			#"MemberType": 8,
			#"Definition": "string Title=person:86%,deer:55%"
		#}
		
		#$exif_title = $exif_title_obj.Definition.Replace('string Title=', '')
		# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

		# VERIFY & EXTRACT EXIF 'PROPERTIES'
		
		# Sometimes these properties do not exit
		if (Get-Member -InputObject $exif_obj -name "Title" -Membertype Properties) {
			$rtn_obj['title'] = $exif_obj.Title
		}	
		# ... apparently redundant to $exif_obj.Title	
		if (Get-Member -InputObject $exif_obj -name "Subject" -Membertype Properties) {
			$rtn_obj['subject'] = $exif_obj.Subject
		}					
		# ... like, '1.1827877.399365-64..22474.22402.1705779033000'
		if (Get-Member -InputObject $exif_obj -name "Comments" -Membertype Properties) {
			$rtn_obj['comments'] = $exif_obj.Comments
		}					
		# ... like, '7/28/2023 3:28 PM'
		if (Get-Member -InputObject $exif_obj -name "DateTaken" -Membertype Properties) {
			# Need tp remove extra characters (left-to-right mark, and right-to-left mark)
			# REF: https://stackoverflow.com/a/25477023										
			$datetaken_str = ($exif_obj.DateTaken -Replace [char]8206) -Replace [char]8207
			$datetaken_ary = $datetaken_str -Split ''
			$datetaken_ary = $datetaken_ary[1..($datetaken_ary.Length-2)]
			$rtn_obj['datetaken'] = $datetaken_ary -Join ''

			#For debugging
			#Write-Host ("$fn_name() datetaken_ary = " + ($datetaken_ary | ConvertTo-Json))
		}					

		# These properties always exist...
		$rtn_obj['width']     = [Int]($exif_obj.Width -Replace "[^0-9]" , '')
		$rtn_obj['height']    = [Int]($exif_obj.Height -Replace "[^0-9]" , '')
		
	}
		
	# For debugging
	#Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))

	return $rtn_obj
}

function UpdateDbRecord() {
# PURPOSE:  Execute the Blue Iris JSON 'update' command, which modifies the Alert's database entry 
#           IN ADDITION - this function also modifies the JPG'S EXIF METADATA
# USAGE:    {this_script_name}.ps1 prop cam(s) prop(s)
# PARAMS:   -AlertDbPath      = (string) Blue Iris Alert's database record ID
#           -AlertDbDateSec      = (int) UTC date in seconds
#           -AlertDbMemo      = (string) Blue Iris Alert's database memo text
#           -AlertDbFlags     = (int) Blue Iris Alert's database flags (ORIGINAL VALUES, as returned by cliplist or alertlist)
#           -UpdateJpgExif    = (switch) if used, must also use arguments -JpgFileSpec 
#           -JpgFileSpec      = (string) full path for Alert's JPG file
#           -QuietFlg         = (boolean) true suppresses verbose informational output to the console
#           -FromConsoleFlg   = (boolean) $true notifies function that calling command was executed from the console
#           -DevModeFlg       = (boolean)
# RETURNS:  (string) "success' or 'fail'
# REQUIRES: GetJpgExifMetadata()
# KEY POINT: 
#   For Alerts that have the 'Protected' flag enabled, BI's JSON 'update' command automatically 
#   sets the Alert's JPG and DAT files to read-only (if they exist). 
#   THIS SCRIPT WILL UNSET THE READ-ONLY ATTRIBUTES FOR THESE FILES IF THEY WERE NOT ORIGINALLY READ-ONLY

	param(
		[String]$AlertDbPath,
		[Int]$AlertDbDateSec,
		[Int]$AlertDbFlags,
		[String]$AlertDbMemo,
		[Switch]$UpdateJpgExif,
		[String]$JpgFileSpec,
		[Boolean]$FromConsoleFlg,
		[Boolean]$QuietFlg,
		[Boolean]$DevModeFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() AlertDbPath = '$AlertDbPath'"
	#Write-Host "$fn_name() AlertDbDateSec = $AlertDbDateSec"
	#Write-Host "$fn_name() AlertDbFlags = $AlertDbFlags"
	#Write-Host "$fn_name() AlertDbMemo = '$AlertDbMemo'"
	#Write-Host ("$fn_name() AlertDbMemo len = " + $AlertDbMemo.Length)
	#Write-Host "$fn_name() UpdateJpgExif = '$UpdateJpgExif'"
	#Write-Host "$fn_name() JpgFileSpec = '$JpgFileSpec'"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	
	if ($UpdateJpgExif -and !$JpgFileSpec) {
		if ($FromConsoleFlg) {
			Write-Warning "Arguments -UpdateJpgExif requires -JpgFileSpec"
		}
		return 'fail'
	}
	
	$rtn_str = 'fail'  # Assume failure

	# GET THE INCOMING READ-ONLY ATTRIBUTES FOR THE ALERT'S JPG & DAT FILES (if they exist)
	
	# KEY POINT: 
	# Alert's with an enabled 'Protected' flag will automatically set their JPG and DAT files to read-only

	# ... set defaults
	$jpg_originally_readonly_flg = $false
	$dat_originally_readonly_flg = $false	
	# ... get current attributes
	if ($JpgFileSpec) {
		$jpg_exists_flg = (Test-Path -LiteralPath "$JpgFileSpec")
		if ($jpg_exists_flg) {
			$obj = Get-ItemProperty -LiteralPath "$JpgFileSpec" | Select-Object IsReadOnly
			$jpg_originally_readonly_flg = $obj.IsReadOnly
		}
		$dat_filespec = $JpgFileSpec.Replace('.jpg', '.dat')
		$dat_exists_flg = (Test-Path -LiteralPath "$dat_filespec")
		if ($dat_exists_flg) {
			$obj = Get-ItemProperty -LiteralPath "$dat_filespec" | Select-Object IsReadOnly
			$dat_originally_readonly_flg = $obj.IsReadOnly
		}
	} else {
		$jpg_exists_flg = $false
		$jpg_originally_readonly_flg = $false
		$dat_exists_flg = $false
		$dat_originally_readonly_flg = $false		
	}

	# For debugging
	#Write-Host ("$fn_name() JPG BEFORE is read-only = " + $jpg_originally_readonly_flg)
	#Write-Host ("$fn_name() DAT BEFORE is read-only = " + $dat_originally_readonly_flg)

	# TEMPORARILY CHANGE SET THE ALERT'S DATABASE MEMO

	# KEY POINTS: 
	# 1) The JPG's metadata will NOT be changed if the Alert's memo is not actually changed
	#    so we temporarily change the memo to a dummy value
	# 2) Alert's flags will be removed if param 'flags' is not included in the JSON 'update' command

	if ($UpdateJpgExif) {

		$jsonCmd = @{
			cmd   = 'update'
			path  = $AlertDbPath
			date  = [String]($AlertDbDateSec*1000)  # must be string milliseconds
			flags = $AlertDbFlags
			memo  = 'xxx'
		}		
		# Returns a powershell PSCustomObject
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-QuietFlg $QuietFlg `
			-FromConsoleFlg $FromConsoleFlg

		# For debugging
		#Write-Host ("$fn_name() jsonCmd = " + ($jsonCmd | ConvertTo-Json))
		#Write-Host ("$fn_name() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
		
		if (!($json_response_obj.result) -match 'success') {
			return 'fail'
		}
		
	}

	# EXECUTE JSON 'UPDATE' CMD

	# KEY POINT: This action INDIRECTLY updates the JPG file's Exif metadata
	# How? Blue Iris refreshes the JPG's Exif metadata when the Alert's DB fields are JSON 'updated'
	# We will verify this change

	# 1) Make jsonCmd object
	if ([Int]$AlertDbDateSec -gt 0) {
#Write-Host "$fn_name() updating a non-zero date db rec"
		$jsonCmd = @{
			cmd   = 'update'
			path  = $AlertDbPath
			date  = [String]($AlertDbDateSec*1000)  # must be string milliseconds
			flags = $AlertDbFlags
			memo  = $AlertDbMemo
		}
	} else {
#Write-Host "$fn_name() updating a zero date db rec"
		# BI's JSON 'update' cmd will not accept a [date] of '0' (zero), so we use '1'
		# This provides the desired result; the JSON 'clipstats' cmd shows [date] => 0
		$jsonCmd = @{
			cmd   = 'update'
			path  = $AlertDbPath
			date  = '1'
			flags = $AlertDbFlags
			memo  = $AlertDbMemo
		}
	}		

	# 2) Execute the JSON all
	#    Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("$fn_name() jsonCmd = " + ($jsonCmd | ConvertTo-Json))
	#Write-Host ("$fn_name() json_response_obj = " + ($json_response_obj | ConvertTo-Json))

	# VERIFY CHANGE TO JPG'S METADATA 
	if ($json_response_obj.result -match 'success') {

		if ($UpdateJpgExif) {

			#For Debugging
			#Write-Host "$fn_name() JpgFileSpec = '$JpgFileSpec'"
			
			# Returns hashtable like @{'title: str, 'subject': str, 'comments': str ... }
			# NOTE: function tests existence of dat_filespec
			$exif_obj = GetJpgExifMetadata -FilePath $jpg_filespec

			# For debugging
			#Write-Host ("exif_obj type = " + $exif_obj.GetType())
			#Write-Host ("exif_obj = " + ($exif_obj | ConvertTo-Json))

			$exif_title = $exif_obj.title
			
			# For debugging
			#Write-Host ("exif_title = " + ($exif_title | ConvertTo-Json))
			
			if ($exif_title -eq $AlertDbMemo) {
				$rtn_str = 'success'
			} else {
				# For debugging
				#Write-Host "$fn_name() exif title ($exif_title) <> AlertDbMemo ($AlertDbMemo)"
			}
			
		} else {
			$rtn_str = 'success'
		}
		
	}

	# For debugging
	#$obj = Get-ItemProperty -LiteralPath "$JpgFileSpec" | Select-Object IsReadOnly
	#Write-Host ("$fn_name() JPG AFTER is read-only = " + $obj.IsReadOnly)
	#$obj = Get-ItemProperty -LiteralPath $JpgFileSpec.Replace('.jpg','.dat') | Select-Object IsReadOnly
	#Write-Host ("$fn_name() DAT AFTER is read-only = " + $obj.IsReadOnly)
		
	# RESTORE THE OUTGOING READ-ONLY ATTRIBUTES FOR THE ALERT'S JPG & DAT FILES (if they exist)
	if ($jpg_exists_flg -and !$jpg_originally_readonly_flg) {	
		Set-ItemProperty (Get-Item "$JpgFileSpec") -Name IsReadOnly -Value $false
	}
	if ($dat_exists_flg -and !$dat_originally_readonly_flg) {	
		Set-ItemProperty (Get-Item "$dat_filespec") -Name IsReadOnly -Value $false
	}

	return $rtn_str
}

function Convert-FromUnixDate() {
# Returns string like '02/01/2024 21:34:00'

	param (
		[Int]$UnixDate,
		[String]$FormatStr  # like, "yyyy-MM-dd"
	)

	if ($FormatStr) {
		$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate)) | Get-Date -f $FormatStr
	} else {
		$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate))
	}
   
   return $rtn_str
   
}

function Convert-ToUnixDate() {
# PURPOSE:    To convert a simple date string to a Unix timestamp
# PARAMS:     -DateStr   = a date string, e.g.,  '01/08/2024 11:10:42'
#             -FormatStr = exact pattern match to -DateStr value, like "MM/dd/yyyy HH:mm:ss"     
# RETURNS:    (int) Unix epoch timestamp (seconds)
# SOURCE:      modified from https://stackoverflow.com/a/72945413
# NOTES:  

	param (
		[String]$DateStr,   # a date string, e.g.,  '01/08/2024 11:10:42'
		[String]$FormatStr  # exact pattern match to -DateStr value, like "MM/dd/yyyy HH:mm:ss"
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	if ($FormatStr) {
		$format_str = $FormatStr
	} else {
		$format_str = "MM/dd/yyyy HH:mm:ss"
	}

	# For debugging.
	#Write-Host "entered function $fn_name() ..."
	#Write-Host "$fn_name() DateStr = '$DateStr'"
	#Write-Host "$fn_name() FormatStr = '$FormatStr'"
	#Write-Host "$fn_name() format_str = '$format_str'"

	# Because there's no timezone component in $DateStr (e.g. "z" or "zz"),
	# ParseExact() converts it to a datetime with "Kind = DateTimeKind.Unspecified"
	# ... see http://tinyurl.com/mw8375fp
	$timestamp = [datetime]::ParseExact($DateStr, $format_str, $null)
	# DateStr    : 01/08/2024 11:10:42
	# Kind       : Unspecified
	
	# For debugging.
	#Write-Host "$fn_name() timestamp = '$timestamp'"

	# tz_obj.Id                         : like, SA Eastern Standard Time
	# tz_obj.DisplayName                : like, (UTC-03:00) Cayenne, Fortaleza
	# tz_obj.StandardName               : like, SA Eastern Standard Time
	# tz_obj.DaylightName               : like, SA Eastern Summer Time
	# tz_obj.BaseUtcOffset              : like, -03:00:00
	# tz_obj.SupportsDaylightSavingTime : like, False
	$tz_obj = (Get-TimeZone)

	# This is the magic bit - treat $timestamp as a local time in timezone $tz.obj, 
	# and convert it to UTC using the BaseUtcOffset and daylight saving rules for $tz_obj
	$utc_time = [TimeZoneInfo]::ConvertTimeToUtc($timestamp, $tz_obj);
	# DateTime    : 09 July 2022 13:11:12
	# Kind        : Utc

	# now convert it to a Unix epoch timestamp
	$epoch_time = [Int]($utc_time | Get-Date -UFormat "%s")
	
	return $epoch_time
}

function IsJpegFileValid {
# PURPOSE: To confirm that a JPG file is a valid JPG image
# NOTES:
#    Using 2 methods:
#    1. Test image type
#       REFS:  http://learningpcs.blogspot.com/2011/07/powershell-validate-jpeg-files.html
#              https://www.powershellgallery.com/packages/PScribo/0.9.1/Content/Src%5CPrivate%5CGet-ImageMimeType.ps1
#    2. Per the JPEG Interchange Format... any JPEG file or stream that is written according to the 
#       JPEG Interchange Format means that they must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) 
#       and end with the EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#       These two markers will not be found anywhere else in a JPEG file/stream.
#       REFS: https://stackoverflow.com/a/46809226
#             https://www.codeproject.com/Questions/421855/Verifying-a-file-is-JPEG
# DISCUSSION:
#    Originally I used only Method #1 in this function. However, I learned later that Method #1 would still return valid
#    if the image bytes did not terminate with the EOI (End-of-Image) marker (0xFF, 0xD9). These JPG files still opened 
#    just fine in every windows viewer I tried (Photos, Paint, Paint3D, PhotoScape X, ACDSee...), HOWEVER, if they were 
#    opened in a web browser (tested Chrome, Edge) the images had a white band across the bottom. 
#    Additionally, web apps like https://app.roboflow.com/ would not import the JPGs.
#    So clearly, Method 1 is not sufficient. For redundancy, I've opted to use both methods. 

	param(
		[String]$FilePath
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#msg.exe $script:server_username /TIME:15 "entered function $fn_name()..."
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() FilePath = '$FilePath'"
	#msg.exe $script:server_username /TIME:30 "$fn_name() FilePath = '$FilePath'"
	
	# Initialize flags
	$method1_flg = $false
	$method2_flg = $false
	
	# ===========================
	# METHOD 1
	# Validate Image file Type

	# WARNING: The following code line IS ESSENTIAL when executing from Task Scheduler (or Blue Iris)
	#          (however works just fine from Powershell ISE)
	#          REF: https://tinyurl.com/2p929nhz
	Add-type -AssemblyName System.Drawing
	
	try {
		$img = [System.Drawing.Image]::FromFile("$FilePath");
		$method1_flg = $img.RawFormat.Equals([System.Drawing.Imaging.ImageFormat]::Jpeg)
		$img.Dispose()  # Release resources

	} catch [OutOfMemoryException] {
		$method1_flg = $false
	}

	# ===========================
	# METHOD 2
	# Validate bytes per the 'JPEG Interchange Format' (see Notes above)
	
	# Import the file to a byte[] array
	$ImageBytes = [System.IO.File]::ReadAllBytes($FilePath)
	
	# Capture the start and end byte pairs.
	#$bytes_rng1 = $ImageBytes[0],$ImageBytes[1]
	#$bytes_rng2 = $ImageBytes[$ImageBytes.Count-2],$ImageBytes[$ImageBytes.Count-1]
	$bytes_rng1 = $ImageBytes[0..1]
	$bytes_rng2 = $ImageBytes[($ImageBytes.Count-2)..($ImageBytes.Count-1)]

	# Define the required JPG start/end markers
	$markers1 = @(255,216) # hexArray @("FF", "D8")
	$markers2 = @(255,217) # hexArray @("FF", "D9")
	<# $hexArray = "FF", "D8"
	$markers1 = @()
	foreach ($hex in $hexArray) {
		$markers1 += [Int32]"0x$hex"
	}
	$hexArray = "FF", "D9"
	$markers2 = @()
	foreach ($hex in $hexArray) {
		$markers2 += [Int32]"0x$hex"
	} #>

	# Validate the start and end byte pairs.
	# ... Note: Method Compare-Object returns null if the arrays are the identical
	$x1 = Compare-Object $bytes_rng1 $markers1
	$x2 = Compare-Object $bytes_rng2 $markers2
	if ($null -eq $x1 -and $null -eq $x2) {
		$method2_flg = $true
	} else {
		$method2_flg = $false
	}
	
	# For testing
	#$method2_flg = $true

	# For debugging
	#Write-Host "markers1 = $markers1, markers2 = $markers2"
	#Write-Host "bytes_rng1 = $bytes_rng1, bytes_rng2 = $bytes_rng2"
	#Write-Host "method1_flg = $method1_flg, method2_flg = $method2_flg"
	#msg.exe $script:server_username /TIME:15 "$fn_name() method1_flg = $method1_flg, method2_flg = $method2_flg"
	
	if ($method1_flg -and $method2_flg) {
		return $true
	} else {
		return $false
	}
}

function GetJpgFileExifDateTaken() {
# PURPOSE:  Get JPG file's Exif metadata DAteTaken
# ARGS:     -FilePath = (string ) Full filespec of JPG file
# RETURNS:  (string) like ''
# REQUIRES: GetJpgExifMetadata()

	param(
		[String]$FilePath  # Full filespec of JPG file
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() FilePath = '$FilePath'"
	
	$rtn_str = 'fail'  # Assume failure
	
	$exif_obj = GetJpgExifMetadata -FilePath $FilePath
	
	# For debugging
	#Write-Host("exif_obj = " + ($exif_obj | ConvertTo-Json))
	#Write-Host("exif_obj.datetaken type = " + ($exif_obj.datetaken).GetType())
	#Write-Host("exif_obj.datetaken value = " + $exif_obj.datetaken)

	if ($exif_obj.datetaken) {
		
		# REF: https://www.sharepointdiary.com/2021/11/date-format-in-powershell.html
		$format_str = "M/d/yyyy h:mm tt"
		$datetaken_sec = `
			Convert-ToUnixDate `
			-DateStr $exif_obj.datetaken `
			-FormatStr "M/d/yyyy h:mm tt"  # must be EXACTLY this
		
		# Convert unix timestamp to string like '02/01/2024 21:34:00'
		#$target_date_str = Convert-FromUnixDate $datetaken_sec "M/d/yyyy h:mm:ss tt"
		$rtn_str = Convert-FromUnixDate $datetaken_sec

	}
	
	return $rtn_str
	
}

function SetJpgFileTimes() {
# PURPOSE:  Set JPG file's CreationTime, LastWriteTime & LastAccessTime equal its Exif metadata DAteTaken
# ARGS:     -FilePath = (string ) Full filespec of JPG file
# RETURNS:  (string) 'success', or 'fail'
# REQUIRES: GetJpgExifMetadata()

	param(
		[String]$FilePath  # Full filespec of JPG file
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	#Write-Host "$fn_name() FilePath = '$FilePath'"

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	
	$rtn_str = 'fail'  # Assume failure
	
	$exif_obj = GetJpgExifMetadata -FilePath $FilePath
	
	# For debugging
	#Write-Host("exif_obj = " + ($exif_obj | ConvertTo-Json))
	#Write-Host("exif_obj.datetaken type = " + ($exif_obj.datetaken).GetType())
	#Write-Host("exif_obj.datetaken value = " + $exif_obj.datetaken)

	if ($exif_obj.datetaken) {
		
		# REF: https://www.sharepointdiary.com/2021/11/date-format-in-powershell.html
		$format_str = "M/d/yyyy h:mm tt"
		$datetaken_sec = `
			Convert-ToUnixDate `
			-DateStr $exif_obj.datetaken `
			-FormatStr "M/d/yyyy h:mm tt"  # must be EXACTLY this
		
		# Convert unix timestamp to string like '02/01/2024 21:34:00'
		#$target_date_str = Convert-FromUnixDate $datetaken_sec "M/d/yyyy h:mm:ss tt"
		$target_date_str = Convert-FromUnixDate $datetaken_sec

		# Just in case, UNSET THE JPG FILE'S READ-ONLY ATTRIBUTE
		$obj = Get-ItemProperty -LiteralPath "$FilePath" | Select-Object IsReadOnly
		$originally_readonly_flg = $obj.IsReadOnly
		
		if ($originally_readonly_flg) {	
			Set-ItemProperty (Get-Item "$FilePath") -Name IsReadOnly -Value $false
		}
		
		(Get-Item "$FilePath").CreationTime = $target_date_str
		(Get-Item "$FilePath").LastWriteTime = $target_date_str
		#(Get-Item "$FilePath").LastAccessTime = $target_date_str

		# Restore the JPG file's original read-only attribute
		if ($originally_readonly_flg) {	
			Set-ItemProperty (Get-Item "$FilePath") -Name IsReadOnly -Value $true
		}

		# For debugging
		#Write-Host "$fn_name() originally_readonly_flg = $originally_readonly_flg ($FilePath)"
		#Write-Host ("JPG file's Exif DateTaken = " + $exif_obj.datetaken + " (" + $datetaken_sec + ")")
		#Write-Host ("JPG file's CreationTime   = " + $target_date_str + " (" + $datetaken_sec + ")")
		#Write-Host ("JPG file's LastWriteTime  = " + $target_date_str + " (" + $datetaken_sec + ")")

		$rtn_str = 'success' 

	}
	
	return $rtn_str
	
}

function RepairDatFile() {
# PURPOSE: Try to repair a DAT file with unexpected structure [unbalanced JPG beg/end markers]
# ARGS:    see param()
# NOTE:    Repair is attempted only when the count of 'end markers' <> the count of 'beg markers'
# RETURNS  [Array] @([Boolean]rtn_flg, [Array]SOI_mrks, [Array]EOI_mrks)

	param (
		[Byte[]]$Bytes,
		[Array]$BegOfImageMarks,
		[Array]$EndOfImageMarks
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "Entered $fn_name()..."
	#Write-Host ("$fn_name() BegOfImageMarks type = " + $BegOfImageMarks.GetType())
	#Write-Host ("$fn_name() BegOfImageMarks cnt = " + $BegOfImageMarks.Count)
	#Write-Host "$fn_name() BegOfImageMarks = $BegOfImageMarks"
	#Write-Host ("$fn_name() EndOfImageMarks type = " + $EndOfImageMarks.GetType())
	#Write-Host ("$fn_name() EndOfImageMarks cnt = " + $EndOfImageMarks.Count)
	#Write-Host "$fn_name() EndOfImageMarks = $EndOfImageMarks"
	
	# Initialize vars
	$rtn_flg = $true  # Assume success
	$SOI_mrks_cnt = $BegOfImageMarks.Count
	$EOI_mrks_cnt = $EndOfImageMarks.Count		

	# Must convert arrays to ArrayLists to use method 'RemoveAt()' (later)
	$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$BegOfImageMarks
	$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EndOfImageMarks

	# For debugging
	#Write-Host ("$fn_name() SOI_mrks_lst type = " + $SOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() SOI_mrks_lst cnt = " + $SOI_mrks_lst.Count)
	#Write-Host "$fn_name() SOI_mrks_lst = $SOI_mrks_lst"
	#Write-Host ("$fn_name() EOI_mrks_lst type = " + $EOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() EOI_mrks_lst cnt = " + $EOI_mrks_lst.Count)
	
	if ($SOI_mrks_cnt -gt $EOI_mrks_cnt) {

		for ($i=0; $i -lt $EOI_mrks_cnt; $i++) {
			$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- end index = [$i] " + $EndOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"

			if ($indices_delta -lt 0) {
				
				$SOI_mrks_lst_new = $SOI_mrks_lst
				$SOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- SOI_mrks_lst_new = $SOI_mrks_lst_new"

				# Recursive call
				$arr1 = [Array]$SOI_mrks_lst_new
				$arr2 = [Array]$EOI_mrks_lst
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2

				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}					
	}
	
	if ($EOI_mrks_cnt -ge $SOI_mrks_cnt) {

		for ($i=0; $i -lt $SOI_mrks_cnt; $i++) {
			$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- beg index = [$i] " + $BegOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"
			
			if ($indices_delta -lt 0) {
				
				$EOI_mrks_lst_new = $EOI_mrks_lst
				$EOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- EOI_mrks_lst_new = $EOI_mrks_lst_new"

				# Recursive call
				$arr1 = [Array]$SOI_mrks_lst
				$arr2 = [Array]$EOI_mrks_lst_new
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2
					
				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}
	}

	return @($rtn_flg, [Array]$SOI_mrks_lst, [Array]$EOI_mrks_lst)

}

function ExtractDatImages() {
# PURPOSE:  Extract all JPG files from a Blue Iris *.DAT file, and save them to disk
#           with filenames like 'img01.jpg', 'img02.jpg', ...
# ARGS:     see param lines below
# RETURNS:  1. (hashtable) @{ source=filespec ; data=@(jpgfilelist) }
#           2. or (string) if warning/error (NOTE: string must begin with 'WARNING:' or 'ERR/')
# NOTES:    1. TBD
#
# ATTRIBUTIONS:
#    Per Ken @blueirissupport... "The DAT file is custom serialized format. 
#    To extract JPGs: search the byte content for blocks that start with 'FF D8 FF E0' and go through 'FF D9'. 
#    To extract the JSON string: after the bitmap there are 2 QWORDs (8 bytes each), then the STRING for the JSON 
#    (comprised of a DWORD length then the characters)."
#
#    Per the JPEG Interchange Format... any JPEG file or stream is written according to the JPEG Interchange 
#    Format must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) and end with the 
#    EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#    These two markers will not be found anywhere else in a JPEG file/stream.
#    REF: https://stackoverflow.com/a/46809226
#
# KEY POINT: 
#    JPG images in a BI *.DAT file are stored in reverse chronological order
	
    param(
		[String]$AlertFile,          # Alert file, like 'GT.20230610_070000.41437.17-0.jpg' (via JSON 'clipstats')
		[String]$AlertPath,          # folder path for Blue Iris 'DAT' files (full path) (via user-setting $bi_alerts_folder_path)
		[Boolean]$LatestFlg,         # use DAT file from the most recent alert (that has a DAT file)
		[String]$DestPath,           # folder for the extracted JPG files
		[Int]$DatImageNo,            # specific images to extract	
		[Int]$DatReqdImages,         # number of images that must be found before extracting any images	
		[Int]$ImagesCnt,             # number of images to be extracted from BVR file, if the DAT file determined to be 'bad'	
		[Boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
		[Boolean]$NukeDatFileFlg,    # true removes DAT file from Blue Iris storage (typ. if used solely to extract JPGs)
		[Boolean]$NoArchiveDatFlg,   # true suppresses the archiving of Blue Iris DAT file (if used to create GIF)
		[Boolean]$QuietFlg,          # true suppresses verbose output to the console
		[Boolean]$FromConsoleFlg,    # true notifies function that calling command was executed from the console
		[Boolean]$DevModeFlg
)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() AlertFile = '$AlertFile'"
	#Write-Host "$fn_name() AlertPath = '$AlertPath'"
	#Write-Host "$fn_name() LatestFlg = $LatestFlg"
	#Write-Host "$fn_name() DatAgeThresholdSec = $DatAgeThresholdSec"
	#Write-Host "$fn_name() DestPath = '$DestPath'"
	#Write-Host "$fn_name() DatImageNo = $DatImageNo"
	#Write-Host ("$fn_name() DatReqdImages type = " + $DatReqdImages.GetType())
	#Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
	#Write-Host "$fn_name() NukeDatFileFlg = $NukeDatFileFlg"
	#Write-Host "$fn_name() NoArchiveDatFlg = $NoArchiveDatFlg"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	
	# VALIDATE ARGS
	$DestPath = $DestPath.TrimEnd('\')
	
	# Special handling for this script
	if ($DevModeFlg) {
		$FromConsoleFlg = $true
		$QuietFlg = $false
	} else {
		$FromConsoleFlg = $false
		$QuietFlg = $false
	}
	
	# INITIALIZE VARS
	$bkup_path = "$DestPath\_dat_archive"
	$dat_file_found_flg = $false  # Default
	$rtn_obj = [ordered]@{}
	# ... extract camname
	$str = Split-Path $AlertFile -leaf
	$arr = $str.Split(".")
	$cam = $arr[0]	
	
	# For debugging
	#Write-Host "$fn_name() str = '$str'"
	#Write-Host ("$fn_name() arr = " + ($arr | ConvertTo-Json))
	#Write-Host "$fn_name() cam = '$cam'"
	
	# DELETE ANY EXISTING JPG FILES FROM DESTINATION FOLDER
	$jpg_path =  $DestPath
	if (!(Test-Path -LiteralPath "$jpg_path")) {
		New-Item -ItemType Directory -Force -Path $jpg_path
	}
	
	# For debugging
	#Write-Host "$fn_name() jpg_path = '$jpg_path'"

	# DELETE EXISTING *.JPG FILES ASYNCHRONOUSLY
	Get-Childitem -Path ($jpg_path + "\img*.jpg") | Remove-Item -Force
	
	# USE THE 'DAT' FILE IMPLIED BY THE DATABASE (VIA JSON CLIPSTATS COMMAND)

	$dat_file = "$AlertPath\$AlertFile".Replace('.jpg','.dat')
	if (Test-Path -LiteralPath "$dat_file") {
		$dat_file_found_flg = $true
	}

	# PROCEED IF 'DAT' FILE FOUND
	if ($dat_file_found_flg) {

		# For debugging
		<# if ($FromConsoleFlg -and !$QuietFlg) {
			Write-Host ""
			Write-Host "$fn_name() Alert AI 'DAT' File Processing Summary:"
			Write-Host "---------------------------------------"
			Write-Host "$fn_name() Found DAT file: '$dat_file'"
		} #>
	
		# LOAD THE 'DAT' FILE
		if ($PsVersion -ge 7) {
			#$bytes = [System.IO.File]::ReadAllBytes("$dat_file")
			# Ref: https://stackoverflow.com/a/63782608/13499984
			$bytes = Get-Content -LiteralPath "$dat_file" -AsByteStream -Raw
		} else {
			$bytes = Get-Content -LiteralPath "$dat_file" -Encoding Byte -Raw
		}

		# For debugging
		#Write-Host ("$fn_name() bytes len = " + $bytes.Count)
		#Write-Host $bytes
		#$msg = $bytes.Count
		#msg.exe $script:server_username /TIME:60 "$fn_name() bytes len = $msg"
		
		# JPG BYTE MARKER NEEDLES
		# ... JPGs start
		$hexArray = "FF", "D8", "FF", "E0"
		$search_ary1 = @()
		foreach ($hex in $hexArray) {
			$search_ary1 += [Int32]"0x$hex"
		}
		# ... JPGs end
		$hexArray = "FF", "D9"
		$search_ary2 = @()
		foreach ($hex in $hexArray) {
			$search_ary2 += [Int32]"0x$hex"
		}

		# For debugging.
		#Write-Host "$fn_name() jpg data starting byte pattern = $search_ary1"
		#Write-Host "$fn_name() jpg data ending byte byte pattern = $search_ary2"
		
		$found_ary1 = FindBytes -Bytes $bytes -Search $search_ary1 -All
		$found_ary2 = FindBytes -Bytes $bytes -Search $search_ary2 -All
		if ($found_ary1) {
			$found_ary1_cnt = $found_ary1.Count
		} else {
			$found_ary1 = @()
			$found_ary1_cnt = 0
		}
		if ($found_ary2) {
			$found_ary2_cnt = $found_ary2.Count
		} else {
			$found_ary2 = @()
			$found_ary2_cnt = 0
		}

		# For debugging.
		#Write-Host "$fn_name() jpg data start indices = $found_ary1"
		#Write-Host "$fn_name() jpg data end indices = $found_ary2"
		#msg.exe $script:server_username /TIME:60 "$fn_name()... `nfound_ary1_cnt = $found_ary1_cnt `nfound_ary2_cnt = $found_ary2_cnt"
		#Write-Host "$fn_name() found_ary1_cnt = $found_ary1_cnt"
		
		# Validate expected DAT images count
		$bad_DAT_file_flg = $false
		if ($found_ary1_cnt -lt $DatReqdImages) {
			$bad_DAT_file_flg = $true
			if ($found_ary1_cnt -eq 0) {
				$bad_DAT_msg = "DAT file has no images"
			} else {
				$bad_DAT_msg = "DAT file has fewer images than required ($found_ary1_cnt/$DatReqdImages)"
			}
			$bad_DAT_msg += "`nTrying to extract images from BVR clip instead..."
		}
		# Validate expected DAT structure
		# ... expect $found_ary1 like '25 113202'
		# ...        $found_ary2 like '113054 227880'
		if (!$bad_DAT_file_flg) {
			$i = 0
			foreach ($jpg_idx in $found_ary1) {
				if ($i -gt 0) {
					if ([Int]$jpg_idx -ge [Int]$found_ary2[$i - 1]) {
						$bad_DAT_file_flg = $true
						$bad_DAT_msg = "DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
						break
					}
				}
			}
		}
		
		# For debugging
		#Write-Host "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"
		#msg.exe $script:server_username /TIME:60 "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"

		$extract_success_flg = $false  # Assume failure

		if (!$bad_DAT_file_flg) {

			# DAT file has at least one JPG file
			if ($found_ary1_cnt -gt 0) {

				# DAT file structure is NOT as expected
				if ($found_ary1_cnt -ne $found_ary2_cnt) {
				
					# Try to repair DAT file marker arrays before giving up
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $found_ary1 `
						-EndOfImageMarks $found_ary2
						
					# For debugging
					#Write-Host "$fn_name() result_flg = $result_flg"

					if ($result_flg) {
						$found_ary1 = $SOI_mrk_idxs_ary
						$found_ary2 = $EOI_mrk_idxs_ary
						$found_ary1_cnt = $SOI_mrk_idxs_ary.Count
						$found_ary2_cnt = $EOI_mrk_idxs_ary.Count

						if ($FromConsoleFlg) {
							if ($found_ary1_cnt -gt $found_ary2_cnt) {
								Write-Warning "$fn_name() REPAIRED 'beg marker' indices: now = $jpg_SOI_mrk_idxs"
							} else {
								Write-Warning "$fn_name() REPAIRED 'end marker' indices; now = $jpg_EOI_mrk_idxs"
							}
						}
					}
				}

				# DAT file structure is as expected
				if ($found_ary1_cnt -eq $found_ary2_cnt) {

					if ($found_ary1_cnt -ge $DatReqdImages) {

						if ($ImagesCntUsedFlg) {
							if ($FromConsoleFlg) {
								Write-Warning "$fn_name() Argument '-ImagesCnt' ignored; it applies only to GIFs created from BVR files"
							}
						}

						$rtn_obj['source'] = $dat_file
						$jpgs = @()
						$jpgs_cnt = 0
						$i = 0
						foreach ($item in $found_ary1) {
							
							$bounding_idx1 = $item
							$bounding_idx2 = $found_ary2[$i] + 1  # Include the EOI marker!
							$img_bytes = $bytes[$bounding_idx1..$bounding_idx2]
							
							if ($img_bytes.Count -gt 0) {	
							
								# For debugging
								#Write-Host ("$fn_name() processing DAT img #" + ($found_ary1_cnt - $i))

								# Build JPG's filespec
								$jpg_filename_idx = $found_ary1_cnt - $i
								$jpg_filename_sfx = ([String]$jpg_filename_idx).PadLeft(2,'0')
								$jpg_filespec = "$jpg_path\img$jpg_filename_sfx.jpg"
							
								# Save the JPG file to disk					
								if ($PsVersion -ge 6) {
									# REF: https://stackoverflow.com/a/63782608/13499984
									Set-Content "$jpg_filespec" -Value $img_bytes -AsByteStream 
								} else {
									Set-Content "$jpg_filespec" -Value $img_bytes -Encoding Byte
								}

								# Confirm that the extracted data is a valid JPG file
								$jpg_valid_flg = `
									IsJpegFileValid `
									-FilePath $jpg_filespec
								
								if (!$jpg_valid_flg) {
									if ($FromConsoleFlg) {
										Write-Warning "$fn_name() DAT file export aborted because '$jpg_filespec' is NOT a valid JPG"
									}
									break
								} else {
									$jpgs_cnt++
									$jpgs += $jpg_filespec
									
									# Provide feedback
									if ($DevModeFlg) {
										if ($FromConsoleFlg) {
											#Write-Host "(dev mode) $fn_name() JPG validated ($jpg_filespec)" -ForegroundColor Cyan
										}
									}
								
									# For debugging
									#Write-Host "$fn_name() jpgs = $jpgs"
									#Write-Host "$fn_name() '$jpg_filespec' is a valid JPG"
								}

								if ($DatImageNo) {
									if ($jpg_filename_idx -eq $DatImageNo) {
							
										# For debugging
										#Write-Host "$fn_name() found target DAT img #$DatImageNo"
										
										break
									}
								}
								
							}
							$i++
						}

						$rtn_obj['data'] = $jpgs
						
						if ($jpgs.Count -gt 0) {
							$extract_success_flg = $true
						}

						# For debugging
						#Write-Host "$fn_name() Saved $jpgs_cnt JPG files to folder '$jpg_path'"
						
					} else {
						$rtn_obj = "FAIL: DAT file contains only $found_ary1_cnt/$DatReqdImages required JPGs"							
					}			

				} else {
					$rtn_obj = "FAIL: DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
				}
			} else {
				$rtn_obj = "FAIL: No JPGs found in DAT file '$dat_file'"			
			}		
		} else {
			$rtn_obj = "FAIL: " + $bad_DAT_msg
		}
	} else {
		$rtn_obj = "FAIL: Unable to find DAT file '$dat_file'"
	}

	return $rtn_obj
}

function FindDatConfimedImageMemo() {
# PURPOSE:     To find the 'confirmed alert' memo string in a Blue Iris DAT file
# PARAMS:      -Bytes = [array] bytes between images in the Blue Iris DAT file
# RETURNS:     [string] memo string  
# ATTRIBUTION: Thanks to EvanVanVan@ipcamtalk for identifying the confirmed JPG's memo in the DAT file

	param (
		[byte[]]$Bytes  # decimals array
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "entered $fn_name()..."
	#Write-Host ("$fn_name() Bytes = " + ($Bytes | ConvertTo-Json))
	#Write-Host ("$fn_name() Bytes cnt = " + $Bytes.Length)

	# Start performance timer
	$timer = [Diagnostics.Stopwatch]::StartNew()

	# For debugging
	$bytes_as_str = [System.Text.Encoding]::UTF8.GetString($bytes_after)
	#Write-Host ("$fn_name() Bytes as string = `n" + $bytes_as_str)
	#Write-Host ("$fn_name() Bytes as string = `n" +  ($bytes_as_str -creplace '\P{IsBasicLatin}'))
	
	$rtn_str = $null
	
	# Initialize bytes ary indices for the embedded memo
	$SOM_idx = 0
	$EOM_idx = 0
	
	# Initialize byte array target values
	$hex = "25"  # percent character (%)
	$target1 = [Int32]"0x$hex"	
	$hex = "00"  
	$target2 = [Int32]"0x$hex"
	$hex = "3A"  # colon character (:)
	$target3 = [Int32]"0x$hex"
	
	# For debugging
	#Write-Host "$fn_name() target1 = $target1"
	#Write-Host "$fn_name() target2 = $target2"
	#Write-Host "$fn_name() target3 = $target3"

	# Find the memo string's end index
	# ... by walking backwards to find the first percent (%) character
	for ($idx = $Bytes.Length - 1; $idx -ge 0; $idx--) {
		if ($Bytes[$idx] -eq $target1){
			$EOM_idx = $idx + 1
			break
		}
	}
	
	# Find the memo string's start index
	# ... by walking backward from the $EOM_idx to find first the '00 00 00' byte triplet
	for ($idx = $EOM_idx - 1; $idx -gt 2; $idx--) {
		if  ($Bytes[$idx-3] -eq $target2 -and $Bytes[$idx-2] -eq $target2 -and $Bytes[$idx-1] -eq $target2) {
			$SOM_idx = $idx
			break
		}
	}

	# Create a candidate memo string
	if (($EOM_idx-$SOM_idx) -gt 1) {
		$bytes_after = $Bytes[$SOM_idx..$EOM_idx]
		$bytes_after_as_str = [System.Text.Encoding]::UTF8.GetString($bytes_after)

		# At this stage, a memo string like "vehicle:93%" is actually the following string: "v\u0000e\u0000h\u0000i\u0000c\u0000l\u0000e\u0000:\u00009\u00003\u0000%\u0000"
		# So we must remove all '\0000' null characters from the extracted string
		# REF: https://stackoverflow.com/a/9866769
		$bytes_after_as_str = $bytes_after_as_str.Replace("`0", "")
		$bytes_after_as_str = $bytes_after_as_str.Trim()  # Just in case
						
		# Validate the candidate memo string 
		# ... must contain a colon that is not at either end of the string
		if ($bytes_after -contains $target3) {
			if (!$bytes_after_as_str.StartsWith($target3) -and !$bytes_after_as_str.EndsWith($target3)) {
				$rtn_str = $bytes_after_as_str
			}
		}
	}

	# Stop overall performance timer
	$timer.stop()
	$total_ms  = [math]::Round($timer.elapsed.totalmilliseconds,1).ToString("0.0")
	if ($script:fromConsole -and $script:DevMode) {
		#Write-Host "(dev mode) DAT confirmed image memo extract time (msec) = $total_ms" -ForegroundColor cyan
	}
	
	# For debugging
	#Write-Host "$fn_name() EOM_idx = $EOM_idx"	
	#Write-Host "$fn_name() SOM_idx = $SOM_idx"
	#Write-Host "$fn_name() rtn_str = $rtn_str"
	
	return $rtn_str
}

function ExtractDatConfImgInfo() {
# PURPOSE:  To find the confirmed image number and comment in a DAT (if it exists)
# PARAMS:   see param{} line below
# RETURNS:  [array] @((int)confirmed_image_num, (string)confirmed_image_memo_str, (boolean)bad_DAT_file_flg)
# REQUIRES: Functions: FindDatConfimedImageMemo()

	param (
		[String]$DatPath,
		[Boolean]$FromConsoleFlg,
		[Boolean]$QuietFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "entered $fn_name()..."
	#Write-Host "$fn_name() DatPath = $DatPath"

	# Start performance timer
	$timer = [Diagnostics.Stopwatch]::StartNew()

	$rtn_ary = @(0,'',$false)

	$dat_file_found_flg = $false  # Default
	
	if (Test-Path -LiteralPath "$DatPath") {
	
		# LOAD THE 'DAT' FILE
		if ($PsVersion -ge 7) {
			#$bytes = [System.IO.File]::ReadAllBytes("$dat_file")
			# Ref: https://stackoverflow.com/a/63782608/13499984
			$bytes = Get-Content -LiteralPath "$DatPath" -AsByteStream -Raw
		} else {
			$bytes = Get-Content -LiteralPath "$DatPath" -Encoding Byte -Raw
		}
		
		$fname = Split-Path -leaf "$DatPath"

		# For debugging
		#Write-Host ("$fn_name() DAT filename = $fname")
		#Write-Host ("$fn_name() bytes len = " + $bytes.Count)
		#Write-Host $bytes
		
		# JPG BYTE MARKER NEEDLES
		# ... JPGs start
		$hexArray = "FF", "D8", "FF", "E0"
		$search_ary1 = @()
		foreach ($hex in $hexArray) {
			$search_ary1 += [Int32]"0x$hex"
		}
		# ... JPGs end
		$hexArray = "FF", "D9"
		$search_ary2 = @()
		foreach ($hex in $hexArray) {
			$search_ary2 += [Int32]"0x$hex"
		}

		# For debugging.
		#Write-Host "$fn_name() jpg data starting byte pattern = $search_ary1"
		#Write-Host "$fn_name() jpg data ending byte byte pattern = $search_ary2"
			
		# Find image start/end indices for all images in the DAT files
		# ... expect $jpg_SOI_mrk_idxs like '25 113202'
		# ...        $jpg_EOI_mrk_idxs like '113054 227880'
		$jpg_SOI_mrk_idxs = `
			FindBytes `
			-Bytes $bytes `
			-Search $search_ary1 `
			-Start 0 `
			-All
			
		$jpg_SOI_mrk_idxs_cnt = $jpg_SOI_mrk_idxs.Count

		$jpg_EOI_mrk_idxs = `
			FindBytes `
			-Bytes $bytes `
			-Search $search_ary2 `
			-Start 0 `
			-All

		$jpg_EOI_mrk_idxs_cnt = $jpg_EOI_mrk_idxs.Count

		# For debugging.
		#Write-Host "$fn_name() jpg data start indices (SOI) = $jpg_SOI_mrk_idxs"
		#Write-Host "$fn_name() jpg data end indices (EOI) = $jpg_EOI_mrk_idxs"
		
		# VALIDATE EXPECTED STRUCTURE DEFINITION FOR DAT FILE
		$bad_DAT_file_flg = $false
		# ... 1. Test zero JPGs
		if ($jpg_SOI_mrk_idxs_cnt -eq 0) {
			$bad_DAT_file_flg = $true
		}
		# ... 2. Test for non-strictly increasing markers
		if (!$bad_DAT_file_flg) {  # Not bad yet!
			$i = 0
			foreach ($jpg_SOI_mrk_idx in $jpg_SOI_mrk_idxs) {
				if ($i -gt 0) {
					# Starting with the 2nd found image, we expect its start marker index to be 
					# greater than the previous found image's end marker index
					# ... for images:                      img1,   img2,   img3, ...
					# ... expect $jpg_SOI_mrk_idxs like '    25, 113202, 228028, ...'
					# ...        $jpg_EOI_mrk_idxs like '113054, 227880, ...'
					if ($jpg_SOI_mrk_idx -le $jpg_EOI_mrk_idxs[$i - 1]) {
						$bad_DAT_file_flg = $true
						break
					}
				}
			}
		}
		# ... 3. Test for unexpected structure
		# ...    we expect both sets of markers to have the same number of elements
		if (!$bad_DAT_file_flg) {  # Not bad yet!
			if ($jpg_SOI_mrk_idxs_cnt -ne $jpg_EOI_mrk_idxs_cnt) {
				if ($FromConsoleFlg) {
					$verbose_flg = $true
					if ($QuietFlg) {
						Write-Host ""  # Prettier output
					}
					Write-Warning "'$fname' has an unexpected structure [unbalanced JPG beg/end markers]:"
					Write-Warning "- $jpg_SOI_mrk_idxs_cnt 'beg marker' indices = $jpg_SOI_mrk_idxs"
					Write-Warning "- $jpg_EOI_mrk_idxs_cnt 'end marker' indices = $jpg_EOI_mrk_idxs"
				}
		
				# Try to repair DAT file marker arrays before giving up
				if ($jpg_EOI_mrk_idxs_cnt -ne $jpg_SOI_mrk_idxs_cnt) {
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $jpg_SOI_mrk_idxs `
						-EndOfImageMarks $jpg_EOI_mrk_idxs
						
					# For debugging
					#Write-Host "result_flg = $result_flg"

					if ($result_flg) {
						$jpg_SOI_mrk_idxs = $SOI_mrk_idxs_ary
						$jpg_EOI_mrk_idxs = $EOI_mrk_idxs_ary
						$jpg_SOI_mrk_idxs_cnt = $SOI_mrk_idxs_ary.Count
						$jpg_EOI_mrk_idxs_cnt = $EOI_mrk_idxs_ary.Count

						if ($FromConsoleFlg) {
							if ($jpg_SOI_mrk_idxs_cnt -gt $jpg_EOI_mrk_idxs_cnt) {
								Write-Host "REPAIRED 'beg marker' indices: now = $jpg_SOI_mrk_idxs"
							} else {
								Write-Host "REPAIRED 'end marker' indices; now = $jpg_EOI_mrk_idxs"
							}
						}
					}
				}
			}
		} else {
			$bad_dat_files += $DatPath
			if ($FromConsoleFlg) {
				$verbose_flg = $true
				if ($QuietFlg) {
					Write-Host ""  # Prettier output
				}
				Write-Warning "'$fname' has an unexpected structure:"
				Write-Warning ("counts of jpg marker start/end indices = " +  $jpg_SOI_mrk_idxs_cnt + "/" +  $jpg_EOI_mrk_idxs_cnt)
				if ($jpg_SOI_mrk_idxs_cnt -gt 0) {
					Write-Warning "- $jpg_SOI_mrk_idxs_cnt 'beg marker' indices = $jpg_SOI_mrk_idxs"
				}
				if ($jpg_EOI_mrk_idxs_cnt -gt 0) {
					Write-Warning "- $jpg_EOI_mrk_idxs_cnt 'end marker' indices = $jpg_EOI_mrk_idxs"
				}
			}
		}
		
		if (!$bad_DAT_file_flg) {

			# DAT FILE STRUCTURE IS GOOD (LOOKS AS EXPECTED)
			$dat_images_cnt = $jpg_SOI_mrk_idxs.Count
			$i = 0
			foreach ($jpg_SOI_mrk_idx in $jpg_SOI_mrk_idxs) {
	
				$confirmed_image_num = $dat_images_cnt - $i  # Reminder: JPGS are stored in reverse order chronologically
				$img_no_sfx = ([string]$confirmed_image_num).PadLeft(2,'0')

				# GET BYTES AFTER JPG IMAGES
				if ($i -eq $jpg_SOI_mrk_idxs_cnt - 1) {
					$after_bounding_idx1 = $jpg_EOI_mrk_idxs[$i] + 2
					$after_bounding_idx2 = $bytes.Count - 1
				} else {
					if ($i -eq 0) {
						# Get text before the final image
						$before_bounding_idx1 = 0
						$before_bounding_idx2 = $jpg_SOI_mrk_idxs[$i] - 1
						$bytes_before = $bytes[$before_bounding_idx1..$before_bounding_idx2]
						$bytes_before_cnt = $bytes_before.Count
						$bytes_before_as_str = [System.Text.Encoding]::UTF8.GetString($bytes_before)
						$bytes_before_as_clean_str = $bytes_before_as_str -creplace '\P{IsBasicLatin}'
					} 
					$after_bounding_idx1 = $jpg_EOI_mrk_idxs[$i] + 2
					$after_bounding_idx2 = $jpg_SOI_mrk_idxs[$i+1] - 1
				}
				$bytes_after = $bytes[$after_bounding_idx1..$after_bounding_idx2]
				$bytes_after_cnt = $bytes_after.Count
				$bytes_after_as_str = [System.Text.Encoding]::UTF8.GetString($bytes_after)

				# For debugging
				if ($i -eq 0) {
					<#
					Write-Host "$img_no_sfx : before_bounding_idx1 = $before_bounding_idx1, before_bounding_idx2 = $before_bounding_idx2"
					Write-Host ("$img_no_sfx : bytes_before cnt = " + $bytes_before_cnt)
					#Write-Host ("$img_no_sfx : bytes_before = " + $bytes_before)
					Write-Host ("$img_no_sfx : bytes_before as string = " + $bytes_before_as_str)
					#Write-Host ("$img_no_sfx : bytes_before as 'clean' string = " + $bytes_before_as_clean_str) 
					#>
				}
				<# 
				Write-Host "$img_no_sfx : img_bounding_idx1 = $img_bounding_idx1, img_bounding_idx2 = $img_bounding_idx2"
				Write-Host "$img_no_sfx : after_bounding_idx1 = $after_bounding_idx1, after_bounding_idx2 = $after_bounding_idx2"
				Write-Host ("$img_no_sfx : bytes_after cnt = " + $bytes_after_cnt)
				#Write-Host ("$img_no_sfx : bytes_after[0] = " + $bytes_after[0])
				#Write-Host ("$img_no_sfx : bytes_after[" + ($bytes_after_cnt - 1) + "] = " + $bytes_after[$bytes_after_cnt - 1])
				#Write-Host ("$img_no_sfx : bytes_after = " + $bytes_after)
				#Write-Host ("`n$img_no_sfx : bytes_after as string = `n" + $bytes_after_as_str)
				#>
				
				$confirmed_image_memo_str = `
					FindDatConfimedImageMemo `
					-Bytes $bytes_after
						
				if ($confirmed_image_memo_str) {
					$rtn_ary = @([Int]$confirmed_image_num, [String]$confirmed_image_memo_str, $bad_DAT_file_flg)
					break
				}
				
				$i++
			}
		} else {
			$rtn_ary = @(0,'',$bad_DAT_file_flg)
		}
	}	

	# Stop overall performance timer
	$timer.stop()
	$total_ms  = [math]::Round($timer.elapsed.totalmilliseconds,1).ToString("0.0")
	if ($script:fromConsole -and $script:DevMode) {
		#Write-Host "(dev mode) DAT confirmed alert image memo extract time (msec) = $total_ms" -ForegroundColor cyan
	}
	
	return $rtn_ary
}

function SleepWithStatusBar() {
# Source: https://gist.github.com/ctigeek/bd637eeaeeb71c5b17f4

	param (
		[double]$Seconds,
		[String]$Activity,
		[String]$Message
	)
	
    $doneDT = (Get-Date).AddSeconds($Seconds)
    while($doneDT -gt (Get-Date)) {
        $secondsLeft = $doneDT.Subtract((Get-Date)).TotalSeconds
        $percent = ($Seconds - $secondsLeft) / $Seconds * 100
 		Write-Progress -Activity $Activity -Status $Message -SecondsRemaining $secondsLeft -PercentComplete $percent
        [System.Threading.Thread]::Sleep(500)
    }
    Write-Progress -Activity $Activity -Status $Message -SecondsRemaining 0 -Completed
}

function ExtractClipImage() {
# PURPOSE:  Extract single JPG file from a Blue Iris *.BVR file, and save it to disk w/ filename like 'img01.jpg'
# ARGS:     see param lines below
# RETURNS:  (string 'success' or 'fail'
# REQUIRES: function: SleepWithStatusBar()
# APPROACH: Asynchronously execute Blue Iris HTTP Interface '/file/clips/' commands
	
    param(
		[String]$BvrFile,          # Alert clipstats 'path' - a BVR file, like 'GT.20230610_070000.bvr'
		[Int]$Offset,              # Alert clipstats 'offset' - in (msec)
		[Int]$JpgImagesCnt,        # maximum number of images that may be extracted from the BVR file over duration of Alert
		[Int]$JpgImageNo,          # image # to be usedto be extracted from the BVR file (between 1 and 10)
		[Int]$DelayMs,             # delay time to buffer BVR (typically the Alert duration in msec)
		[Int]$DelayPct,            # percentage of $DelayMs to wait before extracting JPGs from clip
		[Int]$Height,              # height of JPGs extracted from clip
		[String]$DestPath,         # folder for the extracted JPG files
		[Boolean]$QuietFlg,        # true suppresses verbose output to the console
		[Boolean]$FromConsoleFlg,  # true notifies function that calling command was executed from the console
		[Boolean]$DevModeFlg
)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() BvrFile = '$BvrFile'"
	#Write-Host "$fn_name() JpgImagesCnt = $JpgImagesCnt"
	#Write-Host "$fn_name() JpgImageNo = $JpgImageNo"
	#Write-Host "$fn_name() DelayMs = $DelayMs"
	#Write-Host "$fn_name() DelayPct = $DelayPct"
	#Write-Host "$fn_name() Height = $Height"
	#Write-Host "$fn_name() DestPath = '$DestPath'"
	
	# VALIDATE ARGS
	$DestPath = $DestPath.TrimEnd('\')
	
	# Special handling for this script
	if ($DevModeFlg) {
		$FromConsoleFlg = $true
		$QuietFlg = $false
	} else {
		$FromConsoleFlg = $false
		$QuietFlg = $false
	}
	
	# DELETE ANY EXISTING JPG FILES FROM DESTINATION FOLDER
	$jpg_path =  $DestPath
	if (!(Test-Path -LiteralPath "$jpg_path")) {
		New-Item -ItemType Directory -Force -Path $jpg_path
	}
	
	# For debugging
	#Write-Host "$fn_name() jpg_path = '$jpg_path'"

	# Delete Existing *.Jpg Files Asynchronously
	Get-Childitem -Path ($jpg_path + "\img*.jpg") | Remove-Item -Force
		
	# Calculate time delta between each extracted BVR image
	$step_ms = [math]::floor($DelayMs / $JpgImagesCnt)
	
	# Build arrays FOR A RUNSPACE that will asynchronously execute many URLs 
	$urls = @()   # Blue Iris HTTP Interface commands - to extract single JPGs from an alert's BVR
	$files = @()  # JPG filenames, serialized like 'img01.jpg' to 'img10.jpg' 
	for ($j=0; $j -lt $JpgImagesCnt; $j++) {
		
		$time = $Offset + ($j * $step_ms)

		if ($j -eq $JpgImageNo-1) {
			# Build the URL			
			if ($Height -gt 0) {
				$uri = $bi_webserver + "/file/clips/" + $BvrFile + "?time=" + $time + "&cache=1&h=" + $Height
			} else {
				$uri = $bi_webserver + "/file/clips/" + $BvrFile + "?time=" + $time + "&cache=1"
			}

			# Build the full filespec for the URL's output
			$jpg_filename_sfx = ([String]($j + 1)).PadLeft(2,'0')
			$filespec = $DestPath + "\img" + $jpg_filename_sfx + ".jpg"
			
			# For debugging
			#Write-Host "$fn_name() filespec = $filespec" 

			$urls  += $uri
			$files += $filespec
		}
	}

	# For debugging
	#Write-Host ("urls cnt = " + $urls.Count)
	#Write-Host ("files = `n" + ($files | Out-String))
	
	# KEY POINT; A 'delay' is necessary so that HTTP '/clips' commands successfully retrieve current images
	# My testing suggests that this delay must be at least as long as the duration of the alert clip we are using.
	# REF: In response to an email inquiry regarding this observation, Blue Iris support (Ken) replied
	#     "If recording continuous or with the "continuous +" modes the recording is delayed by the duration of the pre trigger buffer."
	#$delay_ms = $DelayMs*($DelayPct/100)
	$delay_ms = $DelayMs/$JpgImagesCnt*($DelayPct/100)
	#Write-Host "delay_ms = $delay_ms"
	
	# For testing
	#$delay_ms = 0
	
	if ($delay_ms -gt 0) {
		if ($fromConsole) {
			$delay_sec = $delay_ms / 1000
			$activity_str = "Retrieving JPGs..."
			$sleep_msg = "waiting for clip buffering ($delay_sec seconds)..."
			SleepWithStatusBar -Seconds ($delay_ms/1000) -Activity $activity_str -Message $sleep_msg
		} else {
			Start-Sleep -Milliseconds $delay_ms
		}
	}

	# CALL LIST OF URLS ASYNCHRONOUSLY
	# REF: https://stackoverflow.com/a/31525044/13499984

	$maxConcurrentJobs = $JpgImagesCnt
	$content = $urls

	# Create a RUNSPACE POOL where $maxConcurrentJobs is the 
	# maximum number of runspaces allowed to run concurrently    
	$RunspacePool = [runspacefactory]::CreateRunspacePool(1,$maxConcurrentJobs)

	# Open the runspace pool (very important)
	$RunspacePool.Open()

	# Provide feedback for a noticeably slow process
	if ($FromConsoleFlg -and !$QuietFlg) {
		Write-Host "Extracting JPGs from video clip file..."
	}

	# Initialize collectors
	$ps_instances = @()  
	$jobs = New-Object System.Collections.ArrayList
	
	$j = 0
	foreach ($url in $content) {
		
		$file = $files[$j]

		# For debugging
		#Write-Host "$fn_name() file = $file"
		
		# Create a new PowerShell instance
		# REF: https://devblogs.microsoft.com/scripting/beginning-use-of-powershell-runspaces
		$ps_instance = [powershell]::Create()
		
		# Add the instance to a collector (to facilitate disposal later)
		$ps_instances += $ps_instance

		# Tell the instance to execute in our runspace pool
		$ps_instance.RunspacePool = $RunspacePool

		# Attach some code to the PS instance
		# ... "Invoke-RestMethod is perfect for quick APIs that have no special response information such as Headers or Status Codes, 
		#     whereas Invoke-WebRequest gives you full access to the Response object and all the details it provides."
		#$ps_instance.AddCommand("Invoke-WebRequest").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
		# NOTE: w/o Out-Null, the next line injects "Invoke-RestMethod" objects to the output stream
		$ps_instance.AddCommand("Invoke-RestMethod").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
		
		# BEGIN EXECUTION ASYNCHRONOUSLY
		#$ps_instance.BeginInvoke()
		#$ps_instance.BeginInvoke() | Out-Null
		$handle = $ps_instance.BeginInvoke()

		# For debugging
		#Write-Host ("handle = " + ($handle | ConvertTo-Json))

		# Fill collector
		$job_obj = [PSCustomObject]@{
			instance = $null
			data = $null
		}
		$job_obj.instance = $ps_instance
		$job_obj.data = $handle
		# NOTE: w/o Out-Null, the next line injects an integer to the output stream
		$jobs.Add($job_obj) | Out-Null  

		# Give feedback on progress
		if ($FromConsoleFlg -and !$QuietFlg) {
			#Write-Host ("Initiated request for {0}" -f $url)
			Write-Host "Initiated request for $url"
			#Write-Host ("Initiated request for $url`n" + ($handle | ConvertTo-Json))
			#Write-Host ("handle = " + ($handle | ConvertTo-Json -Depth 16))
			#Write-Host ("Available Runspaces in RunspacePool: {0}" -f $RunspacePool.GetAvailableRunspaces())
			<# Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
				$_.data.iscompleted -ne 'Completed'
			}).Count) #>
		}
		
		$j++
	}

	# For debugging
	#Write-Host ("Jobs: " + ($jobs | ConvertTo-Json))

	# WAIT UNTIL ALL JPG FILES CREATED
	$jpg_filename_sfx = ([String]$JpgImageNo).PadLeft(2,'0')
	$filespec = $DestPath + "\img" + $jpg_filename_sfx + ".jpg"
	$iter = 0
	$max_tries = 20
	$success_flg = $true
	$sleep_ms = 25
	do { 

		# Check status of jobs in runspace pool
		<# Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
			$_.data.iscompleted -ne 'Completed'
		}).Count) #>

		if (-not (Test-Path $filespec)) {
			Start-Sleep -Milliseconds $sleep_ms
			$iter++
		}
		if ($iter -gt $max_tries) {
			$success_flg = $false
			break
		}
	} until (Test-Path $filespec)

	# Feedback async status
	if ($success_flg) {
		$rtn_str = 'success'
		#$ascync_msg = "SUCCESS: $JpgImagesCnt asynchronous web requests; Final JPG ready after $iter $sleep_ms-msec waits ($filespec)"
		$ascync_msg = "SUCCESS: 1 asynchronous web requests; Final JPG ready after $iter $sleep_ms-msec waits ($filespec)"
		if ($FromConsoleFlg) {
			if (!$QuietFlg) {
				Write-Host $ascync_msg
			} 
			if ($DevModeFlg) {
				Write-Host "(dev mode) $ascync_msg" -ForegroundColor Cyan
			}
		}
	} else {
		$rtn_str = 'fail'
		#$ascync_msg = "FAIL: $JpgImagesCnt asynchronous web requests; Final JPG NOT ready after $iter $sleep_ms-msec waits ($filespec)"
		$ascync_msg = "FAIL: 1 asynchronous web requests; Final JPG NOT ready after $iter $sleep_ms-msec waits ($filespec)"
	}			
	
	#Write-Host (“Available Runspaces in RunspacePool: {0}” -f $RunspacePool.GetAvailableRunspaces())

	# Perform a proper cleanup to avoid possible resources issues / memory leaks
	<# $return = $job_obj | ForEach {
		$_.instance.EndInvoke($_.handle)
		$_.instance.Dispose()
	}
	$job_obj.clear() #>
	foreach ($ps_instance in $ps_instances) {
		$ps_instance.Dispose()  # Release resources
	}
	$ps_instances.Clear()
	
	return $rtn_str
}

function PurgeOldFiles() {
# PURPOSE:  Delete old files (> N days) from a filepath like 'D:\myfolder\*.csv'
# ARGS:     see param lines below
# RETURNS:  (array) @(found_files_cnt, purged_files_cnt)
#           or (string) 'ERR/' message
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	
    param(
		[String]$Pattern,  # [REQUIRED] full filespec pattern for targeted files; like 'D:\myfolder\*.csv'
		[Int]$Days         # [REQUIRED] maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() Pattern = $Pattern"
	#Write-Host "$fn_name() Days = $Days"
	
	if ($null -eq $Pattern -or $null -eq $Days) {
		return "ERR/$fn_name() requires 2 arguments (-Pattern -Days)"
	}
	
	# GET FILES MATCHING PATTERN
	$cur_ary = get-childitem -Path "$Pattern" | Select Name
	if ($cur_ary) {
		$cur_ary_cnt = $cur_ary.Count
	} else {
		$cur_ary_cnt = 0
	}

	# For debugging
	#Write-Host ("$fn_name() cur_ary type = " + $cur_ary.GetType())
	#Write-Host "$fn_name() cur_ary cnt = $cur_ary_cnt"
	#Write-Host ("$fn_name() cur_ary = " + ($cur_ary | ConvertTo-Json))

	# GET FILES MATCHING AGE REQUIREMENT
	$purge_ary = @() 		
	if ($cur_ary_cnt -gt 0 -and $Days -gt 0) {
		
		$limit = (Get-Date).AddDays(-$Days)
		
		# For debugging
		#Write-Host "$fn_name() limit = $limit"

		$purge_obj = Get-ChildItem -Path "$Pattern" | 
			Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
			Select Name

		if ($purge_obj) {
			$purge_ary = $purge_obj.Name
		}
	}
	
	# Get count for return array
	$purge_ary_cnt = $purge_ary.Count
	
	# For debugging
	#Write-Host "$fn_name() purge_ary cnt = $purge_ary"
	#Write-Host ("$fn_name() purge_ary = " + ($purge_ary | ConvertTo-Json))

	# REMOVE FILES MATCHING AGE REQUIREMENT
	if ($purge_ary_cnt -gt 0) {
		if ($Days) {
			$null = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
				Remove-Item -Force
		}
	}
	
	return  @($cur_ary_cnt, $purge_ary.Count)
}

function MakeFlagsDecimalValue() {
# PURPOSE:    To covert to decimal a bit array like [1,1,1,0...] via [2^0, 2^1 ... 2^N]
# PARAMS:     -BinAry = (array) 
# RETURNS:    (integer)

	param (
		[Int[]]$BinAry
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() BinAry = $BinAry"

	$rtn_val = 0

	$i = 0
	foreach($item in $BinAry) {
		if ($item -eq 1) {
			$rtn_val += [Math]::Pow(2, $i)
		}
		$i++
	}

	# For debugging
	#Write-Host ("$fn_nam() rtn_val = " + $rtn_val)
	
	return $rtn_val;
}
  
function SetAlertFlags() {
# PURPOSE:    To set the 'Flag' flag for Alerts that lack it
# NOTE:       The alert's database entry is modified via the JSON 'update' command
# USAGE:      {this_script_name}.ps1 prop cam(s) prop(s)
# PARAMS:     -AlertDbPath    = (string) Blue Iris Alert's database record ID
#             -AlertDbMemo    = (string) Blue Iris Alert's database memo text
#             -AlertDbFlags   = (int) Blue Iris Alert's database flags
#             -QuietFlg       = (boolean) true suppresses verbose informational output to the console
#             -FromConsoleFlg = (boolean) $true notifies function that calling command was executed from the console
# RETURNS:    (string) "success' or 'fail'

	param(
		[String]$AlertDbPath,
		[String]$AlertDbMemo,
		[Int]$AlertDbFlags,
		[Boolean]$QuietFlg,
		[Boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() AlertDbPath = $AlertDbPath"
	#Write-Host "$fn_name() AlertDbMemo = $AlertDbMemo"
	#Write-Host "$fn_name() AlertDbFlags = $AlertDbFlags"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	
	$rtn_str = 'fail'  # Assume failure

	# CHANGE/SET THE ALERT'S DATABASE FLAGS ENTRY
	$jsonCmd = @{
		cmd   = 'update'
		path  = $AlertDbPath
		memo  = $AlertDbMemo
		flags = $AlertDbFlags
	}			
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("jsonCmd = " + ($jsonCmd | ConvertTo-Json))
	#Write-Host ("json_response_obj = " + ($json_response_obj | ConvertTo-Json))

	# RESTORE THE ALERT'S DATABASE MEMO 

	if ($json_response_obj -match 'success') {

		# Note: This action INDIRECTLY updates the JPG file's Exif metadata
		#  How: Blue Iris refreshes the JPG's Exif metadata when the Alert's DB fields are JSON 'updated'
		# We will verify this change next

		$jsonCmd = @{
			cmd   = 'update'
			path  = $AlertDbPath
			memo  = $AlertDbMemo
			flags = $AlertDbFlags
		}					
		# Returns a powershell PSCustomObject
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-QuietFlg $QuietFlg `
			-FromConsoleFlg $FromConsoleFlg

		# For debugging
		#Write-Host ("jsonCmd = " + ($jsonCmd | ConvertTo-Json))
		#Write-Host ("json_response_obj = " + ($json_response_obj | ConvertTo-Json))

		if ($json_response_obj -match 'success') {
			$rtn_str = 'success'
			<# if ($FromConsoleFlg -and !$QuietFlg) {
				Write-Host "$item - Alert's ['flags'] value changed to '$AlertDbFlags'"
			} #>
		} else {
			<# if ($FromConsoleFlg) {
				Write-Warning "$item - Unsuccessful attempt to update the Alert's flags"
			} #>
		}
	}

	return $rtn_str
}

function RemoveArrayItems() {
# PURPOSE:    Remove specified items from an array
# PARAMS:     -Haystack = (array) simple array of values
#             -Needles  = (array) values to be removed from Haystack
# RETURNS:    (array)
# NOTES:      REF: https://devblogs.microsoft.com/scripting/powertip-use-powershell-to-remove-an-item-from-an-array/

	param (
		[Array]$Haystack,
		[Array]$Needles
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#Write-Host "entered function $fn_name()..."
	#Write-Host ("caller: " + ('Caller {0}' -f (Get-PSCallStack)[1].Command))
	# REF: https://stackoverflow.com/a/33570449
	#Write-Host ("caller line number: " + ((Get-PSCallStack)[1].ScriptLineNumber))
	#Write-Host ("$fn_name() Haystack = " + ($Haystack | ConvertTo-Json))
	#Write-Host ("$fn_name() Needles = " + ($Needles | ConvertTo-Json))
	
	if ($null -eq $Haystack -or $null -eq $Needles -eq 0) {
		if ($null -eq $Haystack) {
#Write-Host "$fn_name() made it1A"
#Write-Host ("caller line number: " + ((Get-PSCallStack)[1].ScriptLineNumber))
		} 
		if ($null -eq $Needles) {
#Write-Host "$fn_name() made it1B"
#Write-Host ("caller line number: " + ((Get-PSCallStack)[1].ScriptLineNumber))
		}
		return @()
	}
	if ($Haystack.Count -eq 0 -or $Needles.Count -eq 0) {
		if ($Haystack.Count -eq 0) {
#Write-Host "$fn_name() made it2A"
#Write-Host ("caller line number: " + ((Get-PSCallStack)[1].ScriptLineNumber))
#Write-Host ("$fn_name() Haystack = " + ($Haystack | ConvertTo-Json))
		}
		if ($Needles.Count -eq 0) {
#Write-Host "$fn_name() made it2B"
#Write-Host ("caller line number: " + ((Get-PSCallStack)[1].ScriptLineNumber))
#Write-Host ("$fn_name() Needles = " + ($Needles | ConvertTo-Json))
		}
		return $Haystack
	}

	# Convert $Haystack array to type ArrayList... so we can use the '.RemoveAt(index)' method
	$items = New-Object System.Collections.ArrayList
	foreach ($item in $Haystack) {
		[Void]$items.Add($item)
	}

	# Remove elements from the ArrayList from the bottom-up
	$item_cnt = $Haystack.Count
	for ($idx=$item_cnt-1; $idx -ge 0; $idx--) {
		$item = $items[$idx]
		foreach ($needle in $Needles) {
			if ($item -eq $needle) {
				#Write-Host "match! Haystack item ($item) = needle"
				$items.RemoveAt($idx)
				#Write-Host ("$fn_name() targets after removing index $idx = " + ($items | ConvertTo-Json))
				break
			}
		}
	}

	$rtn_ary = [Array]$items

	# For debugging
	#Write-Host ("$fn_name() rtn_ary = " + ($rtn_ary | ConvertTo-Json))

	return $rtn_ary
}

function LogScriptResults() {
# PURPOSE:  Log script results data to CSV file for protected Alerts activity.
# ARGS:     see param lines below
# RETURNS:  (string) 'success' or 'ERR/unable to write to CSV log'
# NOTES:    1. CSV file is created automatically if it does not already exist. 
#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
	
    param(
		[String]$CsvFilespec,          # [REQUIRED]
		[String]$Timestamp,            # [REQUIRED]
		[Array]$AlLabelsAry,
		[Array]$AlIssuesAry,
		[Array]$AlFixedYnAry,
		[String]$HostStr,
		[String]$CmdArgsStr
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() CsvFilespec = $CsvFilespec"
	#Write-Host "$fn_name() Timestamp = $Timestamp"
	#Write-Host ("$fn_name() AlLabelsAry cnt = " + $AlLabelsAry.Count)
	#Write-Host ("$fn_name() AlLabelsAry = " + ($AlLabelsAry | ConvertTo-Json))
	#Write-Host ("$fn_name() AlIssuesAry cnt = " + $AlIssuesAry.Count)
	#Write-Host ("$fn_name() AlIssuesAry = " + ($AlIssuesAry | ConvertTo-Json))
	#Write-Host ("$fn_name() AlFixedYnAry cnt = " + $AlFixedYnAry.Count)
	#Write-Host ("$fn_name() AlFixedYnAry = " + ($AlFixedYnAry | ConvertTo-Json))
	#Write-Host "$fn_name() HostStr = $HostStr"
	#Write-Host "$fn_name() CmdArgsStr = $CmdArgsStr"

	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$CsvFilespec -or !$Timestamp) {
		$err_msg = "ERR/expecting 2 arguments (CsvFilespec Timestamp)"
		return $err_msg
	}
	
	# Assume success
	$rtn_str = 'success'
		
	# BUILD THE NEW ROWS
	# Append a row for each Alert that was found/fixed
	
	$i = 0
	foreach ($item in $AlLabelsAry) {
		
		# For debugging
		#Write-Host "$fn_name() saving output CSV row index #$i"
		
		try {
			# Note: this object's keys are the CSV header column's labels
			$new_row_obj = [PSCustomObject]@{
				'Timestamp'  = $Timestamp
				'AlertID'    = $AlLabelsAry[$i]
				'Issue(s)'   = $AlIssuesAry[$i]
				'Fixed?'     = $AlFixedYnAry[$i]
				'ScriptHost' = $HostStr
				'ScriptArgs' = $CmdArgsStr
			}

			$i++

			# APPEND NEW ROW TO CSV FILE, creating the file if necessary
			if ($PSVersionTable.PSVersion.Major -ge 7) {
				# REFS: https://tinyurl.com/yvy4mjy6
				#       https://stackoverflow.com/a/70717331
				#       https://stackoverflow.com/a/60681762
				try {
					$new_row_obj | Export-Csv -LiteralPath "$CsvFilespec" -Append -Force -UseQuotes AsNeeded
				} catch {
					$err_msg = "ERR/unable to append row to '$CsvFilespec'"
					return $err_msg
				}
			} else {
				try {
					$new_row_obj | Export-Csv -LiteralPath "$CsvFilespec" -Append -Force -NoTypeInformation
				} catch {
					$err_msg = "ERR/unable to append row to '$CsvFilespec'"
					return $err_msg
				}
			}
		}
		catch {
			$err_msg = "ERR/unable to append row to '$CsvFilespec'"
			return $err_msg
		}
	}
	
	return $rtn_str
}

function CopyFileAsSequence() {
# PURPOSE:  Copy a source file without over writing; append a sequence number '(N)' 
#           to the end of its basename until it is unique in the destination folder 
# PARAMS:   see param lines below
# RETURNS:  (string) new_filespec, or 'ERR/' message
# ATTRIBUTIONS:  https://stackoverflow.com/a/53595350/13499984

	param(
		[String]$SourceFilespec,  # [REQUIRED]
		[String]$TargetFilespec,  # [REQUIRED]
		[String]$DestFolder         # [REQUIRED]
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	# For debugging
	#Write-Host "Entered $fn_name()"
	#Write-Host "$fn_name() SourceFilespec = $SourceFilespec"
	#Write-Host "$fn_name() TargetFilespec = $TargetFilespec"
	#Write-Host "$fn_name() DestFolder = $DestFolder"
	
	$rtn_str = 'fail'  # Assume failure

	# Just in case
	$DestFolder = $DestFolder.TrimEnd('\')

	if (Test-Path -LiteralPath "$SourceFilespec") {

		if (Test-Path -LiteralPath "$DestFolder") {

			if (Test-Path -LiteralPath "$TargetFilespec") {
				
				$fileInfo = Get-Item -Path $TargetFilespec
				$base_name = $fileInfo.BaseName
				$extension = $fileInfo.Extension

				# Get an array of all filenames (name only) of the files with a similar name already present in the destination folder
				$allFiles = @(Get-ChildItem $DestFolder -File -Filter "$base_name*$extension" | Select-Object -ExpandProperty Name)

				# Construct the new filename
				$new_name = $base_name + $extension  # Initialize
				$count = 1
				while ($allFiles -contains $new_name) {
					# add a sequence number in brackets to the end of the basename until it is unique in the destination folder
					$new_name = "{0}({1}){2}" -f $base_name, $count++, $extension
				}

				# Construct the new filespec
				$new_target_fs = Join-Path -Path $DestFolder -ChildPath $new_name

				try {
					Copy-Item -Path $SourceFilespec -Destination $new_target_fs
					$rtn_str = $new_target_fs
				} catch {
					$rtn_str = "ERR/failed to create file '$new_target_fs'"
				}
				
			} else {
				try {
					Copy-Item -Path $SourceFilespec -Destination $TargetFilespec
					$rtn_str = $TargetFilespec
				} catch {
					$rtn_str = "ERR/failed to create file '$TargetFilespec'"
				}
			}
		} else {
			$rtn_str = "ERR/destination folder does not exist ('$DestFolder')"
		}

	} else {
		$rtn_str = "ERR/source file does not exist ('$FileName')"
	}
	
	return $rtn_str
		
}

function RegGetValue() {
# PURPOSE:  Get current value of a registry entry
# ARGS:     see param lines below
# RETURNS:  value, or error message

	param(
		[String]$KeyPath,  # like, 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips'
		[String]$KeyName   # like, 'compact'
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -KeyPath = '$KeyPath'"
	#Write-Host "$fn_name() -KeyName = '$KeyName'"

	try {
		$reg_obj = Get-ItemProperty -Path $KeyPath -Name $KeyName
		$rtn_value = $reg_obj.$KeyName 
	}
	catch {
		$err_msg = "ERR/unable to get registry key '$KeyPath\$KeyName'"
		return $err_msg
	}	
	
	# For debugging
	#$reg_obj
	#Write-Host "key '$KeyName' = $rtn_value"
	
	return $rtn_value
}

function RegSetValue() {
# PURPOSE:  Change value of a registry entry
# ARGS:     see param lines below
# RETURNS:  new value, or error message

	param(
		[String]$KeyPath,  # like, 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips'
		[String]$KeyName,  # like, 'compact'
		$NewValue  
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -KeyPath = '$KeyPath'"
	#Write-Host "$fn_name() -KeyName = '$KeyName'"
	#Write-Host "$fn_name() -NewValue = '$NewValue'"

	# Get old value
	$old_value = `
		RegGetValue `
		-KeyPath $KeyPath `
		-KeyName $KeyName
		
	if ($null -eq $old_value) {
		return $old_value
	} else {
		if ($old_value.GetType() -match 'string') {
			if ($old_value.StartsWith('ERR/')) {
				return $old_value
			}
		}
	}

	# For debugging
	#Write-Host "$fn_name() old_value = $old_value"

	try {
		# Set new value
		Set-ItemProperty -Path $KeyPath -Name $KeyName -Value $NewValue
	}
	catch {
		$err_msg = "ERR/unable to set registry key '$KeyPath\$KeyName'"
		return $err_msg
	}	

	# Verify new value
	$rtn_value = `
		RegGetValue `
		-KeyPath $KeyPath `
		-KeyName $KeyName
		
	if ($null -eq $rtn_value) {
		return $rtn_value
	} else {
		if ($rtn_value.GetType() -match 'string') {
			if ($rtn_value.StartsWith('ERR/')) {
				return $rtn_value
			}
		}
	}
	
	# For debugging
	#$reg_obj1
	#$reg_obj2
	#Write-Host ("key '$KeyName' old value = " + $old_value)
	#Write-Host ("key '$KeyName' new value = " + $rtn_value)
	
	return $rtn_value
	
}

function ToggleDbMaintenance() {
# PURPOSE:  Toggle Blue Iris scheduled setting 'Compact/repair DB at' 
#           (via Settings > Storage tab)
# ARGS:     see param lines below
# RETURNS:  new value, or error message

	param(
		[String]$TargetState  # like 'On' or 'Off'  
	)
	
	if ($TargetState -match 'On') {
		$target_value = 1
	} else {
		$target_value = 0
	}

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -TargetState = '$TargetState'"
	#Write-Host "$fn_name() target_value = $target_value"

	$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips'
	$key_name = 'compact'

	# Get current setting's value
	$key_val_old = `
		RegGetValue `
		-KeyPath $key_path `
		-KeyName $key_name
		
	if ($null -eq $key_val_old) {
		$err_msg = "ERR/unable to GET registry key value '$key_path\$key_name"
		return $err_msg
	} else {
		if ($key_val_old.GetType() -match 'string') {
			if ($key_val_old.StartsWith('ERR/')) {
				$err_msg = $key_val_old
				return $err_msg
			}
		}
	}

	# For debugging
	#Write-Host "$fn_name() key_val_old = $key_val_old"

	# Change the setting
	if ($key_val_old -ne $target_value) {

		# For debugging
		#Write-Host "$fn_name() toggling setting 'Compact/repair DB at' ..."

		$key_val_new = `
			RegSetValue `
			-KeyPath $key_path `
			-KeyName $key_name `
			-NewValue $target_value
					
		# For debugging
		#Write-Host "$fn_name() key_val_new = $key_val_new"

		if ($null -eq $key_val_new) {
			$err_msg = "ERR/unable to set registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($key_val_new.GetType() -match 'string') {
				if ($key_val_new.StartsWith('ERR/')) {
					$err_msg = $key_val_new
					return $err_msg
				}
			}
		}
	}
	
	return 'success'
	
}

# SHOW HELP
if ($Help -and $fromConsole -and !$Quiet) {
	Write-Host "======================================================================================================="
	Write-Host "BLUE IRIS ALERT PROTECTION UTILITY"
	Write-Host "Utility to ensure that your 'protected' Alerts can survive a Database/Regenerate actions;"
	Write-Host "What it does..."
	Write-Host "* makes sure that each protected Alert has a JPG file, creating one if necessary"
	Write-Host "* makes sure that each Alert's JPG & DAT files have their read-only attributes set"
	Write-Host "* makes sure that each Alert's JPG file Exif metadata matches the Alert's database memo text"
	Write-Host "-------------------------------------------------------------------------------------------------------"
	Write-Host "Usage: $ScriptName {no args} ... output a properties table for all 'protected' Alerts"
	Write-Host "       $ScriptName -NoInspect ... faster output (useful when experimenting w/ most filters)"
	Write-Host "       $ScriptName -Cameras | -Pattern | -MemoTag ... show a subset of protected Alerts"
	Write-Host "       ditto plus -Filter* ... show protected Alerts w/ specified (or any) issues"
	Write-Host "       ditto plus -Fix* | FixAll... fix protected Alerts w/ specified (or all) issues"
	Write-Host "Args: -NoInspect ... skip inspection of Alert JPG's for their Exif metadata (faster execution)"
	Write-Host "      -Cameras   ... analyze protected Alerts for specified cameras (comma-delimited)"
	Write-Host "      -Patterns  ... analyze protected Alerts for specified 'AlertID' column substrings"
	Write-Host "      -MemoTag   ... analyze protected Alerts w/ matching substring ('tag') in 'AlertMemo' column [1]"
	Write-Host "      -Flags     ... analyze Alerts with specified flags, like 'F','!C' (and 'P' w/ -AllAlerts)"
	Write-Host "      -FilterNoJPG   ... filter protected Alerts without a JPG file"
	Write-Host "      -FilterNoDAT   ... filter protected Alerts without a DAT file"
	Write-Host "      -FilterNoBoth  ... filter protected Alerts without a JPG file AND without a DAT file"
	Write-Host "      -FilterNoRO    ... filter protected Alerts without a read-only JPG OR DAT file"
	Write-Host "      -FilterNoMatch ... filter protected Alerts when JPG Exif metadata does not match the alert memo"
	Write-Host "      -FilterNoMemo  ... filter protected Alerts that have empty '' or 'NA' memos in the Blue Iris db"
	Write-Host "      -FilterZeroDates ... filter protected Alerts that have a zero-date in the Blue Iris database"
	Write-Host "      -FilterIssues ... filter protected Alerts for all -Filter* issues"
	Write-Host "      -FixJpgExif   ... force Alert's JPG file Exif metadata to match the Alert's database memo [2]"
	Write-Host "      -FixFiles     ... when neither a JPG or DAT exist, create replacement/dummy JPG and DAT files [3]"
	Write-Host "      -FixReadOnly  ... set read-only attribute for JPG & DAT files"
	Write-Host "      -FixMemos     ... fix Alert db records that have empty '' or 'NA' memos (via DAT confirmed img)"
	Write-Host "      -FixZeroDates ... fix Alert db records that have a zero-date (via DAT file date, if exists)"
	Write-Host "      -FixAll       ... performs all fixes: -FixJpgExif, -FixFiles & -FixReadOnly [3]"
	Write-Host "      -SetFlag ... set the database flag 'Flag' for all protected Alerts"
	Write-Host "      -SortAsc ... show results with Alerts descriptors in ascending order"
	Write-Host "      -SourcePath ... alternate path for JPG & DAT files (vs default in the user-settings file) [1,4]"
	Write-Host "      -ClearMarks ... remove the script's custom_memo_mark ('^', by default) from filtered Alerts"
	Write-Host "      -AllAlerts ... process Alerts regardless of 'Protected' flag"
	Write-Host "      -HideTable ... disable display of table of protected Alerts found/processed"
	Write-Host "      -Quiet     ... disable status messaging EXCEPT results table, warnings, and CLI mode output"
	Write-Host "      -NoBiLog   ... disable/override default logging to the Blue Iris logfile"
	Write-Host "      -ShowCLI ... also display the background execution return object (an array of arrays)"
	Write-Host "      -Help ... Show this usage guide"
	Write-Host "NOTES:"
	Write-Host "   [1] Values for these Args must be quoted if they contain spaces"
	Write-Host "   [2] Original JPG files are backed up to a script-managed folder (`$app_path\backup)"
	Write-Host "   [3] These Args may perform a database Repair/Regenerate (req'd per Blue Iris support)"
	Write-Host "   [4] The folder specified by -SourcePath (`$bi_alerts_folder_path) is recursively inspected"
	Write-Host "RETURNS:"
	Write-Host "   Console: Table/CSV (Alerts, Flags, HasJPG, HasDAT, RO(attribute), AlertMemo, JpgExifMemo, Issues)"
	Write-Host "   CLI mode: array of arrays (same items as output CSV file)"
	Write-Host "-------------------------------------------------------------------------------------------------------"
	exit
}

#----------------------------------------------
# HANDLE COMMANDLINE ARGS
if (1 -eq 1) {  # Use to collapse this lengthy code section while editing

	if ($NoRegen -and !$DevMode) {
		$err_msg = "ERR/argument -NoRegen must be used with -DevMode"
		return $err_msg
	}
	<# if ($AllAlerts -and !$DevMode) {
		$err_msg = "ERR/argument -AllAlerts must be used with -DevMode"
		return $err_msg
	} #>
	<# if ($Flags -and !$DevMode) {
		$err_msg = "ERR/argument -Flags must be used with -DevMode"
		return $err_msg
	} #>
	if ($AllAlerts) {
		$p_adj = ' '
	} else {
		$p_adj = ' protected '
	}
	#Write-Host "p_adj = '$p_adj'"

	$flags_arg_ary = @()
	if ($Flags) {
		$flags_arg_ary = $Flags
	}
	$flags_arg_ary_cnt = $flags_arg_ary.Count
	
	# For debuggin
	#Write-Host ("flags_arg_ary cnt = " + $flags_arg_ary_cnt)
	#Write-Host ("flags_arg_ary = " + ($flags_arg_ary | ConvertTo-Json))
	#Write-Host ("flags_arg_ary -contains 'C' = " + ($flags_arg_ary -contains 'C'))
	#Write-Host ("flags_arg_ary -contains '!C' = " + ($flags_arg_ary -contains '!C'))
	
	if ($Patterns -and !$Cameras) {
		$err_msg = "ERR/argument -Patterns must be used with -Cameras"
		return $err_msg
	}
	
	# Make string of params (used for logging to the script log & the Blue Iris logfile)
	$arr = @()
	foreach ($key in $MyInvocation.BoundParameters.keys){
		$value = (get-variable $key).Value 
		if ($value -eq $true) {
			$arr += "-$key"
		} else {
			$arr += "-$key $value"
		}
	}
	$params_str = $arr -Join " "
	#Write-Host "params_str = $params_str"

	$cams_ary = @()
	if ($Cameras) {
		$cams_ary = $Cameras -Split ','
	}
	$cams_arg = $cams_ary -Join ','

	$patterns_ary = @()
	if ($Patterns) {
		$patterns_ary = $Patterns -Split ','
	}

	if ($SourcePath) {
		$bi_alerts_folder_path = $SourcePath.TrimEnd('\')
	}
	#Write-Host "source_path = $bi_alerts_folder_path"

	$filter_args = @()
	if ($FilterIssues)    { $filter_args += "FilterIssues"}
	if ($FilterNoJPG)     { $filter_args += "FilterNoJPG"}
	if ($FilterNoDAT)     { $filter_args += "FilterNoDAT"}
	if ($FilterNoBoth)    { $filter_args += "FilterNoBoth"}
	if ($FilterNoRO)      { $filter_args += "FilterNoRO"}
	if ($FilterNoMatch)   { $filter_args += "FilterNoMatch"}
	if ($FilterNoMemo)    { $filter_args += "FilterNoMemo"}
	if ($FilterZeroDates) { $filter_args += "FilterZeroDates"}
	$filter_type_arg_used_flg = ($filter_args.count -gt 0)
	
	#Write-Host ("filter_args = " + ($filter_args | ConvertTo-Json))
	#Write-Host "filter_type_arg_used_flg = $filter_type_arg_used_flg"

	$fix_args = @()
	if ($FixJpgExif)   { $fix_args += "FixJpgExif"}
	if ($FixFiles)     { $fix_args += "FixFiles"}
	if ($FixReadOnly)  { $fix_args += "FixReadOnly"}
	if ($FixMemos)     { $fix_args += "FixMemos"}
	if ($FixZeroDates) { $fix_args += "FixZeroDates"}
	if ($FixAll)       { $fix_args += "FixAll"}
	$fix_type_arg_used_flg = ($fix_args.count -gt 0)
	
	#Write-Host ("fix_args = " + ($fix_args | ConvertTo-Json))
	#Write-Host "fix_type_arg_used_flg = $fix_type_arg_used_flg"

	if ($filter_args.Count -gt 1) {
		$err_msg = "Please use only one -Filter* type argument at a time"
		return $err_msg
	}
	#if (($FilterNoMatch -or $FilterNoMemo -or $FilterIssues) -and $NoInspect) {
	if (($FilterNoMatch -or $FilterIssues) -and $NoInspect) {
		# Filters require inspection of JPG files Exif metadata
		$err_msg = "Arguments -FilterNoMatch, -FilterNoMemo & -FilterIssues cannot be used w/ argument -NoInspect"
		return $err_msg
	}
	if (($FixJpgExif -or $FixAll -or $FilterIssues) -and $NoInspect) {
		# Filters require inspection of JPG files Exif metadata
		$err_msg = "Arguments -Fix* cannot be used w/ argument -NoInspect"
		return $err_msg
	}
	if ($FixAll -and ($FixJpgExif -or $FixFiles -or $FixReadOnly -or $FixMemos -or $FixZeroDates)) {
		$err_msg = "Argument -FixAll cannot be used w/ any other -Fix* arguments"
		return $err_msg
	}
	if ($ClearMarks -and $fix_type_arg_used_flg) {
		$err_msg = "Argument -ClearMarks cannot be used w/ a -Fix* type argument"
		return $err_msg
	}
	if ($FixZeroDates -and !(IsBiVersion '5.8.4.5')) {
		if ($fromConsole) {
			Write-Warning "Argument -FixZeroDates works best with Blue Iris 5.8.4.5 of higher"
		}
	}
	
	$filter_arg_set_flg = ( `
		$FilterNoJPG -or $FilterNoDAT -or $FilterNoBoth -or `
		$FilterNoRO -or $FilterNoMatch -or $FilterNoMemo -or `
		$FilterZeroDates -or $FilterIssues `
	)
	
	# Simulator Args
	$Sim1_arg_exists = ($cmdline_args_passed -contains 'Sim1')
	$Sim2_arg_exists = ($cmdline_args_passed -contains 'Sim2')
	$Sim3_arg_exists = ($cmdline_args_passed -contains 'Sim3')
	$Sim4_arg_exists = ($cmdline_args_passed -contains 'Sim4')
	$Sim5_arg_exists = ($cmdline_args_passed -contains 'Sim5')
	# Arg -Sim6 requires special handling because it can have a zero value
	$Sim6_arg_exists = ($cmdline_args_passed -contains 'Sim6')
	
	# For debugging
	#Write-Host "Sim6_arg_exists = $Sim6_arg_exists"
		
}
#----------------------------------------------

# ------------------------------------------------------
# MANAGE SCRIPT'S FOLDERS
if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
	# 1. Home folder
	if (!(Test-Path -LiteralPath "$app_path")) {
		New-Item -ItemType Directory -Force -Path $app_path | Out-Null
	}

	# 2. Create subfolder for the script's output CSV file, if necessary.
	$output_path = $app_path + "\_output"
	If (!(Test-Path -PathType container $output_path)) {
		New-Item -ItemType Directory -Path $output_path | Out-Null
	}
	# ... And add/refresh a README to the folder
	$readme_path = "$output_path\_README.txt"
	if (!(Test-Path -LiteralPath "$readme_path")) {
		New-Item -Path "$readme_path" | Out-Null
		$msg = "Script Output Folder" +
			"`n--------------------" +
			"`n`nWHAT? This folder holds a CSV file representation of the script's output to the PowerShell Console." + 
			"`n`nBENEFITS:" + 
			"`n* These CSV files are not constrained by the size of the Console window;" + 
			"`n* These CSV files contain additional columns that cannot be displayed in the output to the Console." +
			"`n`nTip: If you have opted out of timestamping the output.csv files (via user setting `$timestamp_output_csv_files_flg), you can retain a copy of the current 'output.csv' file and prevent its contents from being overwritten, by simply renaming it to make it unique." + 
			"`n`nNote: This folder and this README.txt file are programmatically created."
		$msg | Out-File -FilePath "$readme_path" -Append
	}

	# 3. Backup folder (for BI detabase)
	$backup_db_path = "$app_path\_backup_db_folders"
	if (!(Test-Path -LiteralPath "$backup_db_path")) {
		New-Item -ItemType Directory -Force -Path $backup_db_path | Out-Null
	}
	# ... And add/refresh a README to the folder
	$readme_path = "$backup_db_path\_README.txt"
	if (!(Test-Path -LiteralPath "$readme_path")) {
		New-Item -Path "$readme_path" | Out-Null
		$msg = "BI Database Backup Folder" +
			"`n-------------------------" +
			"`n`nWHAT? This folder holds dated backups of your Blue Iris database folder." + 
			"`n`nBENEFITS:" + 
			"`n* May be useful if something goes wrong" +  
			"`n`nNote: This folder and this README.txt file are programmatically created."
		$msg | Out-File -FilePath "$readme_path" -Append
	}

	# 4. Backup folder (for JPGs)
	$backup_jpgs_path = "$app_path\_backup_jpgs"
	if (!(Test-Path -LiteralPath "$backup_jpgs_path")) {
		New-Item -ItemType Directory -Force -Path $backup_jpgs_path | Out-Null
	}
	# ... And add/refresh a README to the folder
	$readme_path = "$backup_jpgs_path\_README.txt"
	if (!(Test-Path -LiteralPath "$readme_path")) {
		New-Item -Path "$readme_path" | Out-Null
		$msg = "BI Database JPGs Folder" +
			"`n-------------------------" +
			"`n`nWHAT? This folder holds a backup copy of JPGs before their Exif metadata is modified." + 
			"`n`nBENEFITS:" + 
			"`n* May be useful for debugging Exif metadata write operations" +  
			"`n`nNote: This folder and this README.txt file are programmatically created."
		$msg | Out-File -FilePath "$readme_path" -Append
	}
		
	# 5. Create the Script's dummy_jpg folder if necessary
	$dummyjpgs_path = "$app_path\_dummy_jpgs"
	if (!(Test-Path -LiteralPath "$dummyjpgs_path")) {
		New-Item -ItemType Directory -Force -Path $dummyjpgs_path | Out-Null
	}
	# ... and add/refresh a README to the folder
	$readme_path = "$dummyjpgs_path\_README.txt"
	if (!(Test-Path -LiteralPath "$readme_path")) {
		New-Item -Path "$readme_path" | Out-Null
		$msg = "Script 'dummy jpgs' Folder" +
			"`n----------------------------" +
			"`n`nWHAT? This folder holds JPG files extracted from Alert BVR files (clips) or DAT files" + 
			"`n`nNote: This folder and this README.txt file are programmatically created."
		$msg | Out-File -FilePath "$readme_path" -Append
	}
		
	# 6. Create the Script's DevMode folder if necessary
	$devmode_path = "$app_path\_devmode"
	if (!(Test-Path -LiteralPath "$devmode_path")) {
		New-Item -ItemType Directory -Force -Path $devmode_path | Out-Null
	}
	# ... and add/refresh a README to the folder
	if ($DevMode) {
		$readme_path = "$devmode_path\_README.txt"
		if (!(Test-Path -LiteralPath "$readme_path")) {
			New-Item -Path "$readme_path" | Out-Null
			$msg = "Script 'DevMode' Folder" +
				"`n----------------------------" +
				"`n`nWHAT? This folder holds output created when using the -DevMode argument" + 
				"`n`nNote: This folder and this README.txt file are programmatically created."
			$msg | Out-File -FilePath "$readme_path" -Append
		}
	}
}
# ------------------------------------------------------

# Temporarily disable Blue Iris database maintenance
$clip_maintenance_enabled_flg = $false
if ($disable_clip_maintenance) {
	$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips'
	$key_name = 'compact'
	$key_val = `
		RegGetValue `
		-KeyPath $key_path `
		-KeyName $key_name
			
	if ($key_val -eq 1) {
		$clip_maintenance_enabled_flg = $true
		$db_setting_state = 'enabled'
	} else {
		$db_setting_state = 'disabled'
	}
	<# if ($fromConsole -and !$Quiet) {
		Write-Host "Blue Iris database setting 'Compact/repair DB at' is currently $db_setting_state"
	} #>

	if ($key_val -eq 1) {

		$response = `
			ToggleDbMaintenance `
			-TargetState 'Off'
			
		# For debugging
		if ($response -eq 'success') {
			if ($fromConsole -and !$Quiet) {
				Write-Warning "Blue Iris database setting 'Compact/repair DB at' is now disabled"
				Write-Host ""
			}
		} else {
			if ($fromConsole) {
				Write-Warning "Failed to disable Blue Iris database setting 'Compact/repair DB at' -- now $db_setting_state"
				Write-Host ""
			}
		}
	}
}

if ($DevMode) {
	# CREATE DUMMY JPG					
	$jpg_filespec = "$devmode_path\dummy.jpg"
	
	$result_str = `
		MakeDummyJpgFile `
		-JpgFilespec $jpg_filespec `
		-Color 'Yellow' `
		-AddOverlayFlg $add_jpg_overlay `
		-FontName $jpg_overlay_font `
		-FontColor 'Black'
		
	# For debugging
	#Write-Host "MakeDummyJpgFile result_str = $result_str"

	if ($fromConsole) {
		if ($result_str -eq 'success') {
			Write-Host "(dev mode) created example of the yellow dummy JPG file --> '$jpg_filespec'" -ForegroundColor Cyan
		} else {
			Write-Host "(dev mode) HEADS-UP: problem creating demo dummy JPG file --> '$jpg_filespec'" -ForegroundColor Cyan
		}
	}
}

$timer_script_runtime = [Diagnostics.Stopwatch]::StartNew()

# INITIALIZE OUTPUT LISTS
[String[]]$tbl_al_cameras_ary = @()
[String[]]$tbl_al_fnames_ary = @()
[String[]]$tbl_al_paths_ary = @()
[String[]]$tbl_al_labels_ary = @()
[String[]]$tbl_al_hasjpgs_ary = @()
[String[]]$tbl_al_hasdats_ary = @()
[String[]]$tbl_al_haszerodates_ary = @()
[String[]]$tbl_al_dates_ary = @()
[Int[]]$tbl_al_flags_ary = @()
[String[]]$tbl_al_flags_codes_ary = @()
[String[]]$tbl_al_attribs_ary = @()
[String[]]$tbl_al_memos_ary = @()
[String[]]$tbl_al_issues_ary = @()
[String[]]$tbl_jpg_exif_memos_ary = @()
[String[]]$tbl_jpg_exif_memos_match_ary = @()
[String[]]$tbl_dat_memo_image_nos_ary = @()
[String[]]$tbl_dat_memos_ary = @()

# Initialize output csv file collectors
[String[]]$csv_fixed_paths_ary = @()
[String[]]$csv_fixed_labels_ary = @()
#[String[]]$csv_fixed_flags_ary = @()
[String[]]$csv_fixed_issues_ary = @()

# GET BLUE IRIS CAMERA LIST (if not specified by -Cameras argument)
if ($cams_ary.Count -eq 0) {
	$cliplist_hashtable = `
		BIJsonCamListParse `
		-CamProps 'FPS' `
		-QuietFlg $Quiet `
		-FromConsoleFlg $fromConsole
			
	# For debugging
	#Write-Host ("hashtable keys = " + $cliplist_hashtable.Keys)
	#Write-Host ("hashtable = " + ($cliplist_hashtable | ConvertTo-Json))

	$cams_ary = $cliplist_hashtable.Keys
}
$cams_arg = $cams_ary -Join ','
$patterns_arg = $patterns_ary -Join ','

# For debugging
#Write-Host "cams_ary = $cams_ary"
#Write-Host "cams_arg = $cams_arg"

# CREATE LOOKUP OBJECTS, e.g., @{db_path: memo}
# Used for protected alerts
$lookup_cameras_obj = @{}       # Initialize hashtable
$lookup_paths_obj = @{}         # Initialize hashtable
$lookup_files_obj = @{}         # Initialize hashtable
$lookup_labels_obj = @{}        # Initialize hashtable
$lookup_memos_obj = @{}         # Initialize hashtable
$lookup_comments_obj = @{}      # Initialize hashtable
$lookup_dates_obj = @{}         # Initialize hashtable
$lookup_flags_obj = @{}         # Initialize hashtable (decimal values)
$lookup_flags_bin_obj = @{}     # Initialize hashtable (binary values)
$lookup_hasjpgs_obj = @{}       # Initialize hashtable
$lookup_hasdats_obj = @{}       # Initialize hashtable
$lookup_haszerodates_obj = @{}  # Initialize hashtable
$lookup_zdfixes_obj = @{}       # Initialize hashtable
$lookup_attribs_obj = @{}       # Initialize hashtable
$lookup_issues_obj = @{}        # Initialize hashtable
$lookup_fixed_yn_obj = @{}      # Initialize hashtable
$lookup_dat_confimg_num_obj = @{}   # Initialize hashtable
$lookup_dat_confimg_memo_obj = @{}  # Initialize hashtable

$al_protected_wo_jpg_cntr = 0
$al_protected_wo_dat_cntr = 0
$al_protected_wo_jpg_and_dat_cntr = 0

# Start performance timer
$timer_alertlist_inspect = [Diagnostics.Stopwatch]::StartNew()

# GET BLUE IRIS DB ALERTLIST
$my_al_props = @('camera','path','file','flags','memo','date')  # We only need this subset of JSON alertlist params
#$my_al_props = @()  # Should cause BIJsonGetAlertList() to return ALL JSON alertlist params
# Returns hashtable, like @{ Cam1: @{ prop1:[values], ... , propN:[values] }, ... }
$response_obj = `
	BIJsonGetAlertList `
	-Cameras $cams_arg `
	-Patterns $patterns_arg `
	-AlertProps $my_al_props `
	-MemoTag $MemoTag `
	-QuietFlg $Quiet `
	-FromConsoleFlg $fromConsole `
	-DevModeFlg $DevMode `
	-DevModePath $devmode_path

# For debugging
#Write-Host ("BIJsonGetAlertList() response_obj type = " + $response_obj.GetType())
#Write-Host ("BIJsonGetAlertList() response_obj cnt = " + $response_obj.Count)
#Write-Host ("BIJsonGetAlertList() response_obj = " + ($response_obj | ConvertTo-Json))

if ($response_obj.Count -eq 0) {
	if ($Cameras -or $Patterns) {
		$err_msg = "ERR/Blue Iris JSON 'alertlist' cmd returned no Alerts; check arguments -Cameras & -Patterns"
	} else {
		$err_msg = "ERR/Blue Iris returned no Alerts"
	}
	return $err_msg
}		
if ($response_obj.GetType() -match 'string') {
	if ($response_obj.StartsWith('ERR/')) {
		$err_msg = $response_obj	
		if ($fromConsole) {
			Write-Host $err_msg -ForegroundColor Red
			exit
		} else {
			return $err_msg
		}
	} else {
		$err_msg = "ERR/unexpected response from BIJsonGetAlertList()"
		if ($fromConsole) {
			Write-Host $err_msg -ForegroundColor Red
			exit
		} else {
			return $err_msg
		}
	}
}

# PARSE THE ALERTLIST
$alertlist_hashtable = $response_obj
[String[]]$al_cams_ary = $alertlist_hashtable.Keys  # Ensures array contains unique items
		
# For debugging
#Write-Host ("hashtable keys = " + $alertlist_hashtable.Keys)
#Write-Host ("al_cams_ary = " + $al_cams_ary)
#Write-Host ("hashtable = " + ($alertlist_hashtable | ConvertTo-Json))

$filtered_alerts_cnt = 0

foreach ($cam in $al_cams_ary) {
		
	# For debugging
	#Write-Host ("alertlist_hashtable.$cam = " + ($alertlist_hashtable.$cam | ConvertTo-Json))
	
	if (!($alertlist_hashtable.$cam.camera -match 'NA')) {  # This conditional is required in PS5 (PS7 is forgiving)

		$alertlist_cameras_ary = [String[]]$alertlist_hashtable.$cam.camera
		$alertlist_paths_ary   = [String[]]$alertlist_hashtable.$cam.path
		$alertlist_files_ary   = [String[]]$alertlist_hashtable.$cam.file
		$alertlist_dates_ary   = [Int[]]$alertlist_hashtable.$cam.date   # PS GetType() before assignment is 'long'
		$alertlist_flags_ary   = [Int[]]$alertlist_hashtable.$cam.flags  # PS GetType before assignment is 'long'
		try {
			$alertlist_memos_ary = [String[]]$alertlist_hashtable.$cam.memo
			
			# For debugging
			#Write-Host ("alertlist_hashtable.$cam.memo = " + ($alertlist_hashtable.$cam.memo | ConvertTo-Json))
			
		} catch {
			$alertlist_memos_ary = 1..$alertlist_paths_ary.Count | % { 'NA' }
			$err_msg = "ERR/JSON 'alertlist' response for camera '$cam' contains no 'memo' parameters"
			return $err_msg
		}
		
		if ($alertlist_paths_ary.Count -ne $alertlist_memos_ary.Count) {
			$err_mag = "ERR/problem importing JSON alertlist"
			return $err_msg
		}
		
	} else {
		
		$alertlist_cameras_ary = [String[]]@()
		$alertlist_paths_ary   = [String[]]@()
		$alertlist_files_ary   = [String[]]@()
		$alertlist_dates_ary   = [Int[]]@()
		$alertlist_flags_ary   = [Int[]]@()
		$alertlist_memos_ary   = [String[]]@()
	}
		
	$filtered_alerts_cnt += $alertlist_paths_ary.Count
		
	# For debugging
	#Write-Host "cam = $cam"
	#Write-Host ("alertlist_cameras_ary = " + ($alertlist_cameras_ary | ConvertTo-Json))
	#Write-Host ("alertlist_cameras_ary cnt = " + $alertlist_cameras_ary.Count)
	#Write-Host ("alertlist_paths_ary type = " + $alertlist_paths_ary.GetType())
	#Write-Host ("alertlist_paths_ary cnt = " + $alertlist_paths_ary.Count)
	#Write-Host ("alertlist_paths_ary = " + ($alertlist_paths_ary | ConvertTo-Json))
	#Write-Host ("alertlist_files_ary = " + ($alertlist_files_ary | ConvertTo-Json))
	#Write-Host ("alertlist_dates_ary = " + ($alertlist_dates_ary | ConvertTo-Json))
	#Write-Host ("alertlist_dates_ary[0] = " + ($alertlist_dates_ary[0]).GetType())
	#Write-Host ("al_list_flags_ary = " + ($alertlist_flags_ary | ConvertTo-Json))
	#Write-Host ("al_list_flags_ary[0] = " + ($alertlist_flags_ary[0]).GetType())
	#Write-Host ("al_list_memos_ary = " + ($alertlist_memos_ary | ConvertTo-Json))
	
	# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# CHECK FOR NON-UNIQUE 'PATH' VALUES IN THE JSON ALERTLIST OUTPUT
	
	# WHY? because we are using the alertlist [path] values as keys in this script's lookup objects;
	# therefore, they must be unique
	
	# For testing
	#$alertlist_paths_ary += '@.0.jpg'
	#$alertlist_paths_ary += '@.0.jpg'
	#Write-Host ("(alertlist_paths_ary | Select -Unique) = " + (($alertlist_paths_ary | Select -Unique) | ConvertTo-Json))

	#if (($alertlist_paths_ary | Select -Unique).Count -ne $alertlist_paths_ary.Count) {
	[Array]$arr = $alertlist_paths_ary | Select -Unique
	if ($arr.Count -ne $alertlist_paths_ary.Count) {

		# Make string that shows the non-unique paths and count
		$out_str = $alertlist_paths_ary | Group-Object | Where-Object {$_.Count -gt 1} | Select-Object Name, Count | Out-String

		if ($fromConsole) {
			Write-Warning "JSON 'alertlist' response for '$cam' has some non-unique [path] values "
			Write-Host $out_str
			Write-Host "Recommend you try a Database Repair/Regenerate before proceeding" -ForegroundColor Red
			$query = Read-Host -Prompt "Proceed anyway (Y/N)?"
			if (!$query.StartsWith('Y')) { exit }
		} else {
			$err_str = "ERR/JSON 'alertlist' response for '$cam' has some non-unique [path] values"
			$err_str += "`n" + $out_str
			$err_str += "`nRecommend you try a Database Repair/Regenerate before proceeding"
			return $err_str
		}
	}
	# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	# HANDLE SIMULATOR ARG -SIM6 (ZERO DATE)
	$Sim6_active_flg = $false
	if ($Sim6_arg_exists) {
		# ... the simulator only works on < 4 alerts
		if ($filtered_alerts_cnt -le 3) {
			$Sim6_active_flg = $true
			$Sim6_adj_sec = $Sim6			
			if ($fromConsole) {
				Write-Host "Using simulator -Sim6; forces up to 3 Alerts to have a zero-date" -ForegroundColor Cyan
			}
		} else {							
			if ($fromConsole) {
				Write-Host "Unable to use zero-date simulator argument -Sim6; works only with < 4 Alerts" -ForegroundColor Red
			}
		}
	}
	
	# FOR PROTECTED ALERTS ONLY...
	# Make Alert property lookup objects (files, flags, memos, ...)
	$i = 0
	foreach ($item in $alertlist_paths_ary) {  # $item is like "@123456.bvr"

		if ($alertlist_cameras_ary[$i] -contains $cam) {
			
			# Make key for use in lookup_xxx_obj lines below
			$path_key = $item
			
			if ($path_key -eq '@0.jpg' -and !$fix_type_arg_used_flg) {
				
				$file_val = $alertlist_files_ary[$i]
				$label_val = $file_val.Replace('.jpg', '')

				if ($fromConsole) {
					#Write-Host "$label_val - db record is unfixable! (alertlist 'path = @0.jpg'" -ForegroundColor Yellow
				}

				#continue
			}

			# Initialize Alert 'not protected-by-file' flags
			$al_protected_wo_jpg_flg = $false
			$al_protected_wo_dat_flg = $false
			$al_protected_wo_jpg_and_dat_flg = $false

			# INSPECT ALERT'S DATABASE FLAGS
			# ... returned array is like, [0,0,1,0,...]
			$flags_bin_ary = `
				MakeBitArray `
				-DecVal $alertlist_flags_ary[$i] `
				-NoBits 32

			# For debugging
			#Write-Host ("al_list_flags_ary[$i] = " + $alertlist_flags_ary[$i])
			#Write-Host "flags_bin_ary = $flags_bin_ary"				
			<# if ($flags_bin_ary[1] -eq 1) {
				Write-Host "$item alert is Flagged"
			}
			if ($flags_bin_ary[1] -eq 1) {
				Write-Host "$item alert is Protected"
			}
			if ($flags_bin_ary[28] -eq 1) {
				Write-Host "$item alert is Confirmed"
			} #>

			# CONFIRM THAT ALERT IS PROTECTED (OR APPLY -Flags ARGUMENT)
			# Flags: 2^0=Flagged, 2^1=Protected, 2^6=Archive, 2^9=Export, 2^28 Confirmed
			if ($flags_bin_ary[2] -eq 1 -or $AllAlerts) {
				
				# HANDLE THE -Flags ARGUMENT
				if ($flags_arg_ary_cnt -gt 0) {
					
					$test_ary = @()
					$test_ary += ($flags_arg_ary -contains 'F'  -and $flags_bin_ary[1]  -eq 1)
					$test_ary += ($flags_arg_ary -contains '!F' -and $flags_bin_ary[1]  -eq 0)
					$test_ary += ($flags_arg_ary -contains 'P'  -and $flags_bin_ary[2]  -eq 1)
					$test_ary += ($flags_arg_ary -contains '!P' -and $flags_bin_ary[2]  -eq 0)
					$test_ary += ($flags_arg_ary -contains 'C'  -and $flags_bin_ary[28] -eq 1)
					$test_ary += ($flags_arg_ary -contains '!C' -and $flags_bin_ary[28] -eq 0)
				
					# Get counts of True and False elements in $test_ary (in a hashtable $ht)
					# REF: https://stackoverflow.com/a/12640991
					# like, $ht = {         or $ht = {
					#       "False": 4,        "False": 6
					#       "True": 2          }
					#       }
					$test_ary | group | % { $ht = @{} } { $ht[$_.Name] = $_.Count } { $ht } | Out-Null

					# For debugging
					<# if ($i -lt 10) {
						Write-Host ("test_ary = " + ($test_ary | ConvertTo-Json))
						Write-Host ("ht = " + ($ht | ConvertTo-Json))
						Write-Host ("ht['False'] = " + $ht['False'] + ", type = " + $ht['False'].GetType())						
						if ($null -ne $ht['True']) {
							Write-Host ("ht['True'] = " + $ht['True'] + ", type = " + $ht['True'].GetType())						
						} else {
							Write-Host "ht['True'] = null"
						}
					} #>
					
					if ($null -eq $ht['True']) {
						# No 'True' elements found; so skip to the next Alert
						$i++
						continue
					} else {
						# N entries in the -Flags argument require at least N 'True' element (AND conditional)
						$test_ary_true_cnt = $ht['True']
						if ($test_ary_true_cnt -lt $flags_arg_ary_cnt) {
							# Skip to the next Alert
							$i++
							continue
						}
					}
				}

				# Initialize temporary collectors
				[String[]]$tmp_codes_ary = @()
				[String[]]$tmp_issues_ary = @()

				# For debugging
				#Write-Host ("flags_bin_ary = " + ($flags_bin_ary -Join "") + " (" + $alertlist_flags_ary[$i] + ") " + $item)

				# POPULATE LOOKUP COLLECTORS: $lookup_xxx_obj
				# AND OUTPUT TABLE COLLECTORS:  $tbl_al_xxx_ary
				
				# --> xxx = files, labels, paths, memos, dates, dates & zerodates, flags
				if (1 -eq 1) {  # Use to collapse this lengthy code section while editing

					# ... 'cameras'
					$cam_val = $alertlist_cameras_ary[$i]
					if (!$lookup_files_obj[$path_key]) {
						$lookup_cameras_obj.Add($path_key, $cam_val)
						$tbl_al_cameras_ary += $cam_val
					}

					# ... 'files'
					$file_val = $alertlist_files_ary[$i]
					if (!$lookup_files_obj[$path_key]) {
						$lookup_files_obj.Add($path_key, $file_val)
						$tbl_al_fnames_ary += $file_val
					}

					# ... 'labels'
					$label_val = $file_val.Replace('.jpg', '')
					if (!$lookup_labels_obj[$path_key]) {
						$lookup_labels_obj.Add($path_key, $label_val)
						$tbl_al_labels_ary += $label_val
					}

					# ... 'paths'
					$path_val = $alertlist_paths_ary[$i]
					if (!$lookup_paths_obj[$path_key]) {
						$lookup_paths_obj.Add($path_key, $path_val)
						$tbl_al_paths_ary += $path_val
					}

					# ... 'memos'
					$memo_val = $alertlist_memos_ary[$i]
					if (!$lookup_memos_obj[$path_key]) {
						$lookup_memos_obj.Add($path_key, $memo_val)
						$tbl_al_memos_ary += $memo_val
					}

					# ... 'flags' - letter codes, like 'F','P', 'C'
					# Flags: 2^0=Flagged, 2^1=Protected, 2^6=Archive, 2^9=Export, 2^28 Confirmed
					$flags_val = $alertlist_flags_ary[$i]
					if (!$lookup_flags_bin_obj[$path_key]) {
						$lookup_flags_obj.Add($path_key, $flags_val)
						$lookup_flags_bin_obj.Add($path_key, $flags_bin_ary)
						$tbl_al_flags_ary += $alertlist_flags_ary[$i]

						# For debugging
						#Write-Host "flags_bin_ary = $flags_bin_ary"
						
						$arr = @()  # Initialize temporary accumulator
						if ($flags_bin_ary[1] -eq 1) {  # Flagged
							$arr += 'F'
						} else {
							$arr += ''
						}
						if ($flags_bin_ary[2] -eq 1) {  # Protected
							$arr += 'P'
						} else {
							$arr += ''
						}
						if ($flags_bin_ary[28] -eq 1) {  # Confirmed
							$arr += 'C'
						} else {
							$arr += ''
						}
						
						$tbl_al_flags_codes_ary += $arr -Join "" 
					}

					# ... 'dates' and 'zero-dates'
					$date_val = $alertlist_dates_ary[$i]
					if (!$lookup_dates_obj[$path_key]) {
						$lookup_dates_obj.Add($path_key, $date_val)
						$tbl_al_dates_ary += $date_val
					}

					# HANDLE SIMULATOR ARG -SIM6 (ZERO DATE)
					if ($Sim6_active_flg) {
					
						# For debugging
						Write-Host ("lookup_dates_obj[$path_key] BEFORE ZD = " + ($lookup_dates_obj[$path_key] | ConvertTo-Json))
						
						if ($lookup_dates_obj[$path_key] -gt 0) {

							# Update the alert's database record
							# ... returns "success' or 'fail'
							$result_str = `
								UpdateDbRecord `
								-AlertDbPath $path_val `
								-AlertDbDateSec 0 `
								-AlertDbFlags $flags_val `
								-AlertDbMemo $memo_val `
								-FromConsoleFlg $fromConsole `
								-QuietFlg $Quiet `
								-DevModeFlg $DevMode `

							if ($result_str -match 'success') {

								$lookup_dates_obj[$path_key] = 0

								# Update issues collector
								$lookup_issues_obj[$path_key] += 'Zero-date'

								# Update output collector
								$tbl_al_dates_ary[$i] = 0
								
								if ($fromConsole -and !$Quiet) {
									Write-Host "$label_val - Sim6 : db date changed to zero"
								}
								
							} else {
								if ($fromConsole -and !$Quiet) {
									Write-Host "$label_val - Sim6 : db date NOT changed to zero"
								}
							}
						
							# For debugging
							Write-Host ("$label_val - lookup_dates_obj[$path_key] AFTER ZD = " + ($lookup_dates_obj[$path_key] | ConvertTo-Json))
							
						} else {
							Write-Host "$label_val - Nothing to do - Alert's date is already zero"
						}
						
					}
						
					$date_val_sec = $lookup_dates_obj[$path_key]
					if ($date_val_sec -eq 0) {
						$haszerodate_val = $true
						$tbl_al_haszerodates_ary += 'x'
						$tmp_issues_ary += "Zero-date"
					} else {
						$haszerodate_val = $false
						$tbl_al_haszerodates_ary += ''
					}
					if (!$lookup_haszerodates_obj[$path_key]) {
						$lookup_haszerodates_obj.Add($path_key, $haszerodate_val)
					}
				}

				# --> xxx = hasjpgs, hasdats, atrribs
				# Populate collectors: Alert JPG and DAT file read-only attributes
				if (1 -eq 1) {  # Use to collapse this lengthy code section while editing

					$jpg_filespec = $bi_alerts_folder_path + "\" + $file_val
					$jpg_file_exists = (Test-Path -LiteralPath "$jpg_filespec")

					if ($jpg_file_exists) {
						if (!$lookup_hasjpgs_obj[$path_key]) {
							$lookup_hasjpgs_obj.Add($path_key, $true)
							$tbl_al_hasjpgs_ary += "x"
						}
						$obj = Get-ItemProperty -LiteralPath "$jpg_filespec" | Select-Object IsReadOnly
						if ($obj.IsReadOnly) {
							$tmp_codes_ary += 'J'	# JPG file is read-only (may be protected by Blue Iris)
						} else {
							$tmp_codes_ary += ''	
							$tmp_issues_ary += "JPG file not read-only"
						}
					} else {
						if (!$lookup_hasjpgs_obj[$path_key]) {
							$lookup_hasjpgs_obj.Add($path_key, $false)
							$tbl_al_hasjpgs_ary += ''
							#$tbl_al_hasjpgs_ary += $null
						}
						$al_protected_wo_jpg_flg = $true
						$al_protected_wo_jpg_cntr++
						$tmp_codes_ary += ''
						$tmp_issues_ary += 'No JPG file'
					}

					$dat_filespec = $bi_alerts_folder_path + "\" + $file_val.Replace('.jpg','.dat')
					$dat_file_exists = (Test-Path -LiteralPath "$dat_filespec")

					if ($dat_file_exists) {

						if (!$lookup_hasdats_obj[$path_key]) {
							$lookup_hasdats_obj.Add($path_key, $true)
							$tbl_al_hasdats_ary += "x"
						}
						$obj = Get-ItemProperty -LiteralPath "$dat_filespec" | Select-Object IsReadOnly
						if ($obj.IsReadOnly) {
							$tmp_codes_ary += 'D'	# DAT file is read-only (may be protected by Blue Iris)
						} else {
							$tmp_codes_ary += ''	
							$tmp_issues_ary += 'DAT file not read-only'
						}

					} else {

						if (!$lookup_hasdats_obj[$path_key]) {
							$lookup_hasdats_obj.Add($path_key, $false)
							$tbl_al_hasdats_ary += ""
						}
						$al_protected_wo_dat_flg = $true
						$al_protected_wo_dat_cntr++
						$tmp_codes_ary += ''
						$tmp_issues_ary += 'No DAT file'
					}
					
					if ($al_protected_wo_jpg_flg -and $al_protected_wo_dat_flg) {
						$al_protected_wo_jpg_and_dat_flg = $true
						$al_protected_wo_jpg_and_dat_cntr++
					}

					# Update lookup collectors
					$attrib_str = $tmp_codes_ary -Join ''
					$lookup_attribs_obj[$path_key] = $attrib_str
					$lookup_issues_obj[$path_key] = $tmp_issues_ary
					
					# Update output collectors
					$issues_str = ($tmp_issues_ary | Select -Unique) -Join '; '
					$tbl_al_issues_ary += $issues_str
					$tbl_al_attribs_ary += $attrib_str
					
					# For debugging
					#Write-Host ("tmp_codes_ary = " + ($tmp_codes_ary | ConvertTo-Json))
					#Write-Host ("tmp_issues_ary = " + ($tmp_issues_ary | ConvertTo-Json))
				}

			}
			
			$i++
		}
	}

	# Stop performance timer
	$timer_alertlist_inspect.Stop()
	
	# For debugging
	#Write-Host ("lookup_paths_obj* = " + ($lookup_paths_obj | ConvertTo-Json))
	#Write-Host ("tbl_al_issues_ary = " + ($tbl_al_issues_ary | ConvertTo-Json))
	#Write-Host ("alertlist_dates_ary = " + ($alertlist_dates_ary | ConvertTo-Json))
	#Write-Host ("lookup_dates_obj = " + ($lookup_dates_obj | ConvertTo-Json))
	#Write-Host ("lookup_haszerodates_obj = " + ($lookup_haszerodates_obj | ConvertTo-Json))
	#Write-Host ("lookup_issues_obj = " + ($lookup_issues_obj | ConvertTo-Json))
	#Write-Host ("tbl_al_issues_ary = " + ($tbl_al_issues_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_hasdats_ary = " + ($tbl_al_hasdats_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_haszerodates_ary = " + ($tbl_al_haszerodates_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_paths_ary = " + ($tbl_al_paths_ary | ConvertTo-Json))


	$al_protected_cnt = $tbl_al_fnames_ary.Count

	if ($DevMode) {
		if ($fromConsole) {
			$elapsed_sec = ($timer_alertlist_inspect.elapsed.totalseconds).ToString("0.00")
			$msec_per_alert = (1000*$elapsed_sec/$filtered_alerts_cnt).ToString("0.0")
			Write-Host "(dev mode) found alerts inspect time = $elapsed_sec sec ($msec_per_alert msec/alert) ($filtered_alerts_cnt total)" -ForegroundColor Cyan
			Write-Host ("(dev mode) found alerts that are protected = " + (($al_protected_cnt/$filtered_alerts_cnt)).ToString("0.00%") + " ($al_protected_cnt/$filtered_alerts_cnt)") -ForegroundColor Cyan
		}
	}
} 

if ($fromConsole -and !$Quiet) {
	if ($AllAlerts) {
		Write-Host ("Found [ " + $al_protected_cnt + " ]" + $p_adj + "Alert(s) in the database:")
	} else {
		Write-Host ("Found [ " + $al_protected_cnt + " ]" + $p_adj + "Alert(s) in the database (" + $filtered_alerts_cnt + " total Alerts):")
	}
	$verb1 = "lack"
	if ($al_protected_wo_jpg_cntr -eq 1) {
		$verb1 += "s"
	}
	$verb2 = "lack"
	if ($al_protected_wo_dat_cntr -eq 1) {
		$verb2 += "s"
	}
	$verb3 = "lack"
	if ($al_protected_wo_jpg_and_dat_cntr -eq 1) {
		$verb3 += "s"
	}
	$inspect_summary_str = "$al_protected_wo_jpg_cntr $verb1 a JPG file, $al_protected_wo_dat_cntr $verb2 a DAT file, $al_protected_wo_jpg_and_dat_cntr $verb3 both files"
	#Write-Host "Summary: $inspect_summary_str"
	Write-Host "of these... $inspect_summary_str"
}
	
# For debugging
#Write-Host ("tbl_al_fnames_ary = " + ($tbl_al_fnames_ary | ConvertTo-Json))
#Write-Host "tbl_al_labels_ary = $tbl_al_labels_ary"
#Write-Host ("lookup_paths_obj** = " + ($lookup_paths_obj | ConvertTo-Json))
#Write-Host ("lookup_files_obj type = " + $lookup_files_obj.GetType())
#Write-Host ("lookup_files_obj cnt = " + $lookup_files_obj.Count)
#Write-Host ("lookup_files_obj = " + ($lookup_files_obj | ConvertTo-Json))
#Write-Host ("lookup_labels_obj = " + ($lookup_labels_obj | ConvertTo-Json))
#Write-Host ("lookup_memos_obj type = " + $lookup_memos_obj.GetType())
#Write-Host ("lookup_memos_obj cnt = " + $lookup_memos_obj.Count)
#Write-Host ("lookup_memos_obj = " + ($lookup_memos_obj | ConvertTo-Json))
#Write-Host ("lookup_dates_obj = " + ($lookup_dates_obj | ConvertTo-Json))
#Write-Host ("lookup_flags_obj type = " + $lookup_flags_obj.GetType())
#Write-Host ("lookup_flags_obj cnt = " + $lookup_flags_obj.Count)
#Write-Host ("lookup_flags_obj = " + ($lookup_flags_obj | ConvertTo-Json))
#Write-Host ("lookup_flags_bin_obj type = " + $lookup_flags_bin_obj.GetType())
#Write-Host ("lookup_flags_bin_obj cnt = " + $lookup_flags_bin_obj.Count)
#Write-Host ("lookup_flags_bin_obj = " + ($lookup_flags_bin_obj | ConvertTo-Json))
#Write-Host ("all_al_attribs_obj cnt = " + $lookup_attribs_obj.Count)
#Write-Host ("lookup_hasjpgs_obj = " + ($lookup_hasjpgs_obj | ConvertTo-Json))
#Write-Host ("lookup_hasdats_obj = " + ($lookup_hasdats_obj | ConvertTo-Json))
#Write-Host ("lookup_haszerodates_obj = " + ($lookup_haszerodates_obj | ConvertTo-Json))

# GET DAT FILE CONFIRMED IMAGE NUMBERS AND MEMOS
if ($al_protected_cnt -gt 0) {

	#$found_alerts_cnt = $al_protected_cnt - $al_protected_wo_dat_cntr
	$found_alerts_cnt = $al_protected_cnt

	#if ($found_alerts_cnt -gt 0) {

		if (!$NoInspect) {
			if ($fromConsole -and !$Quiet) {
				if ($found_alerts_cnt -gt 1) {
					Write-Host ("`nInspecting DAT files (" + $found_alerts_cnt + $p_adj + "Alerts have DAT files)...")
				} else {
					Write-Host ("`nInspecting DAT files (1" + $p_adj + "Alert has a DAT file)...")
				}
			}
		}

		# Start performance timer
		$timer_dat_confimg_inspect = [Diagnostics.Stopwatch]::StartNew()

		$progress_bar_cntr = 0
		$i = 0
		foreach($item in $tbl_al_paths_ary) {  

			if (!$NoInspect) {
				
				# Show progress bar
				if ($fromConsole) {
					$progress_bar_cntr++
					$WriteProgressParams = @{
						Activity         = "Inspecting DAT files..."
						Status           = "Progress: $progress_bar_cntr of $found_alerts_cnt"
						PercentComplete  = (($progress_bar_cntr / $found_alerts_cnt) * 100)
					}
					# Note: If loop is too fast, you might need a delay to get the progress bar to update
					#Start-Sleep -Milliseconds 150				
					Write-Progress @WriteProgressParams
				}
				
				# Make key for use in lookup_xxx_obj lines below
				$path_key = $item

				$dat_filespec = $bi_alerts_folder_path + "\" + $lookup_files_obj[$path_key].Replace('.jpg','.dat')
				$dat_file_exists = (Test-Path -LiteralPath "$dat_filespec")
				
				if ($dat_file_exists) {
					
					# Returns: [array] @((int)confirmed_image_num, (string)confirmed_image_memo_str, (array)bad_dat_files)

					# ... $confirmed_image_num = 0 if no confirmed image found
					$confirmed_image_num, $confirmed_image_memo_str, $bad_dat_flg = `
						ExtractDatConfImgInfo `
						-DatPath $dat_filespec `
						-FromConsoleFlg $fromConsole `
						-QuietFlg $Quiet

					# For debugging
					#Write-Host ("confirmed_image_num type = " + $confirmed_image_num.GetType())
					#Write-Host "confirmed_image_num = $confirmed_image_num"
					#Write-Host ("confirmed_image_memo_str type = " + $confirmed_image_memo_str.GetType())
					#Write-Host "confirmed_image_memo_str = $confirmed_image_memo_str"
					#Write-Host ("bad_dat_files type = " + $bad_dat_files.GetType())
					#Write-Host "bad_dat_flg = $bad_dat_flg"
					
					if ($confirmed_image_num -ne 0 -and $confirmed_image_memo_str.Length -ne 0) {

						# Update lookup collectors
						$lookup_dat_confimg_num_obj.Add($path_key, $confirmed_image_num)
						$lookup_dat_confimg_memo_obj.Add($path_key, $confirmed_image_memo_str)

						# Update output collectors
						$tbl_dat_memo_image_nos_ary += $confirmed_image_num
						$tbl_dat_memos_ary += $confirmed_image_memo_str

					} else {
						
						# Update lookup collectors
						$lookup_dat_confimg_num_obj.Add($path_key, '')
						$lookup_dat_confimg_memo_obj.Add($path_key, '')
						
						# Update output collectors
						$tbl_dat_memo_image_nos_ary += ''
						$tbl_dat_memos_ary += ''
						
						if ($bad_dat_flg) {
							
							# Update issues collector
							$lookup_issues_obj[$path_key] += 'Bad DAT file'

							# Update output collector
							$tbl_al_issues_ary[$i] = 'Bad DAT file'
							
						}
					}

					# For debugging
					#Write-Host ("lookup_dat_confimg_num_obj = " + ($lookup_dat_confimg_num_obj | ConvertTo-Json))
					#Write-Host ("lookup_dat_confimg_memo_obj = " + ($lookup_dat_confimg_memo_obj | ConvertTo-Json))

				} else {
					
					# Update lookup collectors
					$lookup_dat_confimg_num_obj.Add($path_key, '')
					$lookup_dat_confimg_memo_obj.Add($path_key, '')
					
					# Update output collectors
					$tbl_dat_memo_image_nos_ary += ''
					$tbl_dat_memos_ary += ''
				}

			} else {
				
				# Update output collecotrs
				$tbl_dat_memo_image_nos_ary += 'skipped'
				$tbl_dat_memos_ary += 'skipped'
			}

			$i++
		}

		# For debugging
		#Write-Host ("tbl_dat_memo_image_nos_ary cnt = " + $tbl_dat_memo_image_nos_ary.Count)
		#Write-Host "tbl_dat_memo_image_nos_ary = $tbl_dat_memo_image_nos_ary"
		#Write-Host ("tbl_dat_memos_ary cnt = " + $tbl_dat_memos_ary.Count)
		#Write-Host ("tbl_dat_memos_ary = " + ($tbl_dat_memos_ary | ConvertTo-Json))

		# Stop performance timer
		$timer_dat_confimg_inspect.Stop()

		if ($DevMode) {
			if ($fromConsole) {
				$elapsed_sec = ($timer_dat_confimg_inspect.elapsed.totalseconds).ToString("0.00")
				$msec_per_dat = (1000*$elapsed_sec/$found_alerts_cnt).ToString("0.0")
				Write-Host "(dev mode) found alerts dat confirmed image info gather time = $elapsed_sec sec ($msec_per_dat msec/DAT) ($found_alerts_cnt total)" -ForegroundColor Cyan
			}
		}
		
	#}
}

# INSPECT JPG EXIF METADATA
if ($al_protected_cnt -gt 0) {
	
	$found_alerts_cnt = $al_protected_cnt - $al_protected_wo_jpg_cntr
	
	if (!$NoInspect) {
		if ($fromConsole -and !$Quiet) {
			if ($found_alerts_cnt -gt 0) {
				if ($found_alerts_cnt -gt 1) {
					Write-Host ("`nInspecting JPG files Exif metadata (" + $found_alerts_cnt + $p_adj + "Alerts have JPG files)...")
				} else {
					Write-Host ("`nInspecting JPG files Exif metadata (1" + $p_adj + "Alert has a JPG file)...")
				}
			}
		}
	}

	# Initialize counters
	$jpg_mismatch_cntr = 0

	# Start performance timer
	$timer_exif_inspect = [Diagnostics.Stopwatch]::StartNew()
	
	# Initialize progress bar
	$progress_bar_cntr = 0
	
	$i = 0
	foreach ($item in $tbl_al_fnames_ary) {  # $item is like "GT.20231009_130000.1224078.5-0.jpg"

		$path_key = $tbl_al_paths_ary[$i]  # like, '@123456.bvr'
		
		$jpg_filespec = $bi_alerts_folder_path + "\" + $item
		$jpg_file_exists = (Test-Path -LiteralPath "$jpg_filespec")

		# For debugging
		#Write-Host "jpg_filespec = $jpg_filespec"
		#Write-Host "path_key = $path_key"
		#Write-Host ("lookup_issues_obj[$path_key] = " + ($lookup_issues_obj[$path_key] | ConvertTo-Json))
		
		if (!$NoInspect) {
			
			# COMPARE DATABASE MEMO TO JPG EXIF MEMO
			if ($jpg_file_exists) {

				# Show progress bar
				if ($fromConsole) {
					$progress_bar_cntr++
					$WriteProgressParams = @{
						Activity         = "Inspecting JPG files..."
						Status           = "Progress: $progress_bar_cntr of $al_protected_cnt"
						PercentComplete  = (($progress_bar_cntr / $al_protected_cnt) * 100)
					}
					Write-Progress @WriteProgressParams
				}
				
				# GET EXIF METADATA AS SHOWN BY WINDOWS EXPLORER
				# ... SOURCE: https://evotec.xyz/getting-file-metadata-with-powershell-similar-to-what-windows-explorer-provides/

				# Returns hashtable like @{'title: str, 'subject': str, 'comments': str ... }
				# NOTE: function tests existence of dat_filespec
				$exif_obj = GetJpgExifMetadata -FilePath $jpg_filespec

				# For debugging
				#Write-Host ("exif_obj type = " + $exif_obj.GetType())
				#Write-Host ("exif_obj = " + ($exif_obj | ConvertTo-Json))

				$exif_title = $exif_obj.title
				
				if ($exif_title.Length -eq 0) {
					$exif_title = "<blank>"
				}
				
				# ////////////////////////////////////////////////////////////////////////////// 
				# @@@ DEV - for now, we're ignoring the apparently redundant $exif_subject
				# NOTE: Blue Iris JSON 'update' command does not update the JPG Exif 'Subject'
				<# if ($exif_title -eq $exif_subject) {
					$tbl_jpg_exif_memos_ary += $exif_title
				} else {
					$tbl_jpg_exif_memos_ary += $exif_title + "," + $exif_subject
				} #>
				# ////////////////////////////////////////////////////////////////////////////// 

				# ////////////////////////////////////////////////////////////////////////////// 
				# @@@ DEV - Handle Exit 'Comments' 'Property
				#[String[]]$comments_bits = $exif_comments -Split '.'
				
				# For debugging
				#Write-Host ("comments_bits cnt = " + $comments_bits.Count)
				#Write-Host ("comments_bits = " + ($comments_bits | ConvertTo-Json))
				
				<# $jpg_exif_alert_tbd_1 = $comments_bits[0]
				if ($comments_bits.Count -gt 1) {
					$jpg_exif_alert_offset = $comments_bits[1]
					[String[]]$arr = $comments_bits[2] -Split '-'
					$jpg_exif_alert_flags = $arr[0]
					if ($arr.Count -gt 1) {
						$jpg_exif_alert_zones = $arr[1]
					} else {
						$jpg_exif_alert_zones = ''
					}
					if ($arr.Count -gt 4) {
						$jpg_exif_alert_tbd_2 = $comments_bits[4]
						$jpg_exif_alert_duration = $comments_bits[4]
					} else {
						$jpg_exif_alert_tbd_2 = ''
						$jpg_exif_alert_duration = ''
					}
				} else {
					$jpg_exif_alert_offset = ''
					$jpg_exif_alert_flags = ''
				} #>
				# ////////////////////////////////////////////////////////////////////////////// 
				
				# Update output collector
				$tbl_jpg_exif_memos_ary += $exif_title
				
				# COMPARE DATABASE MEMO TO JPG EXIF MEMO

				# For debugging
				#Write-Host ("tbl_al_memos_ary[$i].Trim() = '" + $tbl_al_memos_ary[$i].Trim() + "'")
				#Write-Host ("exif_title.Trim() = '" + $exif_title.Trim() + "'")
				
				if ($tbl_al_memos_ary[$i].Replace('<blank>', '').Trim() -eq $exif_title.Replace('<blank>', '').Trim()) {

					# MATCH!

					# Update output collector
					$tbl_jpg_exif_memos_match_ary += 'match!'

				} else {

					$jpg_mismatch_cntr++
					
					# Update issues collector
					$lookup_issues_obj[$path_key] += 'Memo/JPG Exif mismatch'

					# Update output collectors
					$tbl_jpg_exif_memos_match_ary += $tbl_jpg_exif_memos_ary[$i]
					$issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
					$tbl_al_issues_ary[$i] = $issues_str
					
				}

			} else {

				# Update issues collector
				$lookup_issues_obj[$path_key] += 'No JPG file'
				
				# Update output collectors
				$tbl_jpg_exif_memos_ary += "<NA>"
				$tbl_jpg_exif_memos_match_ary += "<NA>"
				$issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
				$tbl_al_issues_ary[$i] = $issues_str
			}
			
		} else {

			# Update output collectors
			$tbl_jpg_exif_memos_ary += 'skipped'
			$tbl_jpg_exif_memos_match_ary += 'skipped'
		}

		$i++
	}
	
	# Update output collector
	#$issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
	#$tbl_al_issues_ary[$i] = $issues_str
	
	if ($jpg_mismatch_cntr -gt 0) {
		if ($fromConsole -and !$Quiet) {
			Write-Host "Found $jpg_mismatch_cntr Alert JPG files with unexpected metadata (not matching database alert memo)"
		}
	}

	# Stop performance timer
	$timer_exif_inspect.Stop()

	if ($DevMode) {
		if ($fromConsole) {
			$elapsed_sec = ($timer_exif_inspect.elapsed.totalseconds).ToString("0.00")
			$msec_per_file = (1000*$elapsed_sec/$found_alerts_cnt).ToString("0.0")
			Write-Host "(dev mode) found alerts jpg exif vs db memo compare time = $elapsed_sec sec ($msec_per_file msec/file) ($found_alerts_cnt files)" -ForegroundColor Cyan
		}
	}

}

# For debugging
#Write-Host ("lookup_issues_obj AFTER Exif inspect= " + ($lookup_issues_obj | ConvertTo-Json))

# INSPECT DAT FILE CONFIRMED IMAGE MEMOS
if ($al_protected_cnt -gt 0) {
	
	$found_alerts_cnt = $al_protected_cnt - $al_protected_wo_dat_cntr
	
	if (!$NoInspect) {
		if ($fromConsole -and !$Quiet) {
			if ($found_alerts_cnt -gt 0) {
				if ($found_alerts_cnt -gt 1) {
					Write-Host ("`nInspecting DAT file confirmed image memos (" + $found_alerts_cnt + $p_adj + "Alerts have DAT files)...")
				} else {
					Write-Host ("`nInspecting DAT file confirmed image memos (1" + $p_adj + "Alert has a DAT file)...")
				}
			}
		}
	}

	# Initialize counters\
	$dat_mismatch_cntr = 0

	# Start performance timer
	$timer_dat_memo_inspect = [Diagnostics.Stopwatch]::StartNew()
	
	# Initialize progress bar
	$progress_bar_cntr = 0
	
	$i = 0
	foreach ($item in $tbl_al_fnames_ary) {  # $item is like "GT.20231009_130000.1224078.5-0.jpg"

		$path_key = $tbl_al_paths_ary[$i]  # like, '@123456.bvr'

		# For debugging
		#Write-Host "path_key = $path_key"
		#Write-Host ("lookup_issues_obj[$path_key] = " + ($lookup_issues_obj[$path_key] | ConvertTo-Json))
		
		if (!$NoInspect) {
			
			# COMPARE DATABASE MEMO TO DAT FILE CONFIRMED IMAGE MEMO 
			if ($lookup_dat_confimg_memo_obj[$path_key]) {

				$confirmed_image_memo_str = $lookup_dat_confimg_memo_obj[$path_key]
				
				# For debugging
				#Write-Host ("tbl_al_memos_ary[$i].Trim() = '" + $tbl_al_memos_ary[$i].Trim() + "'")
				#Write-Host ("confirmed_image_memo_str.Trim() = '" + $confirmed_image_memo_str.Trim() + "'")
				
				$test1 = (($tbl_al_memos_ary[$i].Replace('<blank>', '').Trim()).Length -eq 0)
				$test2 = (($confirmed_image_memo_str.Trim()).Length -eq 0)
				if ($test1 -and !$test2) {

					$dat_mismatch_cntr++
					
					# Update issues collector
					$lookup_issues_obj[$path_key] += 'BlankMemo/DatImgMemo mismatch'

					# Update output collector
					$issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
					$tbl_al_issues_ary[$i] = $issues_str
					
				}
				
			}
			
		} else {

			# Update output collectors
			$tbl_jpg_exif_memos_ary += 'skipped'
			$tbl_jpg_exif_memos_match_ary += 'skipped'
		}

		$i++
	}
	
	# Update output collector
	#$issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
	#$tbl_al_issues_ary[$i] = $issues_str
	
	if ($jpg_mismatch_cntr -gt 0) {
		if ($fromConsole -and !$Quiet) {
			Write-Host "Found $jpg_mismatch_cntr Alert JPG files with unexpected metadata (not matching database alert memo)"
		}
	}

	# Stop performance timer
	$timer_dat_memo_inspect.Stop()

	<# if ($DevMode) {
		if ($fromConsole) {
			$elapsed_msec = ($timer_dat_memo_inspect.elapsed.totalmilliseconds).ToString("0.000")
			$msec_per_file = ($elapsed_msec/$found_alerts_cnt).ToString("0.000")
			Write-Host "(dev mode) found alerts dat confirmed image memo vs db memo compare time = $elapsed_msec msec ($msec_per_file msec/file) ($found_alerts_cnt files)" -ForegroundColor Cyan
		}
	} #>

}

# For debugging
#Write-Host ("lookup_issues_obj AFTER DAT confirmed image inspect= " + ($lookup_issues_obj | ConvertTo-Json))

# APPLY FILTERS 
if ($filter_arg_set_flg) {

	# REMOVE ENTRIES FROM ORIGINAL FILTERED LISTS
	# NOTE: must convert arrays to Powershell lists to remove array elements by index

	[Collections.Generic.List[String]]$al_fnames_lst            = $tbl_al_fnames_ary
	[Collections.Generic.List[String]]$al_fnames_lst2           = $tbl_al_fnames_ary
	[Collections.Generic.List[String]]$al_paths_lst             = $tbl_al_paths_ary
	[Collections.Generic.List[String]]$al_dates_lst             = $tbl_al_dates_ary
	[Collections.Generic.List[String]]$al_labels_lst            = $tbl_al_labels_ary
	[Collections.Generic.List[String]]$al_flags_lst             = $tbl_al_flags_ary
	[Collections.Generic.List[String]]$al_flags_codes_lst       = $tbl_al_flags_codes_ary
	[Collections.Generic.List[String]]$al_hasjpgs_lst           = $tbl_al_hasjpgs_ary
	[Collections.Generic.List[String]]$al_hasdats_lst           = $tbl_al_hasdats_ary
	[Collections.Generic.List[String]]$al_haszerodates_lst      = $tbl_al_haszerodates_ary
	[Collections.Generic.List[String]]$al_attribs_lst           = $tbl_al_attribs_ary
	[Collections.Generic.List[String]]$al_memos_lst             = $tbl_al_memos_ary
	[Collections.Generic.List[String]]$jpg_exif_memos_match_lst = $tbl_jpg_exif_memos_match_ary
	[Collections.Generic.List[String]]$dat_memos_lst            = $tbl_dat_memos_ary
	[Collections.Generic.List[String]]$dat_memo_image_nos_lst   = $tbl_dat_memo_image_nos_ary
	[Collections.Generic.List[String]]$al_issues_lst            = $tbl_al_issues_ary

	$removed_indices_ary = @()
	$kept_indices_ary = @()

	$i = 0
	foreach ($item in $tbl_al_fnames_ary) {
		
		$item_clean = $item.Replace('.jpg','')
		
		$path_key = $tbl_al_paths_ary[$i]  # like, '@123456.bvr'

		$keep_flg = $false
		
		$test = ( ($tbl_al_hasjpgs_ary[$i].Length -eq 0) -and ($tbl_al_hasdats_ary[$i].Length -eq 0) )
		if (($FilterNoBoth -or $FilterIssues) -and $test) {
			$keep_flg = $true
		}

		if (($FilterNoJPG -or $FilterIssues) -and $tbl_al_hasjpgs_ary[$i].Length -eq 0) {
			$keep_flg = $true
		}

		if (($FilterNoDAT -or $FilterIssues) -and $tbl_al_hasdats_ary[$i].Length -eq 0) {
			$keep_flg = $true
		}

		$test1 = ($tbl_al_hasjpgs_ary[$i].Length -gt 0 -and !($tbl_al_attribs_ary[$i] -match "J"))
		$test2 = ($tbl_al_hasdats_ary[$i].Length -gt 0 -and !($tbl_al_attribs_ary[$i] -match "D"))
		if (($FilterNoRO -or $FilterIssues) -and ($test1 -or $test2)) {
			$keep_flg = $true
		}

		$test1 = ($tbl_jpg_exif_memos_match_ary[$i] -match "match!")
		$test2 = ($tbl_jpg_exif_memos_match_ary[$i].StartsWith('<'))
		if (($FilterNoMatch -or $FilterIssues) -and !$test1 -and !$test2) {
			$keep_flg = $true
		}

		$test1 = ($tbl_al_memos_ary[$i].TrimStart($custom_memo_mark).Length -eq 0)
		$test2 = ($tbl_al_memos_ary[$i].TrimStart($custom_memo_mark) -eq 'NA')
		
		# For debugging
		#Write-Host ("tbl_al_memos_ary[$i] = '" + $tbl_al_memos_ary[$i] + "'")
		#Write-Host "FilterNoMemo test1 = $test1"
		#Write-Host "FilterNoMemo test2 = $test2"
		
		if (($FilterNoMemo -or $FilterIssues) -and ($test1 -or $test2)) {
			$keep_flg = $true
		}

		if (($FilterZeroDates -or $FilterIssues) -and $tbl_al_haszerodates_ary[$i].Length -gt 0) {
			$keep_flg = $true
		}
		
		$idx = $al_fnames_lst.IndexOf($item)

		if (!$keep_flg) {
			$removed_indices_ary += $idx
			
			# For debugging
			#Write-Host ("removing alert at idx $i - " + $item_clean)
			
		} else {
			$kept_indices_ary += $idx		
		}
	
		$i++
	}
	
	# For debugging
	#Write-Host "removed_indices_ary = $removed_indices_ary"
	#Write-Host "kept_indices_ary = $kept_indices_ary"
	
	if ($fromConsole -and !$Quiet) {
		$arr = @()
		if ($FilterIssues)    { $arr += "-FilterIssues" }
		if ($FilterNoJPG)     { $arr += "-FilterNoJPG" }
		if ($FilterNoDAT)     { $arr += "-FilterNoDAT" }
		if ($FilterNoBoth)    { $arr += "-FilterNoBoth" }
		if ($FilterNoRO)      { $arr += "-FilterNoRO" }
		if ($FilterNoMemo)    { $arr += "-FilterNoMemo" }
		if ($FilterNoMatch)   { $arr += "-FilterNoMatch" }
		if ($FilterZeroDates) { $arr += "-FilterZeroDates" }
		$str = $arr -Join ","
		Write-Host ("`nFilters applied ($str): " + $kept_indices_ary.Count + " protected Alerts remain")
	}
	
	# REMOVE ITEMS
	# ... must remove elements from the bottom up
	[Array]::Reverse($removed_indices_ary)
	foreach ($idx in $removed_indices_ary) {
		
		# For debugging
		#Write-Host "removing idx# $idx"
		
		$al_fnames_lst2.RemoveAt($idx)	
		$al_paths_lst.RemoveAt($idx)	
		$al_dates_lst.RemoveAt($idx)	
		$al_labels_lst.RemoveAt($idx)	
		$al_flags_lst.RemoveAt($idx)	
		$al_flags_codes_lst.RemoveAt($idx)	
		$al_hasjpgs_lst.RemoveAt($idx)	
		$al_hasdats_lst.RemoveAt($idx)	
		$al_haszerodates_lst.RemoveAt($idx)	
		$al_attribs_lst.RemoveAt($idx)	
		$al_memos_lst.RemoveAt($idx)	
		$jpg_exif_memos_match_lst.RemoveAt($idx)			
		$dat_memos_lst.RemoveAt($idx)			
		$dat_memo_image_nos_lst.RemoveAt($idx)			
		$al_issues_lst.RemoveAt($idx)			
	}
	
	$al_fnames_filtered_ary   = [Array]$al_fnames_lst2
	$tbl_al_paths_ary             = [Array]$al_paths_lst
	$tbl_al_dates_ary             = [Array]$al_dates_lst
	$tbl_al_labels_ary            = [Array]$al_labels_lst
	$tbl_al_flags_ary             = [Array]$al_flags_lst
	$tbl_al_flags_codes_ary       = [Array]$al_flags_codes_lst
	$tbl_al_hasjpgs_ary           = [Array]$al_hasjpgs_lst
	$tbl_al_hasdats_ary           = [Array]$al_hasdats_lst
	$tbl_al_haszerodates_ary      = [Array]$al_haszerodates_lst
	$tbl_al_attribs_ary           = [Array]$al_attribs_lst
	$tbl_al_memos_ary             = [Array]$al_memos_lst
	$tbl_jpg_exif_memos_match_ary = [Array]$jpg_exif_memos_match_lst
	$tbl_dat_memos_ary            = [Array]$dat_memos_lst
	$tbl_dat_memo_image_nos_ary   = [Array]$dat_memo_image_nos_lst
	$tbl_al_issues_ary            = [Array]$al_issues_lst

} else {
	$al_fnames_filtered_ary = $tbl_al_fnames_ary
}

# Final count of filtered, protected Alerts
$alerts_filtered_cnt = $al_fnames_filtered_ary.Count

# For debugging
#Write-Host "alerts_filtered_cnt = $alerts_filtered_cnt"
#Write-Host "al_fnames_filtered_ary = $al_fnames_filtered_ary"
#Write-Host ("tbl_al_issues_ary BEFORE = " + ($tbl_al_issues_ary | ConvertTo-Json))
#Write-Host ("tbl_al_hasjpgs_ary = " + ($tbl_al_hasjpgs_ary | ConvertTo-Json))
#Write-Host ("tbl_al_hasdats_ary = " + ($tbl_al_hasdats_ary | ConvertTo-Json))
#Write-Host ("tbl_al_haszerodates_ary = " + ($tbl_al_haszerodates_ary | ConvertTo-Json))
#Write-Host ("tbl_al_attribs_ary = " + ($tbl_al_attribs_ary | ConvertTo-Json))
#Write-Host ("tbl_al_attribs_ary = " + ($tbl_al_attribs_ary | ConvertTo-Json))

# HANDLE ARGUMENT -ClearMarks
if ($ClearMarks -and !$fix_type_arg_used_flg) {

		if ($fromConsole -and !$Quiet) {
			Write-Host ""
			Write-Host "Clearing previously applied custom memo marks ($custom_memo_mark) ..."
		}

		$progress_bar_cntr = 0
		$i = 0
		foreach ($item in $al_fnames_filtered_ary) {  # $item is like "GT.20231009_130000.1224078.5-0.jpg"

			# Show progress bar
			if ($fromConsole) {
				$progress_bar_cntr++
				$WriteProgressParams = @{
					Activity        = "Clearing custom memo marks..."
					Status          = "Progress: $progress_bar_cntr of $al_protected_cnt"
					PercentComplete = (($progress_bar_cntr / $al_protected_cnt) * 100)
				}
				Write-Progress @WriteProgressParams
			}
			
			$item_clean = $item.Replace('.jpg','')

			$path_key = $tbl_al_paths_ary[$i]  # like, '@123456'
						
			$db_al_path = $lookup_paths_obj[$path_key]
			$db_al_date_sec = $lookup_dates_obj[$path_key]  # var name is reminder of units
			$db_al_flags = $lookup_flags_obj[$path_key]  # array is Int[]
			$jpg_filespec = "$bi_alerts_folder_path\$item"

			$db_al_memo = [String]$lookup_memos_obj[$path_key]  # Force to string (could be a number)
			if ($db_al_memo.StartsWith($custom_memo_mark)) {
				$new_memo = $db_al_memo.TrimStart($custom_memo_mark)
			} else {
				$new_memo = $db_al_memo
			}

			# Update the alert's database record
			if (Test-Path -LiteralPath "$jpg_filespec") {
				# ... returns "success' or 'fail'
				$result_str = `
					UpdateDbRecord `
					-AlertDbPath $db_al_path `
					-AlertDbDateSec $db_al_date_sec `
					-AlertDbFlags $db_al_flags `
					-AlertDbMemo $new_memo `
					-UpdateJpgExif `
					-JpgFileSpec $jpg_filespec `
					-FromConsoleFlg $fromConsole `
					-QuietFlg $Quiet `
					-DevModeFlg $DevMode
					
			} else {
				
				$result_str = `
					UpdateDbRecord `
					-AlertDbPath $db_al_path `
					-AlertDbDateSec $db_al_date_sec `
					-AlertDbFlags $db_al_flags `
					-AlertDbMemo $new_memo `
					-FromConsoleFlg $fromConsole `
					-QuietFlg $Quiet `
					-DevModeFlg $DevMode
				
			}
			
			if ($result_str -match 'success') {
				
				# Update lookup object to ensure this change is recognized by subsequent actions
				$lookup_memos_obj[$path_key] = $new_memo
				
				# Update output arrays
				$tbl_al_memos_ary[$i] = $new_memo
				$tbl_jpg_exif_memos_match_ary[$i] = "match!"
				
			} else {
				if ($fromConsole) {
					Write-Host "$item_clean - (JPG) Failed to update the Exif metadata" -ForegroundColor Yellow
				}
			}
			
			$i++
		}
}

# ===========================================================================
# HANDLE 'FIX' ARGUMENTS -FixJpgExif, -FixFiles, -FixReadOnly, -FixMemos, -FixZeroDates, -SetFlag
if (1 -eq 1) {  # Use to collapse this lengthy code section while editing

	# Save a copy b/c we're going to update $tbl_al_issues_ary
	$al_issues_before_ary = $tbl_al_issues_ary

	# ... and modify blank entries for later use in Blue Iris logfile messages
	$i = 0
	foreach ($item in $al_issues_before_ary) {
		if ($item -eq '') {
			$al_issues_before_ary[$i] = '--'
		}
		$i++
	}

	# Initialize collector to track if Alerts with issues that have been fixed
	# values are 'Y', 'N', 'Partial' or ''
	$al_issues_fixed_ary = @()

	if ($alerts_filtered_cnt -gt 0) {

		if ($fix_type_arg_used_flg -or $SetFlag) { 

			if ($fromConsole -and !$Quiet) {
				Write-Host ""
				Write-Host "Repair actions to be attempted..."
				if ($FixZeroDates -or $FixAll) {
					Write-Host "* Fix Alert database records that have a zero-date (via DAT or JPG dates)"
					#Write-Host "  Note: if needed, will create a replacement JPG from Alert's BVR clip (if it exists)"
				}
				if ($FixMemos -or $FixAll) {
					Write-Host "* Fix Alert database records that have a blank '' or 'NA' memo"
					Write-Host "  Note: tries to use 'memo' string from DAT file's confirmed image (if it exists)"
				}
				if ($FixFiles -or $FixAll) {
					Write-Host "* Create replacements for missing JPGs; in this order"
					Write-Host "* in this order - from DAT file, then BVR clip, then HTTP API cmd '/alert/@record'"
					Write-Host "  NOTE missing DAT files cannot be created"
				}
				if ($FixJpgExif -or $FixAll) {
					Write-Host "* Fix Exif metadata (Alert JPGs) to match the database memo"
				}
				if ($FixReadOnly -or $FixAll) {
					Write-Host "* Set read-only file attributes for JPG & DAT files (if needed)"
				}
				if ($SetFlag) {
					Write-Host "* Edit Alert's database record to add the 'Flag' flag"
				} 
				Write-Host ""
				Write-Host "Results..."
			}
			
			# Track if we've added a file b/c this will require Database/Regenerate action
			$added_file_flg = $false
				
			# Reinitialize output collectors
			[String[]]$tbl_al_issues_ary = @()

			# Reinitialize output csv file collectors
			[String[]]$csv_fixed_paths_ary = @()
			[String[]]$csv_fixed_labels_ary = @()
			#[String[]]$csv_fixed_flags_ary = @()
			[String[]]$csv_fixed_issues_ary = @()
				
			$i = 0
			foreach ($item in $al_fnames_filtered_ary) {  # $item is like "GT.20231009_130000.1224078.5-0.jpg"

				# For debugging
				Write-Host ("`n----- processing '" + $item.Replace('.jpg', '') + "' -----")

				# Accumulator for this alert's fix results 
				# (usually all true values; any false values indicate a fix failure)
				$fixed_results_ary = [boolean[]]@()

				$item_clean = $item.Replace('.jpg','')

				$path_key = $tbl_al_paths_ary[$i]  # like, '@123456'
				
				if ($path_key -eq '@0.jpg') {
				
					$file_val = $alertlist_files_ary[$i]
					$label_val = $file_val.Replace('.jpg', '')
					
					$fixed_results_ary += $false
					$tbl_al_issues_ary += "unfixable alert (path=@0.jpg)"

					if ($fromConsole) {
						Write-Host "$label_val - db record is unfixable! (alertlist 'path = @0.jpg')" -ForegroundColor Yellow
					}

					$i++
					continue
				}
				
				# INITIALIZE VARS USED IN 'FIX' SECTIONS BELOW

				# This Alert's params
				$db_al_path  = $lookup_paths_obj[$path_key]
				$db_al_date  = $lookup_dates_obj[$path_key]  # array is Int[]
				$db_al_flags = $lookup_flags_obj[$path_key]  # array is Int[]
				$db_al_memo  = [String]$lookup_memos_obj[$path_key]  # Force to string (could be a number)
				$db_al_flags_orig = $lookup_flags_obj[$path_key]  # backup copy
				
				# For debugging
				#Write-Host "item = $item"
				#Write-Host "path_key = $path_key"
				#Write-Host "`n----- processing path $db_al_path -----"
				#Write-Host ("db_al_path type = " + $db_al_path.GetType())
				#Write-Host "db_al_path = $db_al_path"
				#Write-Host ("db_al_flags type = " + $db_al_flags.GetType())
				#Write-Host "db_al_flags = $db_al_flags"
				#Write-Host "db_al_date = $db_al_date"
				#Write-Host "db_al_memo = '$db_al_memo'"
				#WZrite-Host ("db_al_flags_orig type = " + $db_al_flags_orig.GetType())
				##Write-Host "db_al_flags_orig = $db_al_flags_orig"

				# Make JPG filespec and test existence
				$jpg_filespec = $bi_alerts_folder_path + "\" + $item
				$jpg_file_exists = (Test-Path -LiteralPath "$jpg_filespec")

				# Use JPG filespec to make DAT filespec and test existence
				$dat_filespec = $jpg_filespec.Replace('.jpg','.dat')
				$dat_file_exists = (Test-Path -LiteralPath "$dat_filespec")

				# For debugging
				#Write-Host "dat_filespec = $dat_filespec"
				#Write-Host "dat_file_exists = $dat_file_exists"
				#Write-Host ("DAT CreationTime = " + (Get-Item "$dat_filespec").CreationTime)
				#Write-Host "jpg_filespec = $jpg_filespec"
				#Write-Host "jpg_file_exists = $jpg_file_exists"
				#Write-Host ("JPG CreationTime = " + (Get-Item "$jpg_filespec").CreationTime)

				# JPG files are backed up before making changes to their EXIF metadata
				# ... this variable avoids backing up the file more once this pass
				$orig_jpg_backed_up = $false
				
				# Initialize flag for successful or unfixable zero-date fixes
				$zero_date_fixed_flg = $false
				$zero_date_is_unfixable = $false
				
				# Initialize zero-date-fix flag to attempt JPG file creation if only DAT file exists
				# Key point: We wait do this at the end of all fixes, in case the JPG file must be created first
				$try_zero_date_make_jpg_flg = $false
				
				# Initialize increment seconds for zero-date fixes
				$zd_al_date_fixed_sec = 0
				
				# Initialize database parameters that may be changed via subsequent JSON 'update' cmds
				# ... these are used multiple times via function UpdateDbRecord()
				$new_memo = $db_al_memo
				$new_date_sec = $db_al_date

				# Make date string for replacement/dummy JPG files
				$db_al_date_sec = $db_al_date  # var name is reminder of units
				$db_al_dt_str = Convert-FromUnixDate $db_al_date_sec

				# For testing
				#$db_al_memo = ''
				#$db_al_memo = 'NA'
				#$db_al_memo = '^NA'
				#$FixZeroDates = $true

				# REPAIR ALERTS WITH ZERO DATES
				if ($FixZeroDates -or $FixAll) {
#Write-Host "made it - fixzerodate0"					
					# If using Blue Iris 5.8.4.5+ ...
					# we can fix the zero-date DIRECTLY using the JSOn 'update' cmd.
					# - Repair requires a date from DAT or JPG file. 
					# - The order of preference is 1) DAT file, 2) JPG file.
					
					# If not using Blue Iris 5.8.4.5+ ...
					# we can ONLY HOPE that a the script's db repair/regenerate (or future delete/regenerate) 
					# will pick up the missing date value via the JPG file's 'taken' date (or creation date if lacking)
					# - In this case the script will attempt to create a JPG, event if arg -FixFiles is not used

					if ($db_al_date_sec -eq 0) {
#Write-Host "made it - fixzerodate1"					

						# BI's JSON 'update' cmd did not support the 'date' param until update 5.8.4.5
						if (IsBiVersion '5.8.4.5') {
#Write-Host "made it - fixzerodate2"					

							# TRY TO GET A NEW DATABASE DATE
							# ... Requires a DAT or JPG file
							
							if ($dat_file_exists -or $jpg_file_exists) {
#Write-Host "made it - fixzerodate3"			
								$try_zero_date_make_jpg_flg = ($dat_file_exists -and !$jpg_file_exists)

								# Preference is given to the DAT file, if it exists
								if ($dat_file_exists) {
#Write-Host "made it - fixzerodate4"					
									
									$datefix_file_type = 'DAT'
									
									# Get DAT file's earliest date attribute; like, '01/08/2024 20:10:42'
									if ((Get-Item "$dat_filespec").CreationTime -le (Get-Item "$dat_filespec").LastWriteTime) {
										$target_date_str = (Get-Item "$dat_filespec").CreationTime
									} else {
										$target_date_str = (Get-Item "$dat_filespec").LastWriteTime
									}
									
									# For debugging
									#Write-Host "dat_filespec = $dat_filespec"
									#Write-Host "dat_filespec actual CreationTime = $target_date_str"

								} else {
#Write-Host "made it - fixzerodate5"					
									
									$datefix_file_type = 'JPG'

									# Get JPG file's DateTaken; like, '01/08/2024 11:10 PM'
									$result_str = `
										GetJpgFileExifDateTaken `
										-FilePath $jpg_filespec
																			
									if ($result_str -ne 'fail') {
										$target_date_str = $result_str
										$jpg_datefix_type = 'DateTaken'
									} else {
										# Get JPG file's CreationTime; like, '01/08/2024 20:10:42'								
										$target_date_str = (Get-Item "$dat_filespec").CreationTime
										$jpg_datefix_type = 'CreationTime'
									}

									# For debugging
									#Write-Host "jpg_filespec = $jpg_filespec"
									#Write-Host "GetJpgFileExifDateTaken() result_str - $result_str"
									#Write-Host "jpg_filespec actual CreationTime = $target_date_str"

								}

								# Make target Unix timestamp (seconds)
								$target_date_sec = Convert-ToUnixDate $target_date_str
								
								# Make the new date to be used to update the database record
								$new_date_sec = $target_date_sec
								
								# HANDLE SIMULATOR ARG -SIM6 (ZERO DATE)
								if ($Sim6_arg_exists) {
									$new_date_sec += $Sim6_adj_sec
									$lookup_zdfixes_obj.Add($path_key, $new_date_sec)
								} else {
									$lookup_zdfixes_obj.Add($path_key, 0)
								}

								# Make the new datetime string (used for feedback)
								$new_date_str = Convert-FromUnixDate $new_date_sec

								# For debugging
								#Write-Host "Alert's unix timestamp = $db_al_date_sec"
								#Write-Host "File used to get target timestamp = $datefix_file_type"
								#Write-Host ("$datefix_file_type file's unix timestamp = $target_date_sec")
								#Write-Host ("Alert's target unix timestamp + $Sim6 (Sim6) = $new_date_sec")
								#Write-Host "Alert's target_date_str = $new_date_str"
								$target_date_rev_str = Convert-FromUnixDate $new_date_sec
								#Write-Host ("Alert's target unix timestamp to dt = $target_date_rev_str (just checking!)")

								# TRY TO UPDATE THE ALERT'S DATABASE RECORD WITH GOOD DATE

								if (Test-Path -LiteralPath "$jpg_filespec") {
#Write-Host "made it - fixzerodate6"					
									
									# Returns "success' or 'fail'
									$result_str = `
										UpdateDbRecord `
										-AlertDbPath $db_al_path `
										-AlertDbDateSec $new_date_sec `
										-AlertDbFlags $db_al_flags `
										-AlertDbMemo $db_al_memo `
										-UpdateJpgExif `
										-JpgFileSpec $jpg_filespec `
										-FromConsoleFlg $fromConsole `
										-QuietFlg $Quiet `
										-DevModeFlg $DevMode
										
								} else {
#Write-Host "made it - fixzerodate7"					

									$result_str = `
										UpdateDbRecord `
										-AlertDbPath $db_al_path `
										-AlertDbDateSec $new_date_sec `
										-AlertDbFlags $db_al_flags `
										-AlertDbMemo $db_al_memo `
										-FromConsoleFlg $fromConsole `
										-QuietFlg $Quiet `
										-DevModeFlg $DevMode
									
								}

								# For debugging
								#Write-Host "result_str = '$result_str'"

								if ($result_str -match 'success') {
#Write-Host "made it - fixzerodate8"					

									$zero_date_fixed_flg = $true

									# NOTE: Aborted trying validate the JPG file's new datetaken change
									# Why? the JPG's Exif 'DateTaken' value does not include seconds 
									#      (e.g., '8:33 AM' instead of '8:33:28 AM)'
									#      which makes UNIX timestamp comparisons unnecessarily complicated

									# Update issues collectors
									$fixed_results_ary += $true
									$lookup_haszerodates_obj[$path_key] = $false
									$lookup_issues_obj[$path_key] = `
										RemoveArrayItems `
										-Haystack $lookup_issues_obj[$path_key] `
										-Needles @('Zero-date')

									# Update output collectors
									$tbl_al_haszerodates_ary[$i] = ''
								
									# Add to output csv file collectors
									if (!($csv_fixed_labels_ary -contains $item_clean)) {
										$csv_fixed_labels_ary += $item_clean
										$csv_fixed_paths_ary += $path_key
									}
									
									if ($fromConsole -and !$Quiet) {
										Write-Host "$item_clean - (DB) zero-date changed to '$new_date_str' (via $datefix_file_type)"
									}
									
								} else {
#Write-Host "made it - fixzerodate9"					

									# Update issues collectors
									$fixed_results_ary += $false
									#$lookup_issues_obj[$path_key] += "Zero-date Alert not fixed; bad JPG Exif datetaken"
									$lookup_issues_obj[$path_key] += "Zero-date not fixed"
									$lookup_issues_obj[$path_key] = `
										RemoveArrayItems `
										-Haystack $lookup_issues_obj[$path_key] `
										-Needles @('Zero-date')

									if ($fromConsole) {
										Write-Host "$item_clean - (DB) unable to fix zero-date via JSON 'update' cmd; tried '$new_date_str' (via $datefix_file_type)" -ForegroundColor Yellow
									}
								}

							} else {
#Write-Host "made it - fixzerodate10"	

								$zero_date_is_unfixable = $true
								$try_zero_date_make_jpg_flg = $false
								
								# Update issues collectors
								$fixed_results_ary += $false
								#$lookup_issues_obj[$path_key] += "Zero-date Alert not fixed; requires DAT or JPG file"
								$lookup_issues_obj[$path_key] += "Zero-date not fixed"
								$lookup_issues_obj[$path_key] = `
									RemoveArrayItems `
									-Haystack $lookup_issues_obj[$path_key] `
									-Needles @('Zero-date')
								
								if (!$delete_unfixable_zerodate_alerts) {
#Write-Host "made it - fixzerodate11"					
									
									if ($fromConsole) {
										Write-Host "$item_clean - (DB) unable to fix zero-date; requires an existing DAT or JPG file" -ForegroundColor Yellow
									}

								} else {
#Write-Host "made it - fixzerodate12"					

									# DELETE unfixable zero-date Alert from database (if requested)
									
									# BUT FIRST, WE MUST UNPROTECT THE ALERT'S DATABASE RECORD 
									
									$db_al_flags_orig = $lookup_flags_obj[$path_key]   # array is Int[]
									$flags_bin_ary = $lookup_flags_bin_obj[$path_key]  # like ['@123456': [0,0,1,0,...]
									$flags_bin_ary_new = $flags_bin_ary
									
									# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
									if ($flags_bin_ary[2] -ne 0) {
										
										# THE ALERT IS PROTECTED!
										# ... so we need to unprotect it now
										$flags_bin_ary_new[2] = 0										
										$db_al_flags_new = `
											MakeFlagsDecimalValue `
											-BinAry $flags_bin_ary_new
									
										# For debugging
										#Write-Host "db_al_flags_new = $db_al_flags_new"
									
										# Update the Alert's flags; returns boolean
										$result_flg = `
											SetAlertFlags `
											-AlertDbPath $db_al_path `
											-AlertDbMemo $db_al_memo `
											-AlertDbFlags $db_al_flags_new `
											-QuietFlg $Quiet `
											-FromConsoleFlg $fromConsole
										
										if (!$result_flg) {
#Write-Host "made it - fixzerodate13"					

											# Update issues collector
											$lookup_issues_obj[$path_key] += "Alert's 'Flag' flag not changed"
											Write-Host "$item_clean - (DB) Unsuccessful attempt to change the Alert's 'Flag' flag" -ForegroundColor Yellow

										} else {
#Write-Host "made it - fixzerodate14"					

											# DELETE THE ALERT FROM THE BLUE IRIS DATABASE
											$jsonCmd = @{
												cmd = 'delalert'
												path = $db_al_path
											}
											# Returns a powershell PSCustomObject
											$json_response_obj = `
												BIJsonCall `
												-JsonCmd $jsonCmd `
												-QuietFlg $false `
												-FromConsoleFlg $fromConsole

											# For debugging
											#Write-Host ("json_response_obj = " + ($json_response_obj | ConvertTo-Json))
											
											if ($json_response_obj -match 'ERR/') {
												if ($fromConsole) {
													Write-Host "$item_clean - (DB) unable to delete DB record for this unfixable zero-date Alert" -ForegroundColor Yellow
												}
											} elseif ($json_response_obj.result -eq 'fail') {
												if ($fromConsole) {
													Write-Host "$item_clean - (DB) unable to delete DB record for this unfixable zero-date Alert" -ForegroundColor Yellow
												}
											} else {
#Write-Host "made it - fixzerodate15"					
												# Update issues collector
												$fixed_results_ary += $true
												$lookup_issues_obj[$path_key] += "Zero-date Alert deleted (unfixable)"
												
												if ($fromConsole) {
													Write-Host "$item_clean - (DB) deleted DB record for this unfixable zero-date Alert (lacks a DAT or JPG file)" -ForegroundColor Yellow
												}
											}						
										}
									}
								}
							}
					
						} else {							
#Write-Host "made it - fixzerodate16"					
							# Update issues collectors
							$fixed_results_ary += $false
							#$lookup_issues_obj[$path_key] += "Zero-date not fixed; requires BI 5.8.4.5+"
							$lookup_issues_obj[$path_key] += "Zero-date not fixed"
							if ($fromConsole) {
								Write-Host "$item_clean - (DB) unable to fix zero-date record; requires Blue Iris 5.8.4.5+" -ForegroundColor Yellow
							}
						}
					}
					
					# For debugging
					#Write-Host "zero_date_is_unfixable = $zero_date_is_unfixable"
					#Write-Host "try_zero_date_make_jpg_flg = $try_zero_date_make_jpg_flg"
					#Write-Host ("lookup_zdfixes_obj = " + ($lookup_zdfixes_obj | ConvertTo-Json))
					
				}

				# CREATE REPLACEMENT/DUMMY JPG
				# ... Three methods tried: DAT file extraction, Clip file extraction, then HTTP cmd
				#     if all of these fail, a yellow JPG is created to hold the Exif metadata
				if (($FixFiles -or $FixAll) -and !$zero_date_is_unfixable) {
#Write-Host "made it - fixfiles1"

					# For debugging
					#Write-Host "tbl_al_labels_ary = $tbl_al_labels_ary"
					#Write-Host "tbl_al_hasjpgs_ary = $tbl_al_hasjpgs_ary"
					#Write-Host "tbl_al_hasdats_ary = $tbl_al_hasdats_ary"

					# Initialize collectors
					$cur_attribs_str = $tbl_al_attribs_ary[$i]  # like '', 'J', 'D', 'JD'
					$cur_attribs_ary = $cur_attribs_str -Split ''
					$tmp_attribs_ary = @()

					# Abort if there are no files to process
					if ($tbl_al_hasjpgs_ary[$i].Length -eq 0 -or $tbl_al_hasdats_ary[$i].Length -eq 0) {
#Write-Host "made it - fixfiles2"
						
						# For debugging
						#Write-Host ("jpg_filespec = '" + $jpg_filespec + "'")
						#Write-Host ("dat_filespec = '" + $dat_filespec + "'")

						if (!$jpg_file_exists) {
#Write-Host "made it - fixfiles3"
							
							# Initialize vars used by all methods
							$dummy_flg = $false
							$method_str = ''
							$overlay_source_fs = ''
							$found_dat_conf_img_flg = $false
							$orig_jpg_backed_up = $true  # diabled b/c there is no file to backup
							
							$timer_dummyjpg_runtime = [Diagnostics.Stopwatch]::StartNew()

							# TRY TO USE DAT FILE
							if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
								if (!$dummy_flg) {
									if ($dat_file_exists) {
#Write-Host "made it - fixfiles3 (try DAT1)"

										# Provide feedback
										if ($DevMode) {
											if ($fromConsole) {
												Write-Host "(dev mode) Extracting JPGs from DAT file..." -ForegroundColor Cyan
											}
										}
										
										# Get the DAT file's confirmed image number
										$confirmed_image_num = $lookup_dat_confimg_num_obj[$path_key]
										$confirmed_image_memo_str = $lookup_dat_confimg_memo_obj[$path_key]

										# For debugging
										#Write-Host ("confirmed_image_num type = " + $confirmed_image_num.GetType())
										#Write-Host "confirmed_image_num = $confirmed_image_num"
										#Write-Host ("confirmed_image_memo_str type = " + $confirmed_image_memo_str.GetType())
										#Write-Host "confirmed_image_memo_str = $confirmed_image_memo_str"
										
										if ($confirmed_image_num -ne 0 -and $confirmed_image_memo_str.Length -ne 0) {
#Write-Host "made it - fixfiles3 (try DAT2)"
											
											$found_dat_conf_img_flg = $true

											# For debugging
											#Write-Host "using DAT confirmed img num = $confirmed_image_num"

											# Extract just the confirmed JPG
											$extract_obj = `
												ExtractDatImages `
												-AlertFile $item `
												-AlertPath $bi_alerts_folder_path `
												-DestPath $dummyjpgs_path `
												-DatImageNo $confirmed_image_num `
												-DatReqdImages 1 `
												-QuietFlg $Quiet `
												-FromConsoleFlg $fromConsole `
												-DevModeFlg $DevMode
												
										} else {
#Write-Host "made it - fixfiles3 (try DAT3)"
											
											# Extract all the JPGs
											$extract_obj = `
												ExtractDatImages `
												-AlertFile $item `
												-AlertPath $bi_alerts_folder_path `
												-DestPath $dummyjpgs_path `
												-DatReqdImages 1 `
												-QuietFlg $Quiet `
												-FromConsoleFlg $fromConsole `
												-DevModeFlg $DevMode
										}
										
										# For debugging
										#Write-Host ("ExtractDatImages extract_obj type = " + $extract_obj.GetType())
										#Write-Host ("ExtractDatImages extract_obj = " + ($extract_obj | ConvertTo-Json))
										
										# For testing
										#$extract_obj = $null
										#$extract_obj = "ERR/ test"
										#$extract_obj = "FAIL: test"

										if ($extract_obj) {
#Write-Host "made it - fixfiles3 (try DAT4)"

											if (($extract_obj -match 'ERR/') -or ($extract_obj -match 'FAIL:')) {

												if ($fromConsole) {
													Write-Warning "$extract_obj".Replace('FAIL: ','')
												}
												
											} elseif ($extract_obj.Count -gt 0) {
#Write-Host "made it - fixfiles3 (try DAT5)"

												$method_str = ' (via DAT)'

												# MAKE THE EXTRACTED IMAGE FILESPEC
												# ... to be used in the in AddTextToImage()

												if ($found_dat_conf_img_flg) {
													$target_dat_image_no = $confirmed_image_num
												} else {
													# Identify the middle DAT image 
													# ... e.g., if 5 images use image #3, if 6 images, use image #4
													$dat_jpgs_cnt = ($extract_obj['data']).Count
													$target_dat_image_no = [Math]::Ceiling($dat_jpgs_cnt/2)
													
													# For debugging
													#Write-Host "dat_jpgs_cnt = $dat_jpgs_cnt"
												}

												# For debugging
												#Write-Host "target_dat_image_no = $target_dat_image_no"
												
												$img_name = "img"
												if ($target_dat_image_no -gt 0) {
													if ($target_dat_image_no -ge 10) {
														$img_name += $target_dat_image_no.ToString()
													} else {
														$img_name += "0" + $target_dat_image_no.ToString()
													}
												}

												$overlay_source_fs = "$dummyjpgs_path\$img_name.jpg"
												
												# For debugging
												#Write-Host "overlay_source_fs = $overlay_source_fs"

												if (Test-Path -LiteralPath "$overlay_source_fs") {
#Write-Host "made it - fixfiles3 (try DAT6)"

													$dummy_flg = $true
													
													# Provide feedback
													if ($DevMode -and $fromConsole) {
														Write-Host "(dev mode) JPG ready ($overlay_source_fs)" -ForegroundColor Cyan
													}
													
												} else {
#Write-Host "made it - fixfiles3 (try DAT6)"
													if ($fromConsole) {
														Write-Warning "no images extracted from DAT file"
													}
												}

											} else {
												if ($fromConsole) {
													Write-Warning "no images extracted from DAT file"
												}
											}
										} else {
											if ($fromConsole) {
												Write-Warning "DAT extraction failed; function ExtractDatImages() returned null"
											}
										}
									}
								}
							}

							# For testing
							#$dummy_flg = $false
							
							# TRY TO USE A BVR CLIP
							if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
								if (!$dummy_flg) {  
#Write-Host "made it - fixfiles3 (try BVR1)"
								
									# Get the BVR filename
									$clip_stats_obj = `
										BIJsonClipstats `
										-Clips $path_key `
										-Props 'camera,path,file,offset,msec,alertmsec,res' `
										-FromConsoleFlg $DevMode `
										-QuietFlg $Quiet 

									# Expecting response_obj to be a hashtable, like ... 
									#[
									#	{
									#		"result": "success"
									#	},
									#	{
									#		"@152866668.bvr": {
									#			"camera": "DW1",
									#			"path": "@152720932.bvr",
									#			"file": "DW1.20230530_060000.1965063.17-0.jpg",
									#			"offset": 2016632,
									#			"msec": 3599166,
									#			"alertmsec": 27426,
									#			"res": "2688x1520"
									#		}
									#	}
									#]

									if ($clip_stats_obj -is [String]) {
#Write-Host "made it - fixfiles3 (try BVR2)"
										if ($clip_stats_obj.StartsWith("ERR/")) {
#Write-Host "made it - fixfiles3 (try BVR3)"
											$err_msg = $clip_stats_obj
											return $err_msg
										}
									}
									
									if ($DevMode) {
										if ($fromConsole) {
											Write-Host "(dev mode) Clipstats:" -ForegroundColor Cyan
											Write-Host ($clip_stats_obj[1] | ConvertTo-Json) -ForegroundColor Cyan
										}
									}
									
									# For debugging
									#Write-Host ("response_obj type = " + $clip_stats_obj.GetType())
									#Write-Host ("response_obj = " + ($clip_stats_obj | ConvertTo-Json))

									$clipstats_result = $clip_stats_obj[0]  # A single element array
									$clipstats_data = $clip_stats_obj[1]  # An array of arrays
									$clipstats_success_flg = ($clipstats_result -match "result=success")
									$al_path = $clipstats_data[$path_key]['path']
									$al_file = $clipstats_data[$path_key]['file']
									$al_dur = $clipstats_data[$path_key]['alertmsec']
									$al_offset = $clipstats_data[$path_key]['offset']
									
									$al_file_ary = $al_file.Split('.')
									$bvr_filename = $al_file_ary[0] + "." + $al_file_ary[1] + ".bvr"

									# Handle multiple New folders
									$bvr_file_exists = $false
									foreach ($new_path in $bi_new_folder_paths_ary) {
										$bvr_filespec = $new_path + "\" + $bvr_filename
										if (Test-Path -LiteralPath "$bvr_filespec") {
											$bvr_file_exists = $true
											break
										}
									}
									
									# For debugging
									#Write-Host "clipstats_result = $clipstats_result"
									#Write-Host ("clipstats_data = " + ($clipstats_data | ConvertTo-Json))
									#Write-Host ("al_path = '" + $al_path + "'")
									#Write-Host ("al_dur = $al_dur")
									#Write-Host ("al_file type = " + $al_file.GetType())
									#Write-Host ("al_file = $al_file")
									#Write-Host ("al_file_ary = " + ($al_file_ary | ConvertTo-Json))
									#Write-Host ("bvr_filename = '" + $bvr_filename + "'")
																
									if ($bvr_file_exists) {
#Write-Host "made it - fixfiles3 (try BVR4)"

										$result_str = `
											ExtractClipImage `
											-BvrFile $bvr_filename `
											-Offset $al_offset `
											-JpgImagesCnt 10 `
											-JpgImageNo $bvr_jpg_image_no `
											-DelayMs $al_dur `
											-DelayPct 100 `
											-Height $bvr_jpg_height `
											-DestPath $dummyjpgs_path `
											-Quiet $Quiet `
											-FromConsoleFlg $fromConsole `
											-DevModeFlg $DevMode
									
										# For debugging
										#Write-Host "ExtractClipImage() result_str = $result_str"

										if ($result_str -eq 'success') {
#Write-Host "made it - fixfiles3 (try BVR5)"

											$dummy_flg = $true
											$jpg_file_exists = $true
											$method_str = ' (via Clip)'

											# Prep image overlay vars
											$filename_sfx = ([String]($bvr_jpg_image_no)).PadLeft(2,'0')
											$overlay_source_fs = $dummyjpgs_path + "\img" + $filename_sfx + ".jpg"
										
										}
										
									} else {
										if ($fromConsole) {
											Write-Warning "BVR file does not exist ($bvr_filespec)"
										}
									}
								}
							}

							# For testing
							#$dummy_flg = $false
							
							# TRY TO USE HTTP CMD - '/alerts/@record' (to get thumbnail image from teh BI database)
							if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
								if (!$dummy_flg) {  
#Write-Host "made it - fixfiles3 (try HTTP1)"

									$dummy_filespec = $dummyjpgs_path + "\http_img.jpg"

									# Delete exiting dummy file
									if (Test-Path -LiteralPath "$dummy_filespec") {
										Remove-Item -LiteralPath "$dummy_filespec" -Force
									}
					
									# Note: We don't bother w/ the '?fulljpeg' option because the JPG would already exist
									# $uri = "/alerts/" + $db_al_path.Replace('.bvr', '') + "?fulljpeg"
									$uri = "/alerts/" + $db_al_path.Replace('.bvr', '')

									# Always returns 'success', so need to test of file actually created
									$result_str = `
										BIHttpCall `
										-HttpCmd $uri `
										-DestPath "$dummy_filespec"

									# For debugging
									#Write-Host "BIHttpCall() result_str = $result_str"

									# Wait a bit for the system file structure to be updated
									#Start-Sleep -Milliseconds 300				
									
									if ($result_str -match 'success' -and (Test-Path -LiteralPath "$dummy_filespec")) {
#Write-Host "made it - fixfiles3 (try HTTP2)"
										$method_str = ' (via HTTP)'
										$dummy_flg = $true
										$overlay_source_fs = $dummy_filespec
									}
								}
							}

							# For testing
							#$dummy_flg = $false

							# MAKE A PLAIN SOLID COLOR JPG
							if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
								if (!$dummy_flg) {
#Write-Host "made it - fixfiles3 (try yellow1)"
								
									# For debugging
									#Write-Host "making solid color JPG"
												
									$result = `
										MakeDummyJpgFile `
										-JpgFilespec $jpg_filespec `
										-Color 'Yellow' `
										-AddOverlayFlg $add_jpg_overlay `
										-FontName $jpg_overlay_font `
										-FontColor 'Black'
								
									if ($result -eq 'success') {
										$dummy_flg = $true
										$jpg_file_exists = $true
										$method_str = ' (yellow)'
										$overlay_source_fs = $jpg_filespec
									}
								}
							}
							
							# Stop performance timer
							$timer_dummyjpg_runtime.Stop()
							$elapsed_sec = ($timer_dummyjpg_runtime.elapsed.totalseconds).ToString("0.00")
							
							if($DevMode -and $fromConsole) {
								Write-Host "(dev mode) replacement/dummy MPG maketime = $elapsed_sec sec" -ForegroundColor Cyan
							}
							
							# Wait a bit for the system's file structure to be updated
							Start-Sleep -Milliseconds 500				

							# For debugging
							#Write-Host "dummy_flg = $dummy_flg"
							#Write-Host "method_str = $method_str"
							#Write-Host "overlay_source_fs = $overlay_source_fs"
							#Write-Host "jpg_filespec = $jpg_filespec"
							
							# FINISH THE DUMMY JPG 
							# ... 1) add overlay and/or copy file to BI Alerts folder
							#     2) update exif metadata
							#     3) set file datetimes
							#     4) make file read-only
							if (1 -eq 1) {  # Use to collapse this lengthy code section while editing

								# Verify success and existence
								if ($dummy_flg -and (Test-Path -LiteralPath "$overlay_source_fs")) {
#Write-Host "made it - fixfiles3 - finish jpg1"
									# For debugging
									#Invoke-Item $overlay_source_fs

									# Set flag to force Blue Iris DB/Regenerate action
									$added_file_flg = $true

									# Update output collector
									$tbl_al_hasjpgs_ary[$i] = 'x'

									# Show results
									if ($fromConsole -and !$Quiet) {
										if ($method_str -match 'yellow') {
											Write-Host ("$item_clean - (JPG) Successfully created dummy JPG file" + $method_str)
										} else {
											Write-Host ("$item_clean - (JPG) Successfully created replacement JPG file" + $method_str)
										}
									}						

									# KEY POINT: New JPGs lack Exif metadata
									# ... (the sole exception is HTTP cmd '/alerts/@record?fulljpeg',
									#      but we never use it, as noted above)
									
									# Update issues collectors
									$fixed_results_ary += $true
									$lookup_issues_obj[$path_key] = `
										RemoveArrayItems `
										-Haystack $lookup_issues_obj[$path_key] `
										-Needles @('No JPG file')
									
									# Add to output csv file collectors
									if (!($csv_fixed_labels_ary -contains $item_clean)) {
										$csv_fixed_labels_ary += $item_clean
										$csv_fixed_paths_ary += $path_key
									}

									# ADD OVERLAY AND/OR COPY JPG TO BLUE IRIS ALERTS FOLDER
									if ($add_jpg_overlay) {
											
										# For debugging
										#Write-Host "overlay_source_fs = $overlay_source_fs"
										#Write-Host "jpg_filespec = $jpg_filespec"

										# Save JPG WITH an overlay
										# KEY POINT: this function copies the JPG to the destination folder
										$result_str = `
											AddTextToImage `
											-SourcePath $overlay_source_fs `
											-DestPath $jpg_filespec `
											-Text1 "Replacement JPG" `
											-Text2 "(created by $ScriptName)" `
											-Location $jpg_text_loc `
											-FontName $jpg_overlay_font `
											-FontStyle 'Regular' `
											-FontSize1 '20' `
											-FontSize2 '16' `
											-ForeColor 'Yellow' `
											-BackColor 'Black' `
											-FromConsoleFlg $fromConsole
											
										if ($result_str -ne 'success') {
											if ($fromConsole ) {
												Write-Warning "replacement/dummy JPG created, but unable to add text overlay"
											}
												
											# For debugging
											#Write-Host "result_str = $result_str"
										}
									} else {
										# Save JPG WITHOUT an overlay
										Copy-Item -LiteralPath "$overlay_source_fs" -destination $("$jpg_filespec") -force
									}
									
									$jpg_file_exists = (Test-Path -LiteralPath $jpg_filespec)

								} else {
									if ($fromConsole) {
										Write-Host "$item_clean - Failed to create dummy JPG file" -ForegroundColor Yellow
									}						
								}

							}
							
						} else {
							if ($cur_attribs_str -match 'J') {
								$tmp_attribs_ary += 'J'
							}
						}
						
						# Update output collector
						if ($tmp_attribs_ary.Count -gt 0) {
							$cur_attribs_str_new = $tmp_attribs_ary -Join ''
							$tbl_al_attribs_ary[$i] = $cur_attribs_str_new
							
							# For debugging
							#Write-Host "updated cur_attribs_str = $cur_attribs_str_new"

						}
					}

				}

				# REPAIR MISSING OR MALFORMED DATABASE MEMOS (USE -FixFiles for JPG MEMOS)
				if (($FixMemos -or $FixAll) -and !$zero_date_is_unfixable) {
#Write-Host "made it - fixmemo1"

					$flags_bin_ary = $lookup_flags_bin_obj[$path_key]   # like ['@123456': [0,0,1,0,...]

					# For debugging
					#Write-Host "path_key = $path_key"
					#Write-Host "flags_bin_ary = $flags_bin_ary"
					
					$flags_bin_ary_new = $flags_bin_ary
					
					# Process AI-Confirmed Alerts only
					if ($flags_bin_ary[28] -eq 1) { 
#Write-Host "made it - fixmemo2"
						
						# HANDLE BLANK OR MULTIPLY MARKED MEMOS
						$test1 = ($db_al_memo.TrimStart($custom_memo_mark).Length -eq 0)
						$test2 = ($db_al_memo.TrimStart($custom_memo_mark) -eq 'NA')
						$test3 = ($db_al_memo.StartsWith($custom_memo_mark+$custom_memo_mark))
						
						# For debugging
						#Write-Host "test1 = $test1;  test2 = $test2;  test3 = $test3"

						if ($test1 -or $test2) {
#Write-Host "made it - fixmemo3"
							
							# Save copy of the original memo (for warnings)
							$db_al_memo_orig = $db_al_memo
						
							# Get the DAT file's confirmed image number and memo
							$confirmed_image_num = $lookup_dat_confimg_num_obj[$path_key]
							$confirmed_image_memo_str = $lookup_dat_confimg_memo_obj[$path_key]

							# For debugging
							#Write-Host ("confirmed_image_num type = " + $confirmed_image_num.GetType())
							#Write-Host "confirmed_image_num = $confirmed_image_num"
							#Write-Host ("confirmed_image_memo_str type = " + $confirmed_image_memo_str.GetType())
							#Write-Host "confirmed_image_memo_str = $confirmed_image_memo_str"
							
							if ($confirmed_image_num -ne 0 -and $confirmed_image_memo_str.Length -ne 0) {
#Write-Host "made it - fixmemo4"
								$found_dat_memo_flg = $true
								# Make image number string (for confirmation MSG)
								#$conf_img_no_padded = ([String]$confirmed_image_num).PadLeft(2,'0')
								$conf_img_str = " (via DAT img #$confirmed_image_num)"
								# Make the new memo string
								$new_memo = $confirmed_image_memo_str

								# Update issues collectors
								$fixed_results_ary += $true
								$lookup_issues_obj[$path_key] = `
									RemoveArrayItems `
									-Haystack $lookup_issues_obj[$path_key] `
									-Needles @('BlankMemo/DatImgMemo mismatch')

							} else {
#Write-Host "made it - fixmemo5"
								$found_dat_memo_flg = $false
								# Make image number string (for confirmation MSG)
								$conf_img_str = ""
								# Make the new memo string
								$new_memo = 'NA'						
							}

							# Mark the memo, if requested.
							if ($apply_custom_memo_mark) {
#Write-Host "made it - fixmemo6"
								if ($custom_memo_mark.Length -gt 0) {
									if ($null -ne $new_memo) {
										#if ($new_memo.Length -gt 0) {
											if (!$new_memo.StartsWith($custom_memo_mark)) {
												$new_memo = $custom_memo_mark + $new_memo
											} else {
												# Note: Method '.TrimStart' will removed multiple occurrences
												$new_memo = $custom_memo_mark + $new_memo.TrimStart($custom_memo_mark, '')
											}
										#}
									}
								}
							}

							# @@@ DEV - For testing
							#$new_memo = $new_memo + "xxx"

							# Update the alert's database record
							
						}

						if ($test3) {
#Write-Host "made it - fixmemo7"
							
							# Save copy of the original memo (for warnings)
							$db_al_memo_orig = $db_al_memo

							$new_memo = $db_al_memo
							
							if ($apply_custom_memo_mark) {
								if ($custom_memo_mark.Length -gt 0) {
									if ($null -ne $new_memo) {
										#if ($new_memo.Length -gt 0) {
											if (!$new_memo.StartsWith($custom_memo_mark)) {
												$new_memo = $custom_memo_mark + $new_memo
											} else {
												# Note: Method '.TrimStart' will removed multiple occurrences
												$new_memo = $custom_memo_mark + $new_memo.TrimStart($custom_memo_mark)
											}
										#}
									}
								}
							}
							
							# @@@ DEV - For testing
							#$new_memo = $new_memo + "xxx"
						}

						if ($test1 -or $test2 -or $test3) {
#Write-Host "made it - fixmemo8"

							# Update the alert's database record
						
							if (Test-Path -LiteralPath "$jpg_filespec") {
#Write-Host "made it - fixmemo9"
								
								# ... returns "success' or 'fail'
								$result_str = `
									UpdateDbRecord `
									-AlertDbPath $db_al_path `
									-AlertDbDate $new_date_sec `
									-AlertDbFlags $db_al_flags `
									-AlertDbMemo $new_memo `
									-UpdateJpgExif `
									-JpgFileSpec $jpg_filespec `
									-FromConsoleFlg $fromConsole `
									-QuietFlg $Quiet `
									-DevModeFlg $DevMode
									
							} else {
#Write-Host "made it - fixmemo10"

								$result_str = `
									UpdateDbRecord `
									-AlertDbPath $db_al_path `
									-AlertDbDate $new_date_sec `
									-AlertDbFlags $db_al_flags `
									-AlertDbMemo $new_memo `
									-FromConsoleFlg $fromConsole `
									-QuietFlg $Quiet `
									-DevModeFlg $DevMode
								
							}

							if ($result_str -match 'success') {
#Write-Host "made it - fixmemo11"
								
								# CONFIRM MEMO CHANGE
								# $exif_obj is hashtable like @{'title: str, 'subject': str, 'comments': str ... }
								$exif_obj = GetJpgExifMetadata -FilePath $jpg_filespec
									
								if ($exif_obj.title -eq $new_memo) {
#Write-Host "made it - fixmemo12"
						
									# Update lookup object to ensure this change is recognized by subsequent actions
									$lookup_memos_obj[$path_key] = $new_memo

									# Update issues collectors
									$fixed_results_ary += $true
									$lookup_issues_obj[$path_key] = `
										RemoveArrayItems `
										-Haystack $lookup_issues_obj[$path_key] `
										-Needles @('Memo/JPG Exif mismatch')
									
									# For debugging
									#Write-Host ("JPG's new exif_title = " + $exif_obj.title)
									
									# Update output arrays
									$tbl_al_memos_ary[$i] = $new_memo
									$tbl_jpg_exif_memos_match_ary[$i] = "match!"
								
									# Add to output csv file collectors
									if (!($csv_fixed_labels_ary -contains $item_clean)) {
										$csv_fixed_labels_ary += $item_clean
										$csv_fixed_paths_ary += $path_key
									}

									if ($fromConsole -and !$Quiet) {
										if ($test1 -or $test2) {
											if ($found_dat_memo_flg) {
												Write-Host "$item_clean - db memo changed from '$db_al_memo_orig' to '$new_memo' (via DAT img #$confirmed_image_num)"
											} else {
												Write-Host "$item_clean - db memo changed from '$db_al_memo_orig' to '$new_memo' (no DAT file available)"
											}
										}
									}
									
								} else {

									# Update issues collectors
									$fixed_results_ary += $false
									#$lookup_issues_obj[$path_key] += "DB memo not changed; unable to update JPG Exif metadata"
									$lookup_issues_obj[$path_key] += "DB memo not changed"
									
									if ($fromConsole) {
										Write-Host "$item_clean - unable to change db memo to '$new_memo'; unable to update JPG Exif metadata" -ForegroundColor Yellow
									}
								}
								
							} else {

								# Update issues collectors
								$fixed_results_ary += $false
								#$lookup_issues_obj[$path_key] += "DB memo not changed; no confirmed images in DAT file"
								$lookup_issues_obj[$path_key] += "DB memo not changed"
								
								if ($fromConsole) {
									Write-Host "$item_clean - unable to change db memo; no confirmed images found in DAT file" -ForegroundColor Yellow
								}
							}

						}

					}
					
				} else {
					$new_memo = $db_al_memo
				}
				
				# MAKE ALERT'S JPG FILE EXIF METADATA MATCH THE ALERT'S DATABASE VALUES
				if (($FixJpgExif -or $FixAll) -and !$zero_date_is_unfixable) {
#Write-Host "made it - fixjpgexif1"

					# For debugging
					#Write-Host ("tbl_al_paths_ary = " + ($tbl_al_paths_ary | ConvertTo-Json))
					#Write-Host ("lookup_memos_obj[$path_key] = " + $lookup_memos_obj[$path_key])
					#Write-Host ("lookup_issues_obj[$path_key] = " + $lookup_issues_obj[$path_key])
					#Write-Host ("tbl_jpg_exif_memos_match_ary = " + ($tbl_jpg_exif_memos_match_ary | ConvertTo-Json))
					#Write-Host ("tbl_jpg_exif_memos_match_ary[$i] = " + $tbl_jpg_exif_memos_match_ary[$i])

					if ($tbl_jpg_exif_memos_match_ary[$i] -ne 'match!') {
#Write-Host "made it - fixjpgexif2"
						
						# PROCEED ONLY IF A JPG FILE EXISTS			
						if ($jpg_file_exists) {
#Write-Host "made it - fixjpgexif3"

							# Update issues collector
							#$lookup_issues_obj[$path_key] += 'Memo/JPG Exif mismatch'
							
							#$db_al_path = $lookup_paths_obj[$path_key]
							#$db_al_memo = [String]$lookup_memos_obj[$path_key]
							#$db_al_flags_orig = $lookup_flags_obj[$path_key]  # array is Int[]
							
							# For debugging
							#Write-Host "path_key = $path_key"
							#Write-Host "db_al_path = $db_al_path"
							#Write-Host "db_al_memo = $db_al_memo"
							#Write-Host "db_al_flags_orig = $db_al_flags_orig"
							#Write-Host "jpg_backed_up = $orig_jpg_backed_up"
							
							# Before updating the Exif metadata...
							# BACKUP THE ORIGINAL JPG, IF NOT DONE ALREADY
							$source_jpg_filespec = $bi_alerts_folder_path + "\" + $lookup_files_obj[$path_key]
							$backup_jpg_filespec = $backup_jpgs_path + "\" + $lookup_files_obj[$path_key]

							# For debugging
							#Write-Host "jpg_filespec = '$jpg_filespec'"
							#Write-Host "source_jpg_filespec = '$source_jpg_filespec'"

							#To avoid losing the original file, we save multiple copies, like 'filename(N).gif'
							$backup_jpg_fs = `
								CopyFileAsSequence `
								-SourceFilespec $source_jpg_filespec `
								-TargetFilespec $backup_jpg_filespec `
								-DestFolder $backup_jpgs_path
								
							# Verify copy
							if ($backup_jpg_fs.StartsWith('ERR')) {
								if ($fromConsole -and !$Quiet) {
									Write-Host "$item_clean - (JPG) Failed to backup file before making changes" -ForegroundColor Yellow
								}
							} else {
								$orig_jpg_backed_up = $true
								if ($fromConsole -and !$Quiet) {
									Write-Host "$item_clean - (JPG) Successfully backed up file before making changes"
								}
							}
								
							# Update the JPG's Exif metadata
							if ($orig_jpg_backed_up) {
#Write-Host "made it - fixjpgexif4"

								# Mark the memo, if requested.
								if ($apply_custom_memo_mark) {
									if ($custom_memo_mark.Length -gt 0) {
										if ($null -ne $new_memo) {
											#if ($new_memo.Length -gt 0) {
												if (!$new_memo.StartsWith($custom_memo_mark)) {
													$new_memo = $custom_memo_mark + $new_memo
												} else {
													# Note: Method '.TrimStart' will removed multiple occurences
													$new_memo = $custom_memo_mark + $new_memo.TrimStart($custom_memo_mark)
												}
											#}
										}
									}
								}

								# Update the alert's database record
								# ... returns "success' or 'fail'
								$result_str = `
									UpdateDbRecord `
									-AlertDbPath $db_al_path `
									-AlertDbDateSec $new_date_sec `
									-AlertDbFlags $db_al_flags `
									-AlertDbMemo $new_memo `
									-UpdateJpgExif `
									-JpgFileSpec $jpg_filespec `
									-FromConsoleFlg $fromConsole `
									-QuietFlg $Quiet `
									-DevModeFlg $DevMode
									
								# For debugging
								#Write-Host "UpdateDbRecord() result_str = '$result_str'"

								if ($result_str -match 'success') {

									# Update issues collectors
									$fixed_results_ary += $true
									$lookup_issues_obj[$path_key] = `
										RemoveArrayItems `
										-Haystack $lookup_issues_obj[$path_key] `
										-Needles @('Memo/JPG Exif mismatch')
									
									# Update output collectors
									$tbl_jpg_exif_memos_ary[$i] = $db_al_memo
									$tbl_al_memos_ary[$i] = $db_al_memo
									$tbl_jpg_exif_memos_match_ary[$i] = "match!"

									# Add to output csv file collectors
									if (!($csv_fixed_labels_ary -contains $item_clean)) {
										$csv_fixed_labels_ary += $item_clean
										$csv_fixed_paths_ary += $path_key
									}
									
									if ($fromConsole -and !$Quiet) {
										Write-Host "$item_clean - (JPG) Exif metadata changed to '$db_al_memo'"
									}						

								} else {
									# Update issues collectors									
									$fixed_results_ary += $false
									$lookup_issues_obj[$path_key] += "JPG Exif not changed"
									if ($fromConsole) {
										Write-Host "$item_clean - (JPG) Failed to update the Exif metadata" -ForegroundColor Yellow
									}
								}
								
							} else {
								# Update issues collectors
								$fixed_results_ary += $false
								$lookup_issues_obj[$path_key] += "JPG Exif not changed"
								if ($fromConsole) {
									Write-Host "$item_clean - (JPG) Failed to update the Exif metadata; unable to backup original JPG file" -ForegroundColor Yellow
								}							
							}
							
						} else {
							# Update issues collectors
							$fixed_results_ary += $false
							$lookup_issues_obj[$path_key] += "JPG file not found"
							# Nothing to do
							#Write-Warning "Unable to find $jpg_filename"
						} 
					}
				}
				
				# MAKE JPG FILE TIMES MATCH ITS EXIF DATE-TAKEN
				# ... This is a good idea, just in case we forget elsewhere to use 
				#     the JPG file's CreationTime instead of its Exif DateTaken
				if ($jpg_file_exists -and !$zero_date_is_unfixable) {
					
					# Returns string like, '01/08/2024 11:10 PM'
					$result_str = `
						SetJpgFileTimes `
						-FilePath $jpg_filespec
						
					if ($result_str = 'success') {
						if ((Get-Item "$jpg_filespec").CreationTime -le (Get-Item "$jpg_filespec").LastWriteTime) {
							$target_date_str = (Get-Item "$jpg_filespec").CreationTime
						} else {
							$target_date_str = (Get-Item "$jpg_filespec").LastWriteTime
						}
					}

					# For debugging
					#Write-Host ("JPG file's final CreationTime = " + (Get-Item "$jpg_filespec").CreationTime)
				}

				# RESTORE DAT & JPG FILES TO READ-ONLY
				if (($FixReadOnly -or $FixAll) -and !$zero_date_is_unfixable) {
#Write-Host "made it - fixreadonly1"

					$cur_attribs_str = $tbl_al_attribs_ary[$i]  # like '', 'D', 'J', 'DJ'

					# For debugging
					#Write-Host "cur_attribs_str = $cur_attribs_str"

					$ro_codes_ary = @()
					
					# JPG FILE
					if (!($cur_attribs_str -match 'J')) {
#Write-Host "made it - fixreadonly2"
						
						if ($jpg_file_exists) {
#Write-Host "made it - fixreadonly3"
						
							# For debugging
							#Write-Host "JPG file is not read-only"
							
							Set-ItemProperty -LiteralPath "$jpg_filespec" -Name IsReadOnly -Value $true
								
							# Add to output csv file collectors
							if (!($csv_fixed_labels_ary -contains $item_clean)) {
								$csv_fixed_labels_ary += $item_clean
								$csv_fixed_paths_ary += $path_key
							}

							# Verify
							$obj = Get-ItemProperty -LiteralPath "$jpg_filespec" | Select-Object IsReadOnly
							if ($obj.IsReadOnly) {
#Write-Host "made it - fixreadonly4"

								$ro_codes_ary += 'J'
								
								# Update issues collectors
								$fixed_results_ary += $true
								$lookup_issues_obj[$path_key] = `
									RemoveArrayItems `
									-Haystack $lookup_issues_obj[$path_key] `
									-Needles @('JPG file not read-only')

								if ($fromConsole -and !$Quiet) {
									Write-Host "$item_clean - (JPG) read-only file attribute enabled"
								}						

							} else {
#Write-Host "made it - fixreadonly5"

								# Update issues collectors
								$fixed_results_ary += $false
								$lookup_issues_obj[$path_key] += "Unable to make JPG file read-only"

								if ($fromConsole) {
									Write-Host "$item_clean - (JPG) Unable to set read-only attribute" -ForegroundColor Yellow

								}
							}

						} else {
Write-Host "made it - fixreadonly6"
							
							# Update issues collectors
							$fixed_results_ary += $false
							$lookup_issues_obj[$path_key] += "JPG file not found"

							if ($fromConsole) {
								Write-Host "$item_clean - (JPG) Could not find file to set read-only attribute: '$jpg_filespec'" -ForegroundColor Yellow
							}
						}

					} else {
						$ro_codes_ary += 'J'
					}
					
					# DAT FILE
					if (!($cur_attribs_str -match 'D')) {

						if ($dat_file_exists) {
						
							# For debugging
							#Write-Host "DAT file is not read-only"

							Set-ItemProperty -LiteralPath "$dat_filespec" -Name IsReadOnly -Value $true
								
							# Update performance csv file collector
							if (!($csv_fixed_labels_ary -contains $item_clean)) {
								$csv_fixed_labels_ary += $item_clean
								$csv_fixed_paths_ary += $path_key
							}

							# Verify
							$obj = Get-ItemProperty -LiteralPath "$dat_filespec" | Select-Object IsReadOnly
							if ($obj.IsReadOnly) {

								$ro_codes_ary += 'D'
								
								# Update issues collectors
								$fixed_results_ary += $true
								$lookup_issues_obj[$path_key] = `
									RemoveArrayItems `
									-Haystack $lookup_issues_obj[$path_key] `
									-Needles @('DAT file not read-only')

								if ($fromConsole -and !$Quiet) {
									Write-Host "$item_clean - (DAT) read-only file attribute enabled"
								}						

							} else {

								# Update issues collectors
								$fixed_results_ary += $false
								$lookup_issues_obj[$path_key] += "Unable to make DAT file read-only"

								if ($fromConsole) {
									Write-Host "$item_clean - (DAT) Unable to set read-only attribute" -ForegroundColor Yellow
								}
								
							}

						} else {
							# Update issues collectors
							$lookup_issues_obj[$path_key] += "DAT file not found"
							$fixed_results_ary += $false
							if ($fromConsole) {
								Write-Host "$item_clean - (DAT) Could not find file to set read-only attribute" -ForegroundColor Yellow
							}
						}

					} else {
						$ro_codes_ary += 'D'
					}
					
					# Update the output 'RO' column value			
					if ($ro_codes_ary.Count -gt 0) {
						$cur_attribs_str_new = $ro_codes_ary -Join ''
						$tbl_al_attribs_ary[$i] = $cur_attribs_str_new
						
						# For debugging
						#Write-Host "updated cur_attribs_str = $cur_attribs_str_new"
						
					}

				}
				
				# SET ALERT'S JSON FLAGS 'FLAG'
				if ($SetFlag) {

					$db_al_flags_orig = $lookup_flags_obj[$path_key]    # array is Int[]
					$flags_bin_ary = $lookup_flags_bin_obj[$path_key]   # like ['@123456': [0,0,1,0,...]

					# For debugging
					#Write-Host "path_key = $path_key"
					#Write-Host "db_al_path = $db_al_path"
					#Write-Host "db_al_memo = $db_al_memo"
					#Write-Host "db_al_flags_orig = $db_al_flags_orig"
					#Write-Host "$path_key flags_bin_ary = $flags_bin_ary"
					
					$flags_bin_ary_new = $flags_bin_ary
					
					if ($flags_bin_ary[1] -ne 1) {
						
						$flags_bin_ary_new[1] = 1
						
						$db_al_flags_new = `
							MakeFlagsDecimalValue `
							-BinAry $flags_bin_ary_new
					
						# For debugging
						#Write-Host "db_al_flags_new = $db_al_flags_new"
					
						# Update the Alert's flags; returns boolean
						$result_flg = `
							SetAlertFlags `
							-AlertDbPath $db_al_path `
							-AlertDbMemo $db_al_memo `
							-AlertDbFlags $db_al_flags_new `
							-QuietFlg $Quiet `
							-FromConsoleFlg $fromConsole
						
						if ($result_flg) {
							if (!$tbl_al_flags_codes_ary[$i] -match 'F') {
								$tbl_al_flags_codes_ary[$i] = "F" + $tbl_al_flags_codes_ary[$i]
							}
							# Add to output csv file collectors
							<# if (!($csv_fixed_labels_ary -contains $item_clean)) {
								$csv_fixed_labels_ary += $item_clean
								$csv_fixed_paths_ary += $path_key
							} #>

							if ($fromConsole -and !$Quiet) {
								Write-Host "$item_clean - Alert's ['flags'] value changed from '$db_al_flags_orig' to '$db_al_flags_new'"
							}						

						} else {
							# Update issues collector
							$lookup_issues_obj[$path_key] += "Alert's 'Flag' flag not changed"
							Write-Host "$item_clean - Unsuccessful attempt to change the Alert's 'Flag' flag" -ForegroundColor Yellow
						}
					}
					
				}
			
				# Update output issues array for output csv file
				#if (!$zero_date_is_unfixable) {
					$al_issues_str = ($lookup_issues_obj[$path_key] | Select -Unique) -Join '; '
					$tbl_al_issues_ary += $al_issues_str
				#}
				
				# For debugging
				#Write-Host ("fixed_results_ary = " + ($fixed_results_ary | ConvertTo-Json))
				#Write-Host ("lookup_issues_obj[$path_key] = " + ($lookup_issues_obj[$path_key] | ConvertTo-Json))
				
				# ASSESS OVERALL RESULT
				if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
					if ($fixed_results_ary -contains $false) {
						if ($fixed_results_ary -contains $true) {
							$al_issues_fixed_ary += 'Partial'
						} else {
							$al_issues_fixed_ary += 'N'
						}
					} else {
						$al_issues_fixed_ary += 'Y'
					}				
					# Update issues accumulator
					$lookup_fixed_yn_obj[$path_key] = $al_issues_fixed_ary
				
					if ($fromConsole) {
						if ($fixed_results_ary -contains $true) {
							#Write-Host ""
						}
					}
				}
				
				# PROCESS THE NEXT PROTECTED ALERT
				$i++
				
			}

			# PERFORM DATABASE REPAIR/REGENERATE
			if ($added_file_flg) {

				if (!$NoRegen) {
					
					# BUT FIRST, BACKUP THE BLUE IRIS DATABASE FOLDER
					# ... Remove excessive sub-folders to keep folder size in check (15 sub-folders max)
					#     REF: https://stackoverflow.com/a/69656558
					#Get-ChildItem -Directory $backup_db_path | Sort-Object CreationTime | Select-Object -SkipLast 14 | Remove-Item -Recurse -WhatIf
					Get-ChildItem -Directory $backup_db_path | Sort-Object CreationTime | Select-Object -SkipLast 14 | Remove-Item -Recurse
					# ... Copy db folder to dated subfolder in $achive_path
					# Using current system time
					#$date_str = [DateTime]::Now.ToString("yyyyMMdd_HHmmss")
					#Copy-Item -LiteralPath "$bi_db_path" -Destination $("$backup_db_path\$date_str") -Recurse
					# Using clips.dat file's last write time
					$obj = Get-ChildItem -Path "$bi_db_path\clips.dat" | select LastWriteTime 
					$date_str = ($obj.LastWriteTime).ToString("yyyyMMdd_HHmmss")
					Copy-Item -LiteralPath "$bi_db_path" -Destination $("$backup_db_path\$date_str") -Recurse
					
					# DO IT
					$http_response_obj = `
						BIHttpCall `
						-HttpCmd "/admin?db=rebuild"		
					
					if ($fromConsole) {
						Write-Host ""
						Write-Host "HEADS-UP: Database Repair/Regenerate action performed (as req'd for file changes in BI-managed folders)" -ForegroundColor Green
					}

				} else {
					if ($fromConsole) {
						Write-Host ""
						Write-Warning "Automatic Database Repair/Regenerate action disabled via -NoRegen argument; YOU ARE RESPONSIBLE"
					}
				}
				
			}			
		}
	}
}
# ===========================================================================

# Restore disable Blue Iris database maintenance
# ... if originally enabled
if ($disable_clip_maintenance -and $clip_maintenance_enabled_flg) {

	$response = `
		ToggleDbMaintenance `
		-TargetState 'On'
		
	if ($response -eq 'success') {
		if ($fromConsole -and !$Quiet) {
			Write-Host ""
			Write-Host "Blue Iris database setting 'Compact/repair DB at' has been re-enabled" -Foregroundcolor Yellow
		}
	} else {
		if ($fromConsole) {
			Write-Warning "Failed to restore Blue Iris database setting 'Compact/repair DB at'"
		}
	}
}

# For debugging
#Write-Host ("tbl_al_issues_ary AFTER = " + ($tbl_al_issues_ary | ConvertTo-Json))
#Write-Host ("csv_fixed_paths_ary type = " + $csv_fixed_paths_ary.GetType())
#Write-Host ("csv_fixed_paths_ary cnt = " + $csv_fixed_paths_ary.Count)
#Write-Host ("csv_fixed_paths_ary = " + ($csv_fixed_paths_ary | ConvertTo-Json))
#Write-Host ("csv_fixed_labels_ary = " + ($csv_fixed_labels_ary | ConvertTo-Json))
#Write-Host ("csv_fixed_flags_ary = " + ($csv_fixed_flags_ary | ConvertTo-Json))
#Write-Host ("csv_fixed_issues_ary = " + ($csv_fixed_issues_ary | ConvertTo-Json))

# RECORD SCRIPT FINDINGS/ACTIONS TO SCRIPT'S OUTPUT CSV FILE ('FIXED ALERTS' LOG)
$csv_fixed_alerts_log_fs = $app_path + "\fixed_alerts_log.csv"
if ($csv_fixed_paths_ary.Count -gt 0) {

	# Build the summary statement & overall result for each fixed Alert
	[String[]]$csv_fixed_issues_ary = @()
	[String[]]$csv_fixed_yn_ary = @()
	foreach ($item in $csv_fixed_paths_ary) {
		$path_key = $item
		[String[]]$arr = @()
		foreach ($x in $lookup_issues_obj[$path_key]) {
			if (!($arr -contains $x)) {
				$arr += $x
			}
		}
		$csv_fixed_issues_ary += $arr -Join '; '
		$csv_fixed_yn_ary = $lookup_fixed_yn_obj[$path_key]
	}

	# ... Prep params for logging
	$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}
	if ($params_str.Length -eq 0) {
		$script_args_str = "<no args>"
	} else {
		$script_args_str = $params_str
	}

	# For debugging
	#Write-Host "`nPREP FOR THE 'FIXED ALERTS' CSV LOG ($csv_fixed_alerts_log_fs)..."
	#Write-Host ("csv_fixed_paths_ary cnt = " + $csv_fixed_paths_ary.Count)
	#Write-Host ("csv_fixed_paths_ary = " + ($csv_fixed_paths_ary | ConvertTo-Json))
	#Write-Host ("csv_fixed_labels_ary cnt = " + $csv_fixed_labels_ary.Count)
	#Write-Host ("csv_fixed_labels_ary = " + ($csv_fixed_labels_ary | ConvertTo-Json))
	#Write-Host ("csv_fixed_issues_ary cnt = " + $csv_fixed_issues_ary.Count)
	#Write-Host ("csv_fixed_issues_ary = " + ($csv_fixed_issues_ary | ConvertTo-Json))
	#Write-Host ("csv_fixed_yn_ary cnt = " + $csv_fixed_yn_ary.Count)
	#Write-Host ("csv_fixed_yn_ary = " + ($csv_fixed_yn_ary | ConvertTo-Json))
	
	# Proceed ONLY if all arrays to be logged are the same length
	if(($csv_fixed_issues_ary.Count, $csv_fixed_yn_ary.Count | Where {$_ -eq $csv_fixed_labels_ary.Count})) {

		# DO IT
		# ... returns 'success' or 'ERR/..."
		$csv_response = `
			LogScriptResults `
			$csv_fixed_alerts_log_fs `
			$timestamp `
			$csv_fixed_labels_ary `
			$csv_fixed_issues_ary `
			$csv_fixed_yn_ary `
			$script_host_str `
			$script_args_str 

		# For debugging
		#Write-Host "csv csv_response = $csv_response"

		if ($csv_response) {
			if ($csv_response.StartsWith("ERR/")) {
				if ($fromConsole) {
					Write-Warning "unable to save 'fixed alerts' CSV log (Reason: $csv_response)"
				}
			}
		} else {
			if ($fromConsole) {
				Write-Warning "unable to save 'fixed alerts' CSV log (Reason unknown)"
			}
		}
		
	} else {
		if ($fromConsole) {
			Write-Warning "unable to save 'fixed alerts' CSV log (Reason: columns not equal length) ($csv_fixed_alerts_log_fs)"
		}
	}
}

# OUTPUT FINAL RESULTS
$show_table_flg = $false  # Default
$large_table_flg = $false  # Default

if ($alerts_filtered_cnt -gt 0) {
	
	# Build 'issues' arrays for both the display table and the output CSV file
	$al_issues_disp_ary = @()  # Initialize collector
	$i = 0
	foreach ($item in $tbl_al_issues_ary) {
		if ($item.Length -gt 0) {
			if ($item -ne '--') {
				$al_issues_disp_ary += "Yes"
			} else {
				$al_issues_disp_ary += "No"
			}
		} else {
			$al_issues_disp_ary += "No"
			
			# Replace the blank value in the output array
			# ... Why use '--' instead of 'No'? 
			#     Because this facilitates CSV file column filtering by the user in external CSV viewers
			#     (the word 'No' is also used in found issues)
			$tbl_al_issues_ary[$i] = '--'
		}
		$i++
	}

	# For debugging
	#Write-Host ("al_issues_disp_ary cnt = " + $al_issues_disp_ary.Count)
	#Write-Host ("al_issues_disp_ary = " + ($al_issues_disp_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_issues_ary cnt = " + $tbl_al_issues_ary.Count)
	#Write-Host ("tbl_al_issues_ary = " + ($tbl_al_issues_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_haszerodates_ary = " + ($tbl_al_haszerodates_ary | ConvertTo-Json))
	#Write-Host ("tbl_dat_memos_ary = " + ($tbl_dat_memos_ary | ConvertTo-Json))
	#Write-Host ("tbl_dat_memo_image_nos_ary = " + ($tbl_dat_memo_image_nos_ary | ConvertTo-Json))
	
	$i = 0
	foreach($tbl_al_memo in $tbl_al_memos_ary) {
		if ($tbl_al_memo.TrimStart($custom_memo_mark).Length -eq 0) {
			$tbl_al_memos_ary[$i] = '<blank>'
		}
		$i++
	}
	#Write-Host ("tbl_dat_memos_ary AFTER = " + ($tbl_dat_memos_ary | ConvertTo-Json))

	# 1. MAKE CSV FILE FOR CONSOLE DISPLAY
	if ($fromConsole) {
		$output_csv_filespec_disp = `
			MakeTable `
			-Path "$output_path" `
			-SortColNo 1 `
			-SortAscFlg $SortAsc `
			-LabelsList $tbl_al_labels_ary `
			-AlertFlagsList $tbl_al_flags_codes_ary `
			-AlertJpgsList $tbl_al_hasjpgs_ary `
			-AlertDatsList $tbl_al_hasdats_ary `
			-AlertAttribsList $tbl_al_attribs_ary `
			-AlertZeroDatesList $tbl_al_haszerodates_ary `
			-AlMemosList $tbl_al_memos_ary `
			-JpgExifMemos $tbl_jpg_exif_memos_match_ary `
			-DatMemosList $tbl_dat_memos_ary `
			-DatMemoImageNosList $tbl_dat_memo_image_nos_ary `
			-IssuesList $al_issues_disp_ary `
			-TimestampFlg $false
	}

	# 2. MAKE FULL-DETAIL CSV FILE (FOR USER'S CSV VIEWER)
	$output_csv_filespec_full = `
		MakeTable `
		-Path "$output_path" `
		-SortColNo 1 `
		-SortAscFlg $SortAsc `
		-LabelsList $tbl_al_labels_ary `
		-AlertFlagsList $tbl_al_flags_codes_ary `
		-AlertJpgsList $tbl_al_hasjpgs_ary `
		-AlertDatsList $tbl_al_hasdats_ary `
		-AlertAttribsList $tbl_al_attribs_ary `
		-AlertZeroDatesList $tbl_al_haszerodates_ary `
		-AlMemosList $tbl_al_memos_ary `
		-JpgExifMemos $tbl_jpg_exif_memos_match_ary `
		-DatMemosList $tbl_dat_memos_ary `
		-DatMemoImageNosList $tbl_dat_memo_image_nos_ary `
		-IssuesList $tbl_al_issues_ary `
		-TimestampFlg $timestamp_output_csv_files_flg

	# 3. OUTPUT TO CONSOLE
	if ($fromConsole) {

		if (!$HideTable) {

			if ($alerts_filtered_cnt -gt 20) {
				# For large tables, give the user the option of displaying them in the Console
				# ... Remember, results are always saved to '$app_path\_output\output.csv'
				Write-Host ""
				$show_table = Read-Host -Prompt "Show table (Y/N)?"
				$show_table_flg = ($show_table -match "Y") 
				$large_table_flg = $true
			} else {
				# Show small tables automatically
				$show_table_flg = $true
			}
			
			if ($show_table_flg) {
				
				# Provide feedback to user
				if (!$Quiet) {
					Write-Host ""
					Write-Host "TABULATED RESULTS: (for details on Issues column 'Yes' entries, see this CSV file:"
					Write-Host "$output_csv_filespec_full)"
					if ($alerts_filtered_cnt -gt 20) {
						if ($ParentProcess -ne "explorer") {
							#if ($Host.UI.RawUI.WindowSize.Height -lt 25) {
								Write-Host "(Note: Hit space bar to show subsequent pages, or Ctrl-C to abort)"
							#}
						}
					}
				}

				# Show table
				if ($ParentProcess -eq "explorer") {

					# RUNNING FROM POWERSHELL-ISE
					ShowTable `
						-Path $output_csv_filespec_disp `
						-SortColNo 1 `
						-SortAscFlg $SortAsc `
						
				} else {

					# *NOT* RUNNING FROM POWERSHELL-ISE

					# Paginate the output
					<# (ShowTable `
						-Path $output_csv_filespec_disp `
						-SortColNo 1 `
						-SortAscFlg $SortAsc ) | Out-Host -Paging #>

					# Alternate pagination method
					(ShowTable `
						-Path $output_csv_filespec_disp `
						-SortColNo 1 `
						-SortAscFlg $SortAsc ) | More
						
				}
				
				if ($auto_open_csv_file) {
					if ($alerts_filtered_cnt -gt $auto_open_csv_file_min_rows) {						
						if ($auto_open_csv_file_in_def_app) {
							Invoke-Expression "$output_csv_filespec_full"
						} else {
							# Open Out-GridView cmdlet UI
							#Import-Csv -LiteralPath "$output_csv_filespec_full" | Out-GridView -Title $output_csv_filespec_full -Wait
							Import-Csv -LiteralPath "$output_csv_filespec_full" | Out-GridView -Title $output_csv_filespec_full
						}
					}
				}

			} else {
				if (!$Quiet) {
					Write-Host "Results saved to file '$output_csv_filespec_full'"
				}
			}
		} else {
			Write-Host (($alerts_cnt).ToString() + " Alert files found via the requested filters")
			if (!$Quiet) {
				Write-Host "`nResults saved to file '$output_csv_filespec_full'"
			}
		}
		
		if (Test-Path -LiteralPath "$output_csv_filespec_disp") {
			Remove-Item -Path "$output_csv_filespec_disp"
		}
	}
} else {
		Write-Host "`nNothing to show"
}

# Stop performance timer
$timer_script_runtime.Stop()
$runtime_elapsed_sec = ($timer_script_runtime.elapsed.totalseconds).ToString("0.00")

if ($DevMode) {
	if ($fromConsole) {
		Write-Host "(dev mode) script runtime = $runtime_elapsed_sec sec" -ForegroundColor Cyan
	}
}

# POST MESSAGE TO BLUE IRIS LOGFILE
if (!$NoBiLog -and $post_to_BI_logfile_flg) {
		
	# For testing
	#$fromConsole = $false
	#$post_to_BI_logfile_in_background_only_flg = $false
	#$post_to_BI_logfile_in_background_only_flg = $true

	# Optionally, messages are posted only if this script is running in the background  
	# (e.g., via Task Scheduler, or Blue Iris actions)
	if ($post_to_BI_logfile_in_background_only_flg -and $fromConsole) {
		$doit_flg = $false
	} else {
		$doit_flg = $true
	}

	# For debugging
	#Write-Host "doit_flg = $doit_flg"	
	#Write-Host ("al_issues_before_ary = " + ($al_issues_before_ary | ConvertTo-Json))
	#Write-Host ("tbl_al_issues_ary = " + ($tbl_al_issues_ary | ConvertTo-Json))

	# Initialize counters
	$issues_cntr = 0
	$remaining_issues_cntr = 0
	$fixed_issues_cntr = 0
	
	if ($doit_flg) {

		# Get count of alerts with issues and fixes
		foreach ($item in $al_issues_before_ary) {
			if ($item -ne '--') {  # '--' = 'No'
				$issues_cntr++
			}
		}
		foreach ($item in $tbl_al_issues_ary) {
			if ($item -ne '--') {  # '--' = 'No'; $al_issues_disp_ary contains 'Yes' &'No'
				$remaining_issues_cntr++
			}
		}		
		if ($remaining_issues_cntr -gt 0) {
			$fixed_issues_cntr = $issues_cntr - $remaining_issues_cntr
		} else {
			$fixed_issues_cntr = $issues_cntr
		}
		
		# For debugging
		#Write-Host "issues_cntr = $issues_cntr"	
		#Write-Host "remaining_issues_cntr = $remaining_issues_cntr"	
		#Write-Host "fixed_issues_cntr = $fixed_issues_cntr"	
		
		# BUILD BI LOGFILE MESSAGE
		# ... Include the script args used
		if ($params_str.Length -eq 0) {
			$script_args_str = "no args"
		} else {
			$script_args_str = $params_str
		}
		# ... Include a summary statement
		$summary_str = "Inspected $alerts_filtered_cnt protected alerts;"
		if (!($fix_type_arg_used_flg -or $SetFlag)) {
			if ($issues_cntr -gt 0) {
				if ($issues_cntr -ne 1) {
					$summary_str += " $issues_cntr w/ issues detected (for details, see '$output_csv_filespec_full')"
				} else {
					$summary_str += " $issues_cntr w/ an issue detected (for details, see '$output_csv_filespec_full')"
				}
			} else {
				$summary_str += " no issues detected"
			}
		} else {
			if ($issues_cntr -gt 0) {
				$summary_str += " fixed " + $fixed_issues_cntr + "/" + $issues_cntr + " issues detected (for details, see '$csv_fixed_alerts_log_fs')"
			} else {
				$summary_str += " no issues detected"
			}
		}
		# ... Finish message string
		$msg_str = "$ScriptBaseName [$script_args_str] [$runtime_elapsed_sec sec] $summary_str "
		# ... remove/replace incompatible substrings
		$msg_str = $msg_str -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"
		
		# Build URL
		$log_level = 0
		if ($post_w_creds_flg) {
			$uri = "http://$bi_webserver/admin?log=$msg_str&level=$log_level&user=$bi_user_name&pw=$bi_user_pw"
		} else {
			$uri = "http://$bi_webserver/admin?log=$msg_str&level=$log_level"
		}

		# Do it
		try {
			$web_response_obj = `
				Invoke-WebRequest `
				-Uri $uri
		} 
		catch {

			# For debugging
			#Write-Host $error
			#Write-Host $_.ScriptStackTrace
			#Write-Host $PSitem

			Write-Host $_.Exception.Message -ForegroundColor Red
			#$err_msg = "ERR/unexpected HTTP Interface command response"
			#return $err_msg
		}
		
		# For debugging
		#Write-Host "msg_str = $msg_str"
		#Write-Host "uri = $uri"
		#Write-Host "web_response: $web_response_obj"	
			
		if ($web_response_obj -match 'signal=' -and $web_response_obj -match 'schedule=' -and $web_response_obj -match 'profile=') {
			# Normal response
			if ($fromConsole -and !$Quiet) {
				if (!$show_table_flg) {
					Write-Host ""
				}
				Write-Host "Message posted to Blue Iris logfile"
				#Write-Host "Message posted to Blue Iris logfile:`n'$log_msg'"
			}
		} else {
			# Unexpected response
			if ($fromConsole) {
				Write-Warning "Failed to post message to Blue Iris logfile; $web_response_obj"
			} else {
				msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nFailed to post message to Blue Iris logfile"
			}
		}
	} else {
		if ($fromConsole) {
			Write-Warning "No message posted to the Blue Iris logfile (not running in background)"
		}
	}
} else {
	if ($fromConsole -and !$Quiet) {
		#Write-Host "No message posted to the Blue Iris logfile"
	}
}

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if ($enable_bg_messages) {
	if (!$fromConsole) {
		$msg_sec = 600
		$bg_msg = "Script '$ScriptName' is finished" + 
		"`nscript runtime = $runtime_elapsed_sec sec" + 
		"`nprotected alerts checked = $alerts_filtered_cnt" + 
		"`nprotected alerts w/ issues = $issues_cntr" + 
		"`nprotected alerts w/ issues fixed = $fixed_issues_cntr" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		msg.exe $server_username /TIME:$msg_sec $bg_msg
	}
}

# PERFORM MAINTENANCE ON FILE ARCHIVE FOLDERS
if (1 -eq 1) {  # Use to collapse this lengthy code section while editing
	if ($fromConsole -and !$Quiet) {
		Write-Host ""
		Write-Host "/////////////////////////////////////////////////////////////////////////////////"
		Write-Host "Archived output files maintenance summary:"
	}

	$purge_filespec = $output_path + "\*_20*.csv"

	# Note: $response_obj may be a string if function returns an error; otherwise expect string[].
	$response_obj = `
		PurgeOldFiles `
		-Pattern $purge_filespec `
		-Days $archive_lifetime_days
		
	#Write-Host "purge_filespec = '$purge_filespec'"
	#Write-Host ("response_obj type = " + $response_obj.GetType())
	#Write-Host ("response_obj cnt = " + $response_obj.Count)
	#Write-Host ("response_obj = " + ($response_obj | ConvertTo-Json))

	if ($response_obj -is [Array]) {
		if ($fromConsole -and !$Quiet) {
			Write-Host ("Count of archived CSV files matching '$purge_filespec' = " + $response_obj[0])
			if ($timestamp_output_csv_files_flg -and $archive_lifetime_days -gt 0) {
				Write-Host ("Count of archived CSV files pruned (age >= $archive_lifetime_days days) = " + $response_obj[1])
			} else {
				if ($fromConsole) {
					Write-Host "NOTE: Archive auto-pruning disabled by setting `$archive_lifetime_days = 0"
				}
			}
		}
	} else {
		if ($fromConsole) {
			Write-Warning "CSV file archive management issue`nPurgeOldFiles() returned: $response_obj"
		} else {
			$rtn_status_str = 'Warning - CSV file archive management issue'
		}
	}
}

# SHOW BACKGROUND MODE RESPONSE
if ($ShowCLI) {
	Write-Host "~~~~~`nOutput as returned from background..."
	@($tbl_al_labels_ary, $tbl_al_flags_ary, $tbl_al_hasjpgs_ary, $tbl_al_hasdats_ary, $tbl_al_attribs_ary, $tbl_al_memos_ary, $tbl_dat_memos_ary, $tbl_dat_memo_image_nos_ary, $tbl_jpg_exif_memos_match_ary) | ConvertTo-Json	
} else {
	if (!$fromConsole) {
		return $tbl_al_labels_ary, $tbl_al_flags_ary, $tbl_al_hasjpgs_ary, $tbl_al_hasdats_ary, $tbl_al_attribs_ary, $tbl_al_memos_ary, $tbl_dat_memos_ary, $tbl_dat_memo_image_nos_ary, $tbl_jpg_exif_memos_match_ary
	}
}

if ($fromConsole -and !$Quiet) {
	Write-Host ""
}

#Stop-Transcript